<?php
/**
 * YANA library
 *
 *
 * Software:        Yana PHP-Framework 3.0.0
 * Author:          Dipl. Inf. Thomas Meyer
 * Homepage:        http://www.yanaframework.net
 * Copyright:       (c) 2009  Thomas Meyer
 * License:         GNU GPL  http://www.gnu.org/licenses/
 *
 * This program: can be redistributed and/or modified under the
 * terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see http://www.gnu.org/licenses/.
 *
 * You may use this software for free. In case you like it, or want to
 * use it for a commercial purpose, please consider making a donation.
 * http://yanaframework.net/donate.php
 *
 * This notice MAY NOT be removed.
 *
 *
 * @package  yana
 * @license  http://www.gnu.org/licenses/gpl.txt
 */

/**
 * counter wrapper
 *
 * This class implements persistent counters - optional with IP checking -
 * for various purposes: e.g. visitor counts, statistics, polls.
 *
 * @access      public
 * @package     yana
 * @subpackage  utilities
 */
class Counter extends Object
{
    /**#@+
     * @access private
     * @ignore
     */

    /** @var string  */ var $directory = "";
    /** @var SML     */ var $file      = null;
    /** @var SML     */ var $ip_file   = null;
    /** @var boolean */ var $use_ip    = true;

    /**#@-*/

    /**
     * create a new instance
     *
     * Use the $use_ip parameter with the setting YANA_COUNTER_USE_IP
     * to create a counter with IP checking (does'nt change on reload)
     * or with the setting YANA_COUNTER_IGNORE_IP to create a counter
     * without IP checking (always changes when an update is called).
     *
     * @param  string  $filename  name of file to safe values to
     * @param  int     $use_ip    count on reload on / off
     */
    function Counter($filename = "", $use_ip = YANA_COUNTER_USE_IP)
    {
        if (!is_string($filename)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($filename)), E_USER_WARNING);
            return false;
        }
        if (!is_int($use_ip)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 2, 'Integer', gettype($use_ip)), E_USER_WARNING);
            return false;
        }

        /* settype to STRING */
        $filename = (string) $filename;
        global $YANA;

        $filename = strtolower(preg_replace("/[^\w\d_]/", "", $filename));
        if (empty($filename)) {
            /* default value */
            $filename = "statistik";
        }

        if (isset($YANA)) {
            $this->directory = $YANA->getVar("COUNTERDIR");
        } else {
            $this->directory = 'config/counter/';
        }
        $this->file = new SML($this->directory.$filename.".cnt", CASE_UPPER);
        $this->ip_file = new SML($this->directory.$filename.".ip", CASE_UPPER);
        $this->use_ip = ($use_ip === YANA_COUNTER_USE_IP);

        if ($this->use_ip && $this->ip_file->exists() && filemtime($this->ip_file->getPath()) + 10000 < time()) {
            /* flush old data */
            $this->ip_file->read();       /* read */
            $this->ip_file->set(array()); /* flush */
            $this->ip_file->write();      /* write empty file */
        }
    }

    /**
     * get counter
     *
     * Returns the current counter settings as:
     *
     * <pre>
     * Array(
     *    'COUNT' => int
     *    'INFO'  => string
     * )
     * </pre>
     *
     * Note: you may use the wildcard '*' to get
     * all counters as an associative, multidimensional array:
     *
     * <pre>
     * Array(
     *    ID1 => Array(
     *        'COUNT' => int
     *        'INFO'  => string
     *    )
     *    ID2 => Array(
     *        'COUNT' => int
     *        'INFO'  => string
     *    )
     *    ...
     * )
     * </pre>
     *
     * @access  public
     * @param   string  $id  counter id
     * @return  array
     */
    function get($id = "*")
    {
        if (!is_string($id)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($id)), E_USER_WARNING);
            return false;
        }

        /* settype to STRING */
        $id = (string) $id;
        if (!preg_match("/^([\w\d_]+|\*)$/i", $id)) {
            /* syntax error */
            return false;
        } else {
            /* all fine - continue */
        }

        if (empty($id) || $id == "*") {
            /* return all */
            $result = $this->file->get();
        } else {
            /* return single entry */
            $result = $this->file->get($id);
        }
        assert('is_array($result); /* unexcpected result */');
        return $result;

    }

    /**
     * get counter info
     *
     * The 'info' field is an optional text value, that
     * describes the counter.
     *
     * This function returns the current description of counter $key.
     *
     * Note: you may use the wildcard '*' to get
     * an associative array for all counters:
     *
     * <pre>
     * Array(
     *    ID1 => string
     *    ID2 => string
     *    ...
     * )
     * </pre>
     *
     * @access  public
     * @param   string  $id  counter id
     * @return  mixed
     */
    function getInfo($id = "*")
    {
        if (!is_string($id)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($id)), E_USER_WARNING);
            return false;
        }

        /* settype to STRING */
        $id = (string) $id;
        if (!preg_match("/^([\w\d_]+|\*)$/i", $id)) {
            /* syntax error */
            return false;
        } else {
            /* all fine - continue */
        }

        if (empty($id) || $id == "*") {
            /* return all */
            $A = $this->file->get();
            $B = array();
            assert('!isset($a); /* cannot redeclare variable $a */');
            assert('!isset($b); /* cannot redeclare variable $b */');
            foreach ( $A as $a => $b )
            {
                if(!empty($b['INFO'])) {
                    $B[$a] = $b['INFO'];
                }
            }
            unset($a,$b);
            asort($B, SORT_STRING);
            return $B;
        } else {
            /* return single entry */
            return $this->file->get("$id.INFO");
        }
    }

    /**
     * get counter value
     *
     * This function returns the current value of counter $key.
     *
     * Note: you may use the wildcard '*' to get
     * an associative array for all values:
     *
     * <pre>
     * Array(
     *    ID1 => int
     *    ID2 => int
     *    ...
     * )
     * </pre>
     *
     * @access  public
     * @param   string  $id  counter id
     * @return  mixed
     */
    function getCount($id = "*")
    {
        if (!is_string($id)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($id)), E_USER_WARNING);
            return false;
        }

        /* settype to STRING */
        $id = (string) $id;
        if (!preg_match("/^([\w\d_]+|\*)$/i", $id)) {
            /* syntax error */
            return false;
        } else {
            /* all fine - continue */
        }

        /**
         * Note: there are two synopsis here.
         * 1) return all entries
         * 2) return a single entry, identified by $id
         */
        if (empty($id) || $id == "*") {
            /* return all entries */
            $A = $this->file->get();
            $B = array();
            assert('!isset($a); /* cannot redeclare variable $a */');
            assert('!isset($b); /* cannot redeclare variable $b */');
            foreach ($A as $a => $b)
            {
                $B[$a] = $b['COUNT'];
            }
            unset($a,$b);
            asort($B, SORT_NUMERIC);
            return $B;
        } else {
            /* return single entry */
            return (int) $this->file->get("$id.COUNT");
        }
    }

    /**
     * Increment/Decrement counter
     *
     * Adds $ammount (defaults to +1) to the counter $id and sets
     * the counter description (the 'info' field)
     * to $info (defaults to "").
     *
     * Note: as you might already have guessed, using a negative
     * value for $ammount decrements the counter.
     *
     * @access  public
     * @param   string  $id       counter id
     * @param   string  $info     optional description to be safed along with the counter
     * @param   int     $ammount  number to add to counter
     * @return  int
     */
    function count($id, $info = "", $ammount = 1)
    {
        if (!is_string($id)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($id)), E_USER_WARNING);
            return false;
        }
        if (!is_string($info)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 2, 'String', gettype($info)), E_USER_WARNING);
            return false;
        }
        if (!is_int($ammount)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 3, 'Integer', gettype($ammount)), E_USER_WARNING);
            return false;
        }

        /* settype to STRING
         *            STRING
         *            INTEGER
         */
        $id      = (string) $id;
        $info    = (string) $info;
        $ammount = (int)    $ammount;

        if (!preg_match("/^[\w\d_]+$/i", $id)) {
            /* syntax error */
            return false;
        }

        /* BEGIN count */
        if (!$this->file->exists()) {
            /* create on first access */
            $this->file->create();
        } else {
            /* continue with existing file */
        }
        $this->file->read();
        $NEW_COUNTER                = array();
        $NEW_COUNTER[$id]           = $this->file->get($id);
        /* settype to INTEGER */
        $NEW_COUNTER[$id]['COUNT'] = (int) $NEW_COUNTER[$id]['COUNT'];
        $NEW_COUNTER[$id]['COUNT'] += $ammount;
        if (!empty($info)) {
            /* add info when avialable */
            $NEW_COUNTER[$id]['INFO'] = $info;
        } else {
            /* Keep info unchanged.
             * Note: info can carry a previously set value
             * and is not necessarily NULL at this point.
             */
        }
        /* END count */

        /* reload detection */
        if ($this->use_ip) {
            /* ip */
            if (!$this->ip_file->exists()) {
                /* create on first access */
                $this->ip_file->create();
            } else {
                /* continue with existing file */
            }
            $this->ip_file->read();
            if (isset($_SERVER['REMOTE_ADDR'])) {
                $ip = preg_replace("/[^\w\d_]/", "_", $_SERVER['REMOTE_ADDR']);
            } else {
                /* This will certainly happen when running the script via the CLI,
                 * e.g. by running a cron-job.
                 *
                 * Note: this is not set to 127.0.0.1 since the remote address
                 * is really undefined - we just guess that it MIGHT be localhost,
                 * but we can't be sure about it.
                 */
                $ip = '0.0.0.0';
            }
            $last_visit       = (int) $this->ip_file->get("$id.$ip");
            $NEW_IP           = array();
            $NEW_IP[$id][$ip] = time() + 10000;
        } else {
            /* continue without checking the remote address for page-reloads */
        }

        if (!$this->use_ip || $last_visit < time()) {
            /* if ip is ignored or no recent entry for this ip found */
            $this->file->insert($NEW_COUNTER);

            if ($this->file->write()) {
                /* write changes to file */

                /* The option "use_ip" triggers wether to
                 * use a counter with IP logging to detect
                 * page reloads (true) or not (false).
                 */
                if ($this->use_ip) {
                    /* save remote address for later review */
                    $this->ip_file->insert($NEW_IP);
                    $this->ip_file->write();
                } else {
                    /* continue without saving the remote address */
                }
                return $NEW_COUNTER[$id]['COUNT'];
            } else {
                trigger_error("Unable to update counter '${id}'.", E_USER_WARNING);
            }
        } else {
            /* page reload detected - file remains unchanged */
        }

        $count = $this->getCount($id);
        assert('is_numeric($count); // Unexpected result: $count.');
        return (int) $count;
    }

    /**
     * compare with another object
     *
     * Returns bool(true) if this object and $anotherObject
     * are equal and bool(false) otherwise.
     *
     * Two instances are considered equal if and only if
     * they are both objects of the same class, they
     * both refer to the same filesystem resource and use
     * the same IP settings.
     *
     * @access   public
     * @param    object  $anotherObject  any object or var you want to compare
     * @return   string
     */
    function equals($anotherObject)
    {
        if (is_object($anotherObject) && method_exists($anotherObject,'getClass') && $anotherObject->getClass() === $this->getClass()) {
            if ($this->file->equals($anotherObject->file) && $this->use_ip === $anotherObject->use_ip) {
                return true;
            } else {
                return false;
            }
        } else {
            return false;
        }
    }

}
// Copyright  2009  Thomas Meyer  http://www.yanaframework.net
?>