<?php
/**
 * YANA library
 *
 *
 * Software:        Yana PHP-Framework 3.0.0
 * Author:          Dipl. Inf. Thomas Meyer
 * Homepage:        http://www.yanaframework.net
 * Copyright:       (c) 2009  Thomas Meyer
 * License:         GNU GPL  http://www.gnu.org/licenses/
 *
 * This program: can be redistributed and/or modified under the
 * terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see http://www.gnu.org/licenses/.
 *
 * You may use this software for free. In case you like it, or want to
 * use it for a commercial purpose, please consider making a donation.
 * http://yanaframework.net/donate.php
 *
 * This notice MAY NOT be removed.
 *
 *
 * @package  yana
 * @license  http://www.gnu.org/licenses/gpl.txt
 */

/**
 * This class is an alternative to csv files.
 *
 * Each line represents a row of data.
 * Tags are used to mark up the fields.
 *
 * The use of tag delimiters inside field values is not allowed.
 * If a tag is found in a value it automatically gets removed.
 *
 * The structure is not as strict as you might know from CSV files,
 * which means the number and names of fields inside a row may
 * vary from one row to another.
 *
 * This provides you with more flexibility. For example:
 * You may decide not to use a field for future entries
 * or just add a new one and you don't have to change your previous
 * entries inside the file.
 *
 * This is also a good way around if you just don't know the
 * structure of future entries yet.
 *
 * @access      public
 * @package     yana
 * @subpackage  file_system
 *
 */
class DatFile extends File
{

    /**
     * retrieve a line of data from the file
     *
     * This returns the dataset in line $lineNr
     * as an associative array, or bool(false) on error.
     *
     * Note that the keys are returned in capital letters.
     *
     * @uses    $datfile->get(0)
     * @access  public
     * @param   int  $lineNr  line to retrieve
     * @return  array
     */
    function get($lineNr)
    {
        if (!is_int($lineNr)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'Integer', gettype($lineNr)), E_USER_WARNING);
            return false;
        }

        $data = $this->content[$lineNr];
        $REF = array();
        while (preg_match("/<(.*)>(.*)<\/.*>/Ui", $data, $matches))
        {
            $matches[2] = preg_replace("/(\S{80})(\S*)/i", "\\1<wbr>\\2", $matches[2]);
            $REF[strtoupper($matches[1])] = $matches[2];
            $data = preg_replace("/<".$matches[1].">.*<\/.*>/U", "", $data);
            $matches = array();
        }
        assert('is_array($REF); /* unexcpected result */');
        return $REF;
    }

    /**
     * insert (append) an entry to the file
     *
     * This appends the new entry to the file.
     * Depending on the setting for argument $append,
     * the entry is appended to the end, or inserted
     * on top of the file.
     *
     * The input array must not be multi-dimensional.
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * @uses    $datfile->insert($new_entry)
     * @access  public
     * @param   array  $newEntry  associative array containing the new entry
     * @param   bool   $append    true = append entry on end of file, false = insert entry on top of file
     * @return  bool
     */
    function insert($newEntry, $append = false)
    {
        if (!is_array($newEntry)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'Array', gettype($newEntry)), E_USER_WARNING);
            return false;
        }
        if (!is_bool($append)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 2, 'Boolean', gettype($append)), E_USER_WARNING);
            return false;
        }

        $txt = $this->_encodeEntry($newEntry);
        if (!$this->isEmpty()) {
            if ($append) {
                $this->content[] = $txt;
            } else {
                array_unshift($this->content, $txt);
            }
        } else {
            $this->content[0] = $txt;
        }
        return true;
    }

    /**
     * update an entry of the file
     *
     * Replaces an existing entry. If the entry does not exist,
     * the function returns bool(false).
     * The input array must not be multi-dimensional.
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * @uses    $datfile->insert($new_entry)
     * @access  public
     * @param   int    $lineNr    line number to update
     * @param   array  $newEntry  associative array containing the new entry
     * @return  bool
     * @since   2.9.6
     */
    function update($lineNr, $newEntry)
    {
        if (!is_int($lineNr)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'Integer', gettype($lineNr)), E_USER_WARNING);
            return false;
        }
        if (!is_array($newEntry)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 2, 'Array', gettype($newEntry)), E_USER_WARNING);
            return false;
        }

        $txt = $this->_encodeEntry($newEntry);
        if (!isset($this->content[$lineNr])) {
            return false;
        } else {
            $this->content[$lineNr] = $txt;
            return true;
        }
    }

    /**
     * remove an entry from the file
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * @uses    $datfile->insert($new_entry)
     * @access  public
     * @param   array  $lineNr  line to retrieve
     * @return  bool
     */
    function remove($lineNr)
    {
        if (!is_int($lineNr)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'Integer', gettype($lineNr)), E_USER_WARNING);
            return false;
        }

        if (isset($this->content[$lineNr])) {
            unset($this->content[$lineNr]);
            return true;
        } else {
            return false;
        }
    }

    /**
     * return file content as string (if available)
     *
     * @uses    $datfile->toString()
     * @access  public
     * @return  string
     */
    function toString()
    {
        if (!$this->exists()) {
            $result = "Dat-file ".$this->getPath()." does not exist";
        } elseif ($this->isEmpty()) {
            $result = "Dat-file ".$this->getPath()." is wether empty or not opened";
        } else {
            $result = implode("\n", $this->content);
        }
        assert('is_array($result); /* unexcpected result */');
        return $result;
    }

    /**
     * return the number of rows in the file
     *
     * @uses    $datfile->length()
     * @access  public
     * @return  int
     */
    function length()
    {
        $result = count($this->content);
        assert('is_int($result); /* unexcpected result */');
        return $result;
    }

    /**
     * encode an entry
     *
     * @access  private
     * @param   array   $entry
     * @return  string
     */
    function _encodeEntry($entry)
    {
        $txt = "";
        foreach ($entry as $key => $element)
        {
            $key = strtoupper($key);
            $element = strip_tags($element);
            $txt .= "<$key>$element</$key>";
        }
        $txt = nl2br($txt);
        $txt = preg_replace("/\s/", " ", $txt);
        $txt .= "\n";
        
        return $txt;
    }

    /**
     * self test
     *
     * This function exists only for debugging purposes.
     * Do not uncomment in productive environments.
     *
     * @access  public
     * @static
     * @param   int  $nr  number of test to take
     * @return  bool
     * @ignore
     */
    function test($nr = 1)
    {
        switch ($nr)
        {
            case 1:
                $dat = new DatFile('test.dat');

                $test1 = array(1 => 'a', 'test' => true);
                $test2 = array(2 => 'b', 'C' => 'd', 'e' => 'f');
                $test3 = array(2 => 'g', 'H' => 'i');
                if (!$dat->insert($test1)) {
                    return '"insert (1)" test failed';
                }
                if (!$dat->insert($test2, true)) {
                    return '"insert (2)" test failed';
                }
                $test1 = array_change_key_case($test1, CASE_UPPER);
                $test2 = array_change_key_case($test2, CASE_UPPER);
                if ($dat->get(0) != $test1) {
                    return '"get (1)" test failed';
                }
                if ($dat->get(1) != $test2) {
                    return '"get (2)" test failed';
                }
                if (!$dat->update(1, $test3)) {
                    return '"update" test failed';
                }
                if ($dat->get(1) != $test3) {
                    return '"get (3)" test failed';
                }

                return true;
            break;
        }
    } /* end of function */

}
// Copyright  2009  Thomas Meyer  http://www.yanaframework.net
?>