<?php
/**
 * YANA library
 *
 *
 * Software:        Yana PHP-Framework 3.0.0
 * Author:          Dipl. Inf. Thomas Meyer
 * Homepage:        http://www.yanaframework.net
 * Copyright:       (c) 2009  Thomas Meyer
 * License:         GNU GPL  http://www.gnu.org/licenses/
 *
 * This program: can be redistributed and/or modified under the
 * terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see http://www.gnu.org/licenses/.
 *
 * You may use this software for free. In case you like it, or want to
 * use it for a commercial purpose, please consider making a donation.
 * http://yanaframework.net/donate.php
 *
 * This notice MAY NOT be removed.
 *
 *
 * @package  yana
 * @license  http://www.gnu.org/licenses/gpl.txt
 */

/**
 * read binary large objects (blobs) from database
 *
 * Example of usage:
 * <code>
 * $db = $YANA->connect('foo');
 * $id = $db->get('foo.1.foo_file');
 *
 * $file = new DbBlob($id);
 * $file->read();
 *
 * // output file to screen
 * print $file->get();
 * // copy file to some destination
 * $file->copy('foo/bar.dat');
 * </code>
 *
 * @access      public
 * @package     yana
 * @subpackage  database
 * @since       2.9.2
 */
class DbBlob extends FileReadonly
{
    /**#@+
     * @access  private
     */

    /** @var int    */ var $size = 0;
    /** @var string */ var $type = 'application/unknown';

    /**#@-*/

    /**
     * read file contents
     *
     * Returns bool(true) on success and
     * bool(false) on error.
     *
     * @access  public
     * @return  bool
     */
    function read()
    {
        $source = $this->getPath();

        if (!is_file($source)) {

            trigger_error("The file '${source}' does not exist (database: blob not found).", E_USER_NOTICE);
            return false;

        } else {
            $this->content = array();

            if (!preg_match('/\.gz$/', $source)) {

                trigger_error("The source '${source}' is not a valid database blob.", E_USER_WARNING);
                return false;

            } else {

                $i = 0;
                $gz = gzopen($source, 'r');
                while (!gzeof($gz))
                {
                    $buffer = gzgets($gz, 4096);
                    switch ($i)
                    {
                        case 0:
                            $buffer = trim($buffer);
                            if (preg_match('/^[\w\.\d\-\_]+$/s', $buffer)) {
                                $this->path = $buffer;
                            } else {
                                trigger_error("Invalid file path: '${buffer}'.", E_USER_NOTICE);
                            }
                        break;

                        case 1:
                            $buffer = trim($buffer);
                            if (is_numeric($buffer)) {
                                $this->size = $buffer;
                            } else {
                                trigger_error("Invalid filesize: '${buffer}'.", E_USER_NOTICE);
                            }
                        break;

                        case 2:
                            $buffer = trim($buffer);
                            if (preg_match('/^\w+\/[\w-]+$/s', $buffer)) {
                                $this->type = $buffer;
                            } else {
                                trigger_error("Invalid MIME-Type: '${buffer}'.", E_USER_NOTICE);
                            }
                        break;

                        default:
                            $this->content[] = $buffer;
                        break;

                    }
                    $i++;
                }
                gzclose($gz);

            }
            $this->md5 = md5(implode('', $this->content));
            return true;
        }
    }

    /**
     * get mime-type of this file
     *
     * Returns bool(false) on error.
     *
     * @access  public
     * @return  string|bool(false)
     */
    function getMimeType()
    {
        if (isset($this->type)) {
            return $this->type;
        } else {
            return false;
        }
    }

    /**
     * get size of this file
     *
     * Returns the size of the file in bytes (from cached value).
     * If an error occurs, bool(false) is returned.
     *
     * @access  public
     * @return  int
     */
    function getFilesize()
    {
        if (isset($this->size)) {
            return $this->size;
        } else {
            return false;
        }
    }

    /**
     * read the current file id from the session vars
     *
     * Returns the database id of a selected file.
     * This function is supposed to be used in combination
     * with the form-generator.
     *
     * Example of usage:
     * <code>
     * $dbBlob = new DbBlob(DbBlob::getFileId());
     * </code>
     *
     * Returns bool(false) on error.
     *
     * @access  public
     * @param   int   $i         index number in files list, of the file to get
     * @param   bool  $fullsize  show full size or thumb-nail (images only)
     * @static
     * @return  string|bool(false)
     * @since   2.9.3
     */
    function getFileId($i = null, $fullsize = null)
    {
        global $YANA;

        /* get default arguments */
        if (is_null($i)) {
            if (isset($YANA)) {
                $i = (int) $YANA->getRequestVar('target');
            } elseif (isset($_GET['target'])) {
                $i = (int) $_GET['target'];
            } else {
                return false;
            }
        }

        if (is_null($fullsize)) {
            if (isset($YANA)) {
                $fullsize = (bool) $YANA->getRequestVar('fullsize');
            } else {
                $fullsize = !empty($_GET['fullsize']);
            }
        }
        
        /* check arguments */
        if (!is_int($i)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'Positive integer', gettype($i)), E_USER_WARNING);
            return false;
        } elseif (!isset($_SESSION['filelist'][$i])) {
            trigger_error("Invalid argument. File '$i' is undefined.", E_USER_WARNING);
            return false;
        }
        if (!is_bool($fullsize)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 2, 'Boolean', gettype($fullsize)), E_USER_WARNING);
            return false;
        }

        $source = new String($_SESSION['filelist'][$i]);
        $source->encrypt('xor', YANA_INSTANCE_ID . session_id());
        if ($source->match('/^(?:thumb|image|file)\.\w+\.(?:png|jpg|jpeg|gif|gz)$/s') === false) {
            return false;
        }
        if (!empty($fullsize)) {
            $source->replace('thumb.', 'image.');
        } else {
            $source->replace('image.', 'thumb.');
        }
        if (isset($YANA)) {
            $source = $YANA->getVar('DBBLOB') . $source->get();
        } else {
            $source = 'config/db/.blob/' . $source->get();
        }
        if (!is_file($source)) {
            /* error - file does not exist */
            return false;
        } else {
            return $source;
        }
    }

    /**
     * copy the file to some destination
     *
     * @access   public
     * @param    string  $destFile   destination to copy the file to
     * @param    bool    $overwrite  setting this to false will prevent existing files from getting overwritten
     * @return   bool
     */
    function copy($destFile, $overwrite = true)
    {
        if (!is_string($destFile)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($destFile)), E_USER_WARNING);
            return false;
        }
        if (!is_bool($overwrite)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 2, 'Boolean', gettype($overwrite)), E_USER_WARNING);
            return false;
        }

        /* validity checking */
        if (empty($destFile) || strlen($destFile) > 512 || !preg_match('/^[\w\d-_\.][\w\d-_\/\.]*$/',$destFile)) {
            trigger_error("Invalid filename '".$destFile."' in ".__CLASS__."->".__FUNCTION__."().", E_USER_WARNING);
            return false;
        }

        if (!$overwrite && file_exists($destFile)) {
            trigger_error("Unable to copy to file '${destFile}'. Another file with the same name does already exist.", E_USER_NOTICE);
            return false;
        } elseif ($overwrite && file_exists($destFile) && !is_writeable($destFile)) {
            trigger_error("Unable to copy to file '${destFile}'. Permission denied.", E_USER_NOTICE);
            return false;
        } else {
            $handle = fopen($destFile, "w+");
            if ($handle === false) {
                return false;
            }
            flock($handle, LOCK_EX);
            if (fwrite($handle, $this->getFileContent()) == false) {
                return false;
            }
            flock($handle, LOCK_UN);
            if (fclose($handle) == false) {
                return false;
            }
            chmod($destFile, 0777);
            return true;
        }
    }
}
// Copyright  2009  Thomas Meyer  http://www.yanaframework.net
?>