<?php
/**
 * YANA library
 *
 *
 * Software:        Yana PHP-Framework 3.0.0
 * Author:          Dipl. Inf. Thomas Meyer
 * Homepage:        http://www.yanaframework.net
 * Copyright:       (c) 2009  Thomas Meyer
 * License:         GNU GPL  http://www.gnu.org/licenses/
 *
 * This program: can be redistributed and/or modified under the
 * terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see http://www.gnu.org/licenses/.
 *
 * You may use this software for free. In case you like it, or want to
 * use it for a commercial purpose, please consider making a donation.
 * http://yanaframework.net/donate.php
 *
 * This notice MAY NOT be removed.
 *
 *
 * @package  yana
 * @license  http://www.gnu.org/licenses/gpl.txt
 */

/**
 * decorator  database Creator
 *
 * This decorator class is intended to create SQL DDL (data definition language)
 * from YANA Framework - database structure files.
 *
 * For this task it provides functions which create specific
 * DDL for various DBS.
 *
 * @access      public
 * @package     yana
 * @subpackage  database
 * @since       2.8.6
 */
class DbCreator extends Object
{

    /**#@+
     * @ignore
     * @access  private
     */

    /** @var DbStructure */ var $structure = null;

    /**#@-*/

    /**
     * create a new instance
     *
     * This class requires an object of class DbStructure
     * as input.
     *
     * Example of usage:
     * <code>
     * // create new instance of this class
     * $dbc = new DbCreator( new DbStructure('guestbook'));
     * // since version 2.9.6 you may also write
     * $db = 'guestbook';
     * $dbc = new DbCreator($db);
     * // create SQL DDL (here: using MySQL syntax)
     * $arrayOfStmts = $dbc->createMySQL();
     * // This will output the result as text
     * print implode("\n", $arrayOfStmts);
     * // This will send the statements to the database
     * $db = new DbStream('guestbook');
     * foreach ($arrayOfStmts as $stmt)
     * {
     *    $db->query($stmt);
     * }
     * </code>
     *
     * @param  string|DbStructure  &$structure  a structure resource
     */
    function DbCreator(&$structure)
    {
        $this->setStructure($structure);
    }

    /**
     * get list of tables that require InnoDB driver
     *
     * This applies to MySQL only.
     *
     * @access  private
     * @return  array
     */
    function getListOfInnoDb()
    {
        $listOfInnoDb = array();

        /**
         *  Loop through tables
         */
        foreach ($this->structure->getTables() as $table)
        {
            $table = strtolower($table);
            $listOfForeignKeys = $this->structure->getForeignKeys($table);
            if (!empty($listOfForeignKeys)) {
                $listOfInnoDb[] = $table;
                /**
                 *  Loop through foreign keys
                 */
                foreach ($listOfForeignKeys as $foreignTable)
                {
                    $listOfInnoDb[] = strtolower($foreignTable);
                }
            }
        }
        $listOfInnoDb = array_unique($listOfInnoDb);
        return $listOfInnoDb;
    }

    /**
     * create SQL for MySQL
     *
     * Returns a numeric array of SQL statements.
     * Each element is a single statement.
     * If you want to send the result to a SQL file
     * you should "implode()" the array to a string.
     *
     * @access  public
     * @return  array
     */
    function createMySQL()
    {
        /* this is the result var that will be returned when finished */
        $SQL = array();

        /* this is for statements, which have to come last */
        $lastSQL = array();

        $listOfInnoDb = $this->getListOfInnoDb();

        /*
         *  Create Table
         */
        assert('!isset($table); /* cannot redeclare variable $table */');
        foreach ($this->structure->getTables() as $table)
        {
            assert('is_string($table) && !empty($table);');
            $stmt = "CREATE TABLE IF NOT EXISTS `".YANA_DATABASE_PREFIX."${table}` (\n";

            $listOfColumns = $this->structure->getColumns($table);
            $listOfUniqueConstraints = $this->structure->getUniqueConstraints($table);
            $listOfForeignKeys = $this->structure->getForeignKeys($table);
            $listOfIndexes = $this->structure->getIndexes($table);

            /*
             *  Create Column
             */
            assert('!isset($column); /* cannot redeclare variable $column */');
            foreach ($listOfColumns as $column)
            {
                $type = $this->structure->getType($table, $column);
                $length = $this->structure->getLength($table, $column);
                $precision = $this->structure->getPrecision($table, $column);
                $default = $this->structure->getDefault($table, $column);
                $comment = $this->structure->getDescription($table, $column);
                $stmt .= "\t`${column}` ";

                /*
                 *  Add Type
                 */
                if ($this->structure->isForeignKey($table, $column)) {
                    $foreignTable = $this->structure->getTableByForeignKey($table, $column);
                    $foreignPrimaryKey = $this->structure->getPrimaryKey($foreignTable);
                    $type = $this->structure->getType($foreignTable, $foreignPrimaryKey);
                    $length = $this->structure->getLength($foreignTable, $foreignPrimaryKey);
                    $precision = $this->structure->getPrecision($foreignTable, $foreignPrimaryKey);

                }
                
                assert('!isset($flags); // Cannot redeclare var $flags');
                $flags = "";
                switch (strtolower($type))
                {
                    case "time":
                        $type   = "INT";
                        $length = 11;
                    break;

                    case "reference":
                        $type   = "VARCHAR";
                        if (empty($length)) {
                            $length = 255;
                        }
                    break;

                    case "image": case "file":
                        $type   = "VARCHAR";
                        $length = 128;
                    break;

                    case "ip":
                        $type   = "VARCHAR";
                        $length = 15;
                    break;

                    case "profile":
                        $type   = "VARCHAR";
                        $length = 128;
                    break;

                    case "mail": case "url": case "uri": case "string":
                        if (empty($length)) {
                            $type = 'TEXT';
                        } else {
                            $type = 'VARCHAR';
                        }
                    break;

                    case "select": case "array": case "text":
                        $type  = 'TEXT';
                        $length = 0;
                    break;

                    case "int": case "integer":
                        $type = "INT";
                        if ($this->structure->isUnsigned($table, $column)) {
                            $flags .= " UNSIGNED";
                        }
                        if ($this->structure->isZerofill($table, $column)) {
                            $flags .= " ZEROFILL";
                        }
                    break;

                    case "float": case "double":
                        $type = "DOUBLE";
                        if ($length > 0 || $precision > 0) {
                            $type .= "($length,$precision)";
                            $length = 0;
                        }
                        if ($this->structure->isUnsigned($table, $column)) {
                            $flags .= " UNSIGNED";
                        }
                        if ($this->structure->isZerofill($table, $column)) {
                            $flags .= " ZEROFILL";
                        }
                    break;

                    case "boolean": case "bool":
                        $type = "BOOL";
                    break;

                    default:
                        if ($length == 0) {
                            $type = 'TEXT';
                        } else {
                            $type = 'VARCHAR';
                        }
                    break;

                } /* end switch */
                $stmt .= "$type";

                /*
                 *  Add Length
                 */
                if (is_int($length) && $length > 0 && $type !== 'text') {
                    $stmt .= "(${length})";
                }

                /*
                 *  Add flags
                 */
                if ($flags !== '') {
                    $stmt .= $flags;
                } else {
                    $stmt .= $flags;
                }
                unset($flags);

                /*
                 *  Decide wether this column is nullable
                 */
                if ($this->structure->isNullable($table,$column) === true) {
                    $stmt .= " NULL";
                } else {
                    $stmt .= " NOT NULL";
                }

                /*
                 *  Add Default-Value
                 */
                if (is_string($default)) {
                    $stmt .= " DEFAULT '${default}'";
                } elseif (is_bool($default)) {
                    $stmt .= " DEFAULT " . ( ($default) ? 'true' : 'false' );
                } elseif (is_scalar($default)) {
                    $stmt .= " DEFAULT ${default}";
                }

                /*
                 *  Add Autonumber
                 */
                if ($this->structure->isAutonumber($table,$column) === true) {
                    $stmt .= " AUTO_INCREMENT";
                }

                /*
                 *  Add Comment
                 */
                if (!empty($comment)) {
                    $stmt .= " COMMENT '".addcslashes($comment,"'")."'";
                }

                $stmt .= ",\n";

            } /* end foreach column */
            unset($column);


            /*
             *  Create Primary Key
             */
            $stmt .= "\tPRIMARY KEY  (`".$this->structure->getPrimaryKey($table)."`)";

            if (!empty($listOfUniqueConstraints)) {
                assert('!isset($i); // Cannot redeclare variable $i');
                assert('!isset($uniqueConstraint); // Cannot redeclare variable $uniqueConstraint');
                foreach ($listOfUniqueConstraints as $i => $uniqueConstraint)
                {
                    $listOfUniqueConstraints[$i] = "`$uniqueConstraint`";
                }
                unset($uniqueConstraint, $i);
                $stmt .= ",\n\tUNIQUE ${table}_uq (".implode(', ',$listOfUniqueConstraints).")";
            }

            $stmt .= "\n";

            /*
             *  Decide for InnoDB or MyISAM
             */
            if (in_array($table, $listOfInnoDb)) {
                $stmt .= ") TYPE=InnoDB;";
            } else {
                $stmt .= ") TYPE=MyISAM;";
            }
            assert('is_array($SQL);');
            $SQL[] = $stmt;
            unset($stmt);

            /*
             *  Create Indexes
             */
            if (!empty($listOfIndexes)) {
                assert('!isset($index); /* cannot redeclare variable $index */');
                assert('!isset($i); /* cannot redeclare variable $i */');
                $i = 0;
                foreach ($listOfIndexes as $index)
                {
                    $indexName = "${table}_${i}_idx";
                    assert('is_array($SQL);');
                    $SQL[] = "ALTER TABLE `".YANA_DATABASE_PREFIX."${table}` ADD INDEX `${indexName}` (`${index}`);";
                    $i++;
                }
                unset($i, $index);
            }

            /*
             *  Create Foreign Key Constraints
             */
            if (!empty($listOfForeignKeys)) {
                foreach ($listOfForeignKeys as $foreignKey => $foreignTable)
                {
                    $foreignKey = strtolower($foreignKey);
                    $foreignPrimaryKey = $this->structure->getPrimaryKey($foreignTable);
                    assert('is_array($lastSQL);');
                    $lastSQL[] = "ALTER TABLE `".YANA_DATABASE_PREFIX."${table}` ADD FOREIGN KEY (`${foreignKey}`) REFERENCES `".YANA_DATABASE_PREFIX."${foreignTable}` (`${foreignPrimaryKey}`);";
                }
            }

        } /* end foreach table */
        unset($table);

        assert('is_array($lastSQL);');
        assert('!isset($stmt); /* cannot redeclare variable $stmt */');
        foreach ($lastSQL as $stmt)
        {
            assert('is_array($SQL);');
            $SQL[] = $stmt;
        }
        unset($stmt);

        return $SQL;
    }

    /**
     * create SQL for PostgreSQL
     *
     * Returns a numeric array of SQL statements.
     * Each element is a single statement.
     * If you want to send the result to a SQL file
     * you should "implode()" the array to a string.
     *
     * @access  public
     * @return  array
     */
    function createPostgreSQL()
    {
        /* this is the result var that will be returned when finished */
        $SQL = array();

        /* this is for statements, which have to come last */
        $lastSQL = array();

        /*
         *  Create Table
         */
        assert('!isset($table); /* cannot redeclare variable $table */');
        foreach ($this->structure->getTables() as $table)
        {
            assert('is_string($table) && !empty($table);');
            $stmt = "CREATE TABLE \"".YANA_DATABASE_PREFIX."${table}\" (\n";

            $listOfColumns      = $this->structure->getColumns($table);
            $listOfIndexes      = $this->structure->getIndexes($table);

            /*
             *  Create Column
             */
            assert('!isset($i); /* cannot redeclare variable $i */');
            for ($i = 0; $i < count($listOfColumns); $i++)
            {
                $column = $listOfColumns[$i];

                $type = $this->structure->getType($table, $column);
                $length = $this->structure->getLength($table, $column);
                $precision = $this->structure->getPrecision($table, $column);
                $default = $this->structure->getDefault($table, $column);
                $comment = $this->structure->getDescription($table, $column);
                $stmt .= "\t\"${column}\" ";

                /*
                 *  Add Type
                 */
                if ($this->structure->isForeignKey($table,$column)) {
                    $foreignTable = $this->structure->getTableByForeignKey($table, $column);
                    $foreignPrimaryKey    = $this->structure->getPrimaryKey($foreignTable);
                    $type = $this->structure->getType($foreignTable, $foreignPrimaryKey);
                    $length = $this->structure->getLength($foreignTable, $foreignPrimaryKey);
                    $precision = $this->structure->getPrecision($foreignTable, $foreignPrimaryKey);
                }
                switch (strtolower($type))
                {
                    case "time":
                        $type   = "integer";
                        $length = 0;
                    break;

                    case "reference":
                        $type   = "varchar";
                        if (empty($length)) {
                            $length = 255;
                        }
                    break;

                    case "image": case "file":
                        $type   = "varchar";
                        $length = 128;
                    break;

                    case "ip":
                        $type   = "varchar";
                        $length = 15;
                    break;

                    case "profile":
                        $type   = "varchar";
                        $length = 128;
                    break;

                    case "mail": case "url": case "uri": case "string":
                        if (empty($length)) {
                            $type = 'text';
                        } else {
                            $type = 'varchar';
                        }
                    break;

                    case "select": case "array": case "text":
                        $type  = 'text';
                        $length = 0;
                    break;

                    case "int": case "integer":
                        $type = "integer";
                        $length = 0;
                    break;

                    case "float": case "double":
                        $type = "double precision";
                        $length = 0;
                    break;

                    case "boolean": case "bool":
                        $type = "boolean";
                    break;

                    default:
                        if ($length == 0) {
                            $type = 'text';
                        } else {
                            $type = 'varchar';
                        }
                    break;

                } /* end switch */
                $stmt .= "$type";

                /*  PostgreSQL:
                 *  Add Length
                 */
                if (is_int($length) && $length > 0 && $type !== 'text') {
                    $stmt .= "(${length})";
                }

                /*  PostgreSQL:
                 *  Decide wether this column is nullable
                 */
                if ($this->structure->isNullable($table,$column) === false) {
                    $stmt .= " NOT NULL";
                }

                /*  PostgreSQL:
                 *  Add Autonumber
                 */
                if ($this->structure->isAutonumber($table,$column) === true) {
                    array_unshift($SQL,"CREATE SEQUENCE ${table}_${column}_sq INCREMENT 1;");
                    $stmt .= " DEFAULT nextval('${table}_${column}_sq')";
                /*  PostgreSQL:
                 *  Add Default-Value
                 */
                } elseif (is_string($default)) {
                    $stmt .= " DEFAULT '${default}'";
                } elseif (is_bool($default)) {
                    $stmt .= " DEFAULT " . ( ($default) ? 'true' : 'false' );
                } elseif (is_scalar($default)) {
                    $stmt .= " DEFAULT ${default}";
                }

                /*  PostgreSQL:
                 *  Add Unique Constraint
                 */
                if ($this->structure->isUnique($table,$column) === true) {
                    $stmt .= " UNIQUE";
                }

                /*  PostgreSQL:
                 *  Add Primary Key
                 */
                if ($this->structure->isPrimaryKey($table,$column) === true) {
                    $stmt .= " PRIMARY KEY";
                }

                /*  PostgreSQL:
                 *  Create Foreign Key Constraints
                 */
                if ($this->structure->isForeignKey($table,$column) === true) {
                    $ftable = $this->structure->getTableByForeignKey($table,$column);
                    assert('is_array($lastSQL);');
                    $lastSQL[] = "ALTER TABLE \"".YANA_DATABASE_PREFIX."${table}\" ADD FOREIGN KEY (\"${column}\") REFERENCES \"".YANA_DATABASE_PREFIX.$this->structure->getTableByForeignKey($table,$column)."\";";
                }

                /*  PostgreSQL:
                 *  Add Comment
                 */
                if (!empty($comment)) {
                    assert('is_array($lastSQL);');
                    $lastSQL[] = "COMMENT ON COLUMN \"".YANA_DATABASE_PREFIX."${table}\".\"${column}\" IS '".addcslashes($comment,"'")."';";
                }

                if ($i<count($listOfColumns) - 1) {
                    $stmt .= ",";
                }
                $stmt .= "\n";

            } /* end foreach column */
            unset($i);

            assert('is_array($SQL);');
            $SQL[] = $stmt.");";
            unset($stmt);

            /*  PostgreSQL:
             *  Create Indexes
             */
            if (!empty($listOfIndexes)) {
                assert('!isset($index); /* cannot redeclare variable $index */');
                foreach ($listOfIndexes as $index)
                {
                    assert('is_array($SQL);');
                    $SQL[] = "CREATE INDEX ${table}_${index}_idx ON \"".YANA_DATABASE_PREFIX."${table}\" (\"${index}\");";
                }
                unset($index);
            }

        } /* end foreach table */
        unset($table);

        assert('is_array($lastSQL);');
        assert('!isset($stmt); /* cannot redeclare variable $stmt */');
        foreach ($lastSQL as $stmt)
        {
            assert('is_array($SQL);');
            $SQL[] = $stmt;
        }
        unset($stmt);

        return $SQL;
    }

    /**
     * create SQL for MS SQL Server
     *
     * Returns a numeric array of SQL statements.
     * Each element is a single statement.
     * If you want to send the result to a SQL file
     * you should "implode()" the array to a string.
     *
     * @access  public
     * @return  array
     */
    function createMSSQL()
    {
        /* this is the result var that will be returned when finished */
        $SQL = array();

        /* this is for statements, which have to come last */
        $lastSQL = array();

        /*
         *  Create Table
         */
        assert('!isset($table); /* cannot redeclare variable $table */');
        foreach ($this->structure->getTables() as $table)
        {
            assert('is_string($table) && !empty($table);');
            $stmt = "CREATE TABLE [dbo].[".YANA_DATABASE_PREFIX."${table}] (\n";

            $listOfColumns = $this->structure->getColumns($table);
            $listOfIndexes = $this->structure->getIndexes($table);

            /*
             *  Create Column
             */
            assert('!isset($i); /* cannot redeclare variable $i */');
            for ($i = 0; $i < count($listOfColumns); $i++)
            {
                $column = $listOfColumns[$i];

                $type = $this->structure->getType($table, $column);
                $length = $this->structure->getLength($table, $column);
                $precision = $this->structure->getPrecision($table, $column);
                $default = $this->structure->getDefault($table, $column);
                $comment = $this->structure->getDescription($table, $column);
                $stmt .= "\t[${column}] ";

                /*  MS-SQL:
                 *  Add Type
                 */
                if ($this->structure->isForeignKey($table,$column)) {
                    $foreignTable = $this->structure->getTableByForeignKey($table, $column);
                    $foreignPrimaryKey = $this->structure->getPrimaryKey($foreignTable);
                    $type = $this->structure->getType($foreignTable, $foreignPrimaryKey);
                    $length = $this->structure->getLength($foreignTable, $foreignPrimaryKey);
                    $precision = $this->structure->getPrecision($foreignTable, $foreignPrimaryKey);
                }
                switch (strtolower($type))
                {
                    case "time":
                        $type   = "integer";
                        $length = 0;
                    break;

                    case "reference":
                        $type   = "varchar";
                        if (empty($length)) {
                            $length = 255;
                        }
                    break;

                    case "image": case "file":
                        $type   = "varchar";
                        $length = 128;
                    break;

                    case "ip":
                        $type   = "varchar";
                        $length = 15;
                    break;

                    case "profile":
                        $type   = "varchar";
                        $length = 128;
                    break;

                    case "mail": case "url": case "uri": case "string":
                        if (empty($length)) {
                            $type = 'text';
                        } else {
                            $type = 'varchar';
                        }
                    break;

                    case "select": case "array": case "text":
                        $type  = 'text';
                        $length = 0;
                    break;

                    case "int": case "integer":
                        $type = "integer";
                        $length = 0;
                    break;

                    case "float": case "double":
                        $type = "float";
                        $length = 0;
                    break;

                    case "boolean": case "bool":
                        $type = "bit";
                    break;

                    default:
                        if ($length == 0) {
                            $type = 'text';
                        } else {
                            $type = 'varchar';
                        }
                    break;

                } /* end switch */
                $stmt .= "$type";

                /*  MS-SQL:
                 *  Add Length
                 */
                if (is_int($length) && $length > 0 && $type !== 'text') {
                    $stmt .= "(${length})";
                }

                /*  MS-SQL:
                 *  Decide wether this column is nullable
                 */
                if ($this->structure->isNullable($table,$column) === false) {
                    $stmt .= " NOT NULL";
                } else {
                    $stmt .= " NULL";
                }

                /*  MS-SQL:
                 *  Add Autonumber
                 */
                if ($this->structure->isAutonumber($table,$column) === true) {
                    $stmt .= " IDENTITY (1,1)";

                /*  MS-SQL:
                 *  Add Default-Value
                 */
                } elseif (is_string($default)) {
                    $stmt .= " DEFAULT '${default}'";
                } elseif (is_bool($default)) {
                    $stmt .= " DEFAULT " . ( ($default) ? '1' : '0' );
                } elseif (is_scalar($default)) {
                    $stmt .= " DEFAULT ${default}";
                }

                /*  MS-SQL:
                 *  Add Unique Constraint
                 */
                if ($this->structure->isUnique($table,$column) === true) {
                    assert('is_array($lastSQL);');
                    $lastSQL[] = "ALTER TABLE [dbo].[".YANA_DATABASE_PREFIX."${table}] ADD CONSTRAINT ${table}_${column}_uq UNIQUE (${column});";
                }

                /*  MS-SQL:
                 *  Add Primary Key
                 */
                if ($this->structure->isPrimaryKey($table,$column) === true) {
                    $stmt .= " CONSTRAINT ${table}_pk PRIMARY KEY";
                }

                /*  MS-SQL:
                 *  Create Foreign Key Constraints
                 */
                if ($this->structure->isForeignKey($table,$column) === true) {
                    $ftable = $this->structure->getTableByForeignKey($table,$column);
                    assert('is_array($lastSQL);');
                    $lastSQL[] = "ALTER TABLE [dbo].[".YANA_DATABASE_PREFIX."${table}] ADD CONSTRAINT ${table}_${ftable}_fk FOREIGN KEY (${column}) REFERENCES [dbo].[".YANA_DATABASE_PREFIX."${ftable}];";
                }

                /*  MS-SQL:
                 *  Add Comment
                 */
                if (!empty($comment)) {
                    /* intentionally left blank */
                }

                if ($i<count($listOfColumns) - 1) {
                    $stmt .= ",";
                }
                $stmt .= "\n";

            } /* end foreach column */
            unset($i);

            assert('is_array($SQL);');
            $SQL[] = $stmt.");";
            unset($stmt);

            /*  MS-SQL:
             *  Create Indexes
             */
            if (!empty($listOfIndexes)) {
                assert('!isset($index); /* cannot redeclare variable $index */');
                foreach ($listOfIndexes as $index)
                {
                    assert('is_array($SQL);');
                    $SQL[] = "CREATE NONCLUSTERED INDEX ${table}_${index}_idx ON [".YANA_DATABASE_PREFIX."${table}] ([${index}]);";
                }
                unset($index);
            }

        } /* end foreach table */
        unset($table);

        assert('is_array($lastSQL);');
        assert('!isset($stmt); /* cannot redeclare variable $stmt */');
        foreach ($lastSQL as $stmt)
        {
            assert('is_array($SQL);');
            $SQL[] = $stmt;
        }
        unset($stmt);

        return $SQL;
    }

    /**
     * create SQL for MS Access
     *
     * Same as DbCreator::createMSSQL()
     *
     * @access  public
     * @return  array
     * @see     DbCreator::createMSSQL()
     */
    function createMSAccess()
    {
        return $this->createMSSQL();
    }

    /**
     * create SQL for IBM DB2
     *
     * Returns a numeric array of SQL statements.
     * Each element is a single statement.
     * If you want to send the result to a SQL file
     * you should "implode()" the array to a string.
     *
     * @access  public
     * @return  array
     */
    function createDB2()
    {
        /* this is the result var that will be returned when finished */
        $SQL = array();

        /* this is for statements, which have to come last */
        $lastSQL = array();

        global $YANA;
        if (isset($YANA)) {
            $file = $YANA->registry->get('reserved_sql_keywords');
            $sql_keywords = file($file->getPath());
            unset($file);
        } else {
            $sql_keywords = array();
        }

        /*
         *  Create Table
         */
        assert('!isset($table); /* cannot redeclare variable $table */');
        foreach ($this->structure->getTables() as $table)
        {
            assert('is_string($table) && !empty($table);');
            $tableName = YANA_DATABASE_PREFIX . $table;
            if (qSearchArray($sql_keywords, $tableName) !== false) {
                $tableName = "\"${tableName}\"";
            }

            $stmt = "CREATE TABLE ${tableName}(\n";

            $listOfColumns = $this->structure->getColumns($table);
            $listOfForeignKeys = $this->structure->getForeignKeys($table);
            $listOfIndexes = $this->structure->getIndexes($table);

            /*
             *  Create Column
             */
            assert('!isset($i); /* cannot redeclare variable $i */');
            for ($i = 0; $i < count($listOfColumns); $i++)
            {
                $column = $listOfColumns[$i];

                if (qSearchArray($sql_keywords, $column) !== false) {
                    $columnName = "\"${column}\"";
                } else {
                    $columnName = $column;
                }

                $type = $this->structure->getType($table, $column);
                $length = $this->structure->getLength($table, $column);
                $precision = $this->structure->getPrecision($table, $column);
                $default = $this->structure->getDefault($table, $column);
                $comment = $this->structure->getDescription($table, $column);
                $stmt .= "\t${columnName} ";

                /*  DB2:
                 *  Add Type
                 */
                if ($this->structure->isForeignKey($table,$column)) {
                    $foreignTable = $this->structure->getTableByForeignKey($table, $column);
                    $foreignPrimaryKey = $this->structure->getPrimaryKey($foreignTable);
                    $type = $this->structure->getType($foreignTable, $foreignPrimaryKey);
                    $length = $this->structure->getLength($foreignTable, $foreignPrimaryKey);
                    $precision = $this->structure->getPrecision($foreignTable, $foreignPrimaryKey);
                }
                switch (strtolower($type))
                {
                    case "time":
                        $type   = "INTEGER";
                        $length = 0;
                    break;

                    case "reference":
                        $type   = "VARCHAR";
                        if (empty($length)) {
                            $length = 255;
                        }
                    break;

                    case "image": case "file":
                        $type   = "VARCHAR";
                        $length = 128;
                    break;

                    case "ip":
                        $type   = "VARCHAR";
                        $length = 15;
                    break;

                    case "profile":
                        $type   = "VARCHAR";
                        $length = 128;
                    break;

                    case "mail": case "url": case "uri": case "string":
                        if (empty($length)) {
                            $type = 'LONG VARCHAR';
                        } else {
                            $type = 'VARCHAR';
                        }
                    break;

                    case "select": case "array": case "text":
                        $type  = 'LONG VARCHAR';
                        $length = 0;
                    break;

                    case "int": case "integer":
                        $type = "INTEGER";
                        $length = 0;
                    break;

                    case "float": case "double":
                        if ($length > 0 || $precision > 0) {
                            $type .= "($length,$precision)";
                            $length = 0;
                        }
                    break;

                    case "boolean": case "bool":
                        $type = "SMALLINT";
                        $length = 1;
                    break;

                    default:
                        if ($length == 0) {
                            $type = 'CLOB';
                        } else {
                            $type = 'VARCHAR';
                        }
                    break;

                } /* end switch */
                $stmt .= "$type";

                /*  DB2:
                 *  Add Length
                 */
                if (is_int($length) && $length > 0 && $type !== 'text') {
                    $stmt .= "(${length})";
                }

                /*  DB2:
                 *  Decide wether this column is nullable
                 */
                if ($this->structure->isNullable($table,$column) === false || $this->structure->isAutonumber($table,$column) === true) {
                    $stmt .= " NOT NULL";
                }

                /*  DB2:
                 *  Add Autonumber
                 */
                if ($this->structure->isAutonumber($table,$column) === true) {
                    $stmt .= " GENERATED ALWAYS AS IDENTITY (START WITH 1, INCREMENT BY 1, NO CACHE)";

                /*  DB2:
                 *  Add Default-Value
                 */
                } elseif (is_string($default)) {
                    $stmt .= " DEFAULT '${default}'";
                } elseif (is_bool($default)) {
                    $stmt .= " DEFAULT " . ( ($default) ? '1' : '0' );
                } elseif (is_scalar($default)) {
                    $stmt .= " DEFAULT ${default}";
                }

                /*  DB2:
                 *  Add Primary Key
                 */
                if ($this->structure->isPrimaryKey($table,$column) === true) {
                    $constraintName = "${table}_pk";
                    if (strlen($constraintName) > 18) {
                        $constraintName = substr($constraintName, strlen($constraintName)-18, strlen($constraintName));
                    }
                    $stmt .= " CONSTRAINT ${constraintName} PRIMARY KEY";
                }

                /*  DB2:
                 *  Add Primary Key
                 */
                if ($this->structure->isUnique($table,$column) === true) {
                    $constraintName = "${table}_${column}_uq";
                    if (strlen($constraintName) > 18) {
                        $constraintName = "${column}_uq";
                    }
                    if (strlen($constraintName) > 18) {
                        $constraintName = substr($constraintName, strlen($constraintName)-18, strlen($constraintName));
                    }
                    $stmt .= " CONSTRAINT ${constraintName} UNIQUE";
                }

                /*  DB2:
                 *  Add Comment
                 */
                if (!empty($comment)) {
                    assert('is_array($lastSQL);');
                    $lastSQL[] = "COMMENT ON ${tableName} (${columnName} IS '".addcslashes($comment,"'")."');";
                }

                if ($i<count($listOfColumns) - 1) {
                    $stmt .= ",";
                }
                $stmt .= "\n";

            } /* end foreach column */
            unset($i);

            assert('is_array($SQL);');
            $SQL[] = $stmt.");";
            unset($stmt);

            /*  DB2:
             *  Create Indexes
             */
            if (!empty($listOfIndexes)) {
                assert('!isset($index); /* cannot redeclare variable $index */');
                foreach ($listOfIndexes as $index)
                {
                    if (qSearchArray($sql_keywords, $index) !== false) {
                        $indexName = "\"${index}\"";
                    } else {
                        $indexName = $index;
                    }
                    assert('is_array($SQL);');
                    $SQL[] = "CREATE INDEX ${table}_${index}_idx ON ${tableName} (${indexName});";
                } /* end foreach index */
                unset($index);
            }

            /*  DB2:
             *  Create Foreign Key Constraints
             */
            if (!empty($listOfForeignKeys)) {
                assert('!isset($i); // Cannot redeclare variable $i');
                $i = 0;
                foreach ($listOfForeignKeys as $foreignKey => $foreignTable)
                {
                    $foreignKey = strtolower($foreignKey);
                    $foreignPrimaryKey = $this->structure->getPrimaryKey($foreignTable);
                    $foreignTableName = YANA_DATABASE_PREFIX . $foreignTable;
                    if (qSearchArray($sql_keywords, $foreignTableName) !== false) {
                        $foreignTableName = "\"${foreignTableName}\"";
                    } else {
                        $foreignTableName = $foreignTableName;
                    }
                    if (qSearchArray($sql_keywords, $foreignKey) !== false) {
                        $foreignKeyName = "\"${foreignKey}\"";
                    } else {
                        $foreignKeyName = $foreignKey;
                    }

                    $constraintName = "${table}_fk_${i}";
                    assert('is_array($lastSQL);');
                    $lastSQL[] = "ALTER TABLE ${tableName} ADD CONSTRAINT ${constraintName} FOREIGN KEY (${foreignKeyName}) REFERENCES ${foreignTableName};";
                    ++$i;
                }
                unset($i);
            }

        } /* end foreach table */
        unset($table);

        assert('is_array($lastSQL);');
        assert('!isset($stmt); /* cannot redeclare variable $stmt */');
        foreach ($lastSQL as $stmt)
        {
            assert('is_array($SQL);');
            $SQL[] = $stmt;
        }
        unset($stmt);

        return $SQL;

    }

    /**
     * create SQL for Oracle
     *
     * Returns a numeric array of SQL statements.
     * Each element is a single statement.
     * If you want to send the result to a SQL file
     * you should "implode()" the array to a string.
     *
     * @access  public
     * @return  array
     */
    function createOracleDB()
    {
        /* this is the result var that will be returned when finished */
        $SQL = array();

        /* this is for statements, which have to come last */
        $lastSQL = array();

        /*
         *  Create Table
         */
        assert('!isset($table); /* cannot redeclare variable $table */');
        foreach ($this->structure->getTables() as $table)
        {
            assert('is_string($table) && !empty($table);');
            $stmt = "CREATE TABLE \"".YANA_DATABASE_PREFIX."${table}\" (\n";

            $listOfColumns      = $this->structure->getColumns($table);
            $listOfIndexes      = $this->structure->getIndexes($table);
            $listOfForeignKeys = $this->structure->getForeignKeys($table);
            $seq_cnt              = 0;

            /*
             *  Create Column
             */
            assert('!isset($i); /* cannot redeclare variable $i */');
            for ($i = 0; $i < count($listOfColumns); $i++)
            {
                $column = $listOfColumns[$i];

                $type = $this->structure->getType($table, $column);
                $length = $this->structure->getLength($table, $column);
                $precision = $this->structure->getPrecision($table, $column);
                $default = $this->structure->getDefault($table, $column);
                $comment = $this->structure->getDescription($table, $column);
                $stmt .= "\t\"${column}\" ";

                /*  Oracle:
                 *  Add Type + Length
                 */
                switch (strtolower($type))
                {
                    case "time":
                        $type   = 'INTEGER';
                    break;

                    case "reference":
                        if (empty($length)) {
                            $type   = "VARCHAR2(255 CHAR)";
                        } else {
                            $type   = "VARCHAR2(${length} CHAR)";
                        }
                    break;

                    case "image": case "file":
                        $type   = "VARCHAR2(128 CHAR)";
                    break;

                    case "ip":
                        $type   = 'VARCHAR2(15 CHAR)';
                    break;

                    case "profile":
                        $type   = "VARCHAR2(128 CHAR)";
                    break;

                    case "mail": case "url": case "uri": case "string":
                        if (empty($length)) {
                            $type = 'CLOB';
                        } else {
                            $type = 'VARCHAR2';
                            if (!empty($length)) {
                                $type .= "(${length} CHAR)";
                            }
                        }
                    break;

                    case "select": case "array": case "text":
                        $type  = 'CLOB';
                    break;

                    case "int": case "integer":
                        $type = 'INTEGER';
                    break;

                    case "float": case "double":
                        $type  = 'NUMBER';
                        if ($length > 0 || $precision > 0) {
                            $type .= "($length,$precision)";
                            $length = 0;
                        }
                    break;

                    case "boolean": case "bool":
                        $type = 'NUMBER(1)';
                    break;

                    default:
                        if ($length == 0) {
                            $type = 'CLOB';
                        } else {
                            $type = 'VARCHAR2';
                            if (!empty($length)) {
                                $type .= "(${length} CHAR)";
                            }
                        }
                    break;

                } /* end switch */
                $stmt .= "$type";

                /*  Oracle:
                 *  Decide wether this column is nullable
                 */
                if ($this->structure->isNullable($table,$column) === false || $this->structure->isAutonumber($table,$column) === true) {
                    $stmt .= " NOT NULL";
                }

                /*  Oracle:
                 *  Add Autonumber
                 */
                if ($this->structure->isAutonumber($table,$column) === true) {
                    $lastSQL[] = "CREATE SEQUENCE \"${table}_sq${seq_cnt}\";";
                    assert('!isset($tmp); /* cannot redeclare variable $tmp */');
                    $tmp  = "CREATE OR REPLACE TRIGGER \"${table}_${column}_inc\"\n";
                    $tmp .= "\tBEFORE INSERT ON \"${table}\"\n";
                    $tmp .= "\tFOR EACH ROW BEGIN\n";
                    $tmp .= "\tSELECT \"${table}_${column}_sq\".nextval INTO :new.\"${column}\" FROM dual;\n";
                    $tmp .= "\tEND;\n;";
                    $lastSQL[] = $tmp;
                    $seq_cnt++;
                    unset($tmp);

                /*  Oracle:
                 *  Add Default-Value
                 */
                } elseif (is_string($default)) {
                    $stmt .= " DEFAULT '${default}'";
                } elseif (is_bool($default)) {
                    $stmt .= " DEFAULT " . ( ($default) ? '1' : '0' );
                } elseif (is_scalar($default)) {
                    $stmt .= " DEFAULT ${default}";
                }

                /*  Oracle:
                 *  Add Unique Constraint
                 */
                if ($this->structure->isUnique($table,$column) === true) {
                    assert('is_array($lastSQL);');
                    $lastSQL[] = "ALTER TABLE \"".YANA_DATABASE_PREFIX."${table}\" ADD CONSTRAINT \"${table}_${column}_uq\" UNIQUE (\"${column}\");";
                }

                /*  Oracle:
                 *  Add Comment
                 */
                if (!empty($comment)) {
                    assert('is_array($lastSQL);');
                    $lastSQL[] = "COMMENT ON COLUMN \"".YANA_DATABASE_PREFIX."${table}\".\"${column}\" IS '".addcslashes($comment,"'")."';";
                }

                $stmt .= ",\n";

            } /* end foreach column */
            unset($i,$seq_cnt);

            /*  Oracle:
             *  Add Primary Key
             */
            $stmt .= "\tCONSTRAINT \"${table}_pk\" PRIMARY KEY (\"".$this->structure->getPrimaryKey($table)."\")\n";

            assert('is_array($SQL);');
            $SQL[] = $stmt.");";
            unset($stmt);

            /*  Oracle:
             *  Create Indexes
             */
            if (!empty($listOfIndexes)) {
                assert('!isset($index); /* cannot redeclare variable $index */');
                $i = 0;
                foreach ($listOfIndexes as $index)
                {
                    assert('is_array($SQL);');
                    $SQL[] = "CREATE INDEX \"${table}_idx${i}\" ON \"".YANA_DATABASE_PREFIX."${table}\" (\"${index}\");";
                    $i++;
                }
                unset($i, $index);
            }

            /*
             *  Create Foreign Key Constraints
             */
            if (!empty($listOfForeignKeys)) {
                $i = 0;
                foreach ($listOfForeignKeys as $foreignKey => $foreignTable)
                {
                    $foreignKey = strtolower($foreignKey);
                    $foreignPrimaryKey = $this->structure->getPrimaryKey($foreignTable);
                    $constraintName = "${table}_fk${i}";
                    assert('is_array($lastSQL);');
                    $lastSQL[] = "ALTER TABLE \"".YANA_DATABASE_PREFIX."${table}\" ADD CONSTRAINT \"$constraintName\" FOREIGN KEY (\"${foreignKey}\") REFERENCES \"".YANA_DATABASE_PREFIX."${foreignTable}\" (\"${foreignPrimaryKey}\");";
                    $i++;
                }
                unset($i, $foreignKey, $foreignTable);
            }

        } /* end foreach table */
        unset($table);

        assert('is_array($lastSQL);');
        assert('!isset($stmt); /* cannot redeclare variable $stmt */');
        foreach ($lastSQL as $stmt)
        {
            assert('is_array($SQL);');
            $SQL[] = $stmt;
        }
        unset($stmt);

        return $SQL;
    }

    /**
     * get the currently set structure file
     *
     * Returns an instance of {link DbStructure}.
     * You can query this object for details on the structure
     * of the currently selected database.
     *
     * @access  public
     * @return  DbStructure
     * @since   2.9.6
     */
    function getStructure()
    {
        return $this->structure;
    }

    /**
     * set a new structure file
     *
     * The argument $structure may either be a reference to an instance of
     * class DbStructure, or a string, which contains the name of a database.
     * The database must be registered in the framework. If so, this function
     * may find the corresponding structure file automatically.
     *
     * Returns bool(true) on success or bool(false) on error.
     *
     * @access  public
     * @param   string|DbStructure  &$structure
     * @return  bool
     * @since   2.9.6
     */
    function setStructure(&$structure)
    {
        if (is_object($structure)) {
            /* this is a simple work-around to use PHP5-like type-checking in PHP4 */
            assert('is_a($structure, "DbStructure");  // Wrong argument type for argument 1. Instance of DbStructure expected.');
            $this->structure =& $structure;

        } elseif (is_string($structure)) {
            $this->structure = new DbStructure($structure);

        } else {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($structure)), E_USER_WARNING);
            return false;

        }

        if ($this->structure->read() !== false) {
            return true;
        } else {
            return false;
        }
    }

    /**
     * self test
     *
     * This function exists only for debugging purposes.
     * Do not uncomment in productive environments.
     *
     * @access  public
     * @static
     * @param   string  $dbName
     *
     * @ignore
     */
    function test($dbName)
    {
        global $YANA;
        $db = $YANA->connect($dbName);
        $dbc = new DbCreator($db->structure);
        $list = get_class_methods($dbc);
        foreach ($list as $i => $method)
        {
            if (!preg_match('/^create/i',$method)) {
                unset($list[$i]);
            }
        }
        print "<h1>self test</h1>\n";
        print "<ol>\n";
        print '<li><a href="#idx1">Object</a></li>';
        $i = 0;
        foreach ($list as $name)
        {
            print "<li><a href=\"#idx${i}\">${name}</a></li>\n";
            $i++;
        }
        print "</ol>\n";
        print "<h2 id=\"idx1\">Object</h2>\n";
        $result = htmlentities(print_r($dbc,true));
        print '<pre>'.$result.'</pre>';
        $i = 0;
        foreach ($list as $name)
        {
            print "<h2 id=\"idx${i}\">${name}</h2>\n";
            $result = $dbc->$name();
            print '<pre>'.print_r(implode("\n",$result),true).'</pre>';
            $i++;
        }
        return true;
    } /* end of function */
}
// Copyright  2009  Thomas Meyer  http://www.yanaframework.net
?>