<?php
/**
 * YANA library
 *
 *
 * Software:        Yana PHP-Framework 3.0.0
 * Author:          Dipl. Inf. Thomas Meyer
 * Homepage:        http://www.yanaframework.net
 * Copyright:       (c) 2009  Thomas Meyer
 * License:         GNU GPL  http://www.gnu.org/licenses/
 *
 * This program: can be redistributed and/or modified under the
 * terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see http://www.gnu.org/licenses/.
 *
 * You may use this software for free. In case you like it, or want to
 * use it for a commercial purpose, please consider making a donation.
 * http://yanaframework.net/donate.php
 *
 * This notice MAY NOT be removed.
 *
 *
 * @package  yana
 * @license  http://www.gnu.org/licenses/gpl.txt
 */

/**
 * table information
 *
 * exports as (example):
 *
 * Array
 * (
 *     [name] => foo
 *     [comment] => table of foos
 *     [length] => 6
 *     [primarykey] => foo_id
 *     [foreignkeys] => Array
 *         (
 *             [0] => Array
 *                 (
 *                     [column] => bar_id
 *                     [foreigntable] => bar
 *                     [foreigncolumn] => bar_id
 *                 )
 *
 *         )
 *     -- columns --
 * )
 *
 * Note: you should set the column's properties
 * before you add them to the table and set
 * the table properties after adding the columns.
 *
 * If you do so, properties that are inherited
 * to or from some columns, like the primary key,
 * are applied automatically.
 * Otherwise you will have to do it yourself.
 *
 * @access      public
 * @package     yana
 * @subpackage  database
 * @ignore
 */
class DbInfoTable extends Object
{
    /**#@+
     * @access  private
     */

    /** @var string */ var $primaryKey = null;
    /** @var string */ var $comment = null;
    /** @var string */ var $table = null;
    /** @var array  */ var $foreignKeys = array();
    /** @var array  */ var $init = array();
    /** @var array  */ var $columns = array();

    /**#@-*/

    /**
     * Constructor
     *
     * This function creates a new instance of this class.
     *
     * @param   string  $table table name
     * @access  public
     */
    function DbInfoTable($table)
    {
        if (!is_string($table)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($table)), E_USER_ERROR);
        }
        $this->table = (string) $table;
    }

    /**
     * get table name
     *
     * Returns table name on success and bool(false) on error.
     *                                                                                                                                                   
     * @access  public
     * @return  string|bool(false)
     */
    function getName()
    {
        if (!is_string($this->table)) {
            return false;
        } else {
            return $this->table;
        }
    }

    /**
     * get initialization record
     *
     * Returns a numeric list of sql statements, which ought to be run, when the table is created,
     * or bool(false) on error.
     *
     * @access  public
     * @return  string|bool(false)
     * @since   2.9.7
     */
    function getInit()
    {
        if (is_array($this->init)) {
            return $this->init;

        } else {
            return false;
        }
    }

    /**
     * set initialization record
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * Note: this does not check the syntax of the statements.
     *
     * @access  public
     * @param   array  $init  numeric list of sql statements
     * @return  bool
     * @since   2.9.7
     */
    function setInit($init)
    {
        if (!is_array($init)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'Array', gettype($comment)), E_USER_NOTICE);
            return false;

        } else {
            $this->init = array_values($init);
            return true;

        }
    }

    /**
     * get comment
     *
     * Returns the comment as a string,
     * or bool(false) on error.
     *
     * @access  public
     * @return  string|bool(false)
     */
    function getComment()
    {
        if (is_string($this->comment)) {
            return $this->comment;

        } else {
            return false;
        }
    }

    /**
     * set comment
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * @access  public
     * @param   string  $comment
     * @return  bool
     */
    function setComment($comment)
    {
        if (!is_string($comment)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($comment)), E_USER_NOTICE);
            return false;

        } else {
            $this->comment = $comment;
            return true;

        }
    }

    /**
     * get the name of the primary key
     *
     * Returns the name of the primary key as a lower-cased string,
     * or bool(false) on error.
     *
     * @access  public
     * @return  string|bool(false)
     */
    function getPrimaryKey()
    {
        if (is_string($this->primaryKey)) {
            return $this->primaryKey;

        } else {
            return false;
        }
    }

    /**
     * set primary key
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * @access  public
     * @param   string  $column  name of column containing the primary key
     * @return  bool
     */
    function setPrimaryKey($column)
    {
        if (!is_string($column)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($column)), E_USER_WARNING);
            return false;
        }

        $column = strtolower($column);

        if (!isset($this->columns[$column])) {
            trigger_error("No such column '$column'.", E_USER_WARNING);
            return false;

        } else {
            $this->columns[$column]->primaryKey = true;
            $this->primaryKey = $column;
            return true;

        }
    }

    /**
     * get array of foreign keys
     *
     * Returns bool(false) on error.
     *
     * @access  public
     * @return  array|bool(false)
     */
    function getForeignKeys()
    {
        if (is_array($this->foreignKeys)) {
            return $this->foreignKeys;

        } else {
            return false;
        }
    }

    /**
     * set a foreign key constraint
     *
     * Sets a foreign key constraint on $column.
     * If $foreignColumn is not set, it is assumed,
     * the name of the referenced column is the same
     * as $column.
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * @access  public
     * @param   string  $column         name of column in source table
     * @param   string  $foreignTable   name of destination table
     * @param   string  $foreignColumn  name of column in destination table
     * @return  bool
     */
    function setForeignKey($column, $foreignTable, $foreignColumn = "")
    {
        if (!is_string($column)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($column)), E_USER_WARNING);
            return false;
        }
        if (!is_string($foreignTable)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 2, 'String', gettype($foreignTable)), E_USER_WARNING);
            return false;
        }
        if (!is_string($foreignColumn)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 3, 'String', gettype($foreignColumn)), E_USER_WARNING);
            return false;
        }

        /* apply default value */
        $column = strtolower($column);
        if ($foreignColumn === "") {
            $foreignColumn = $column;
        } else {
            $foreignColumn = strtolower($foreignColumn);
        }

        if (!isset($this->columns[$column])) {
            trigger_error("No such column '$column'.", E_USER_ERROR);
            return false;

        } else {
            $this->columns[$column]->foreignKey = true;
            $this->columns[$column]->setReference($foreignTable, $foreignColumn);

        }

        $this->foreignKeys[] = array(
            'column' => $column,
            'foreigntable' => $foreignTable,
            'foreigncolumn' => $foreignColumn
        );
        return true;
    }

    /**
     * add column object
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * @access  public
     * @param   DbInfoColumn  $column info object
     * @return  bool
     */
    function addColumn($column)
    {
        if (!is_object($column)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'Object', gettype($column)), E_USER_WARNING);
            return false;

        } else {
            $this->columns[$column->getName()] =& $column;
            
            $column->table =& $this->table;

            if ($column->primaryKey) {
                return $this->setPrimaryKey($column->getName());

            } elseif ($column->foreignKey && @is_array($column->references)) {
                return $this->setForeignKey($column->getName(), $column->references[0], $column->references[1]);

            } else {
                return true;

            }
        }
    }

    /**
     * export object as associative array
     *
     * @access  public
     * @return  array
     */
    function toArray()
    {
        $array = array(
            'name' => $this->table,
            'comment' => $this->comment,
            'length' => count($this->columns),
            'primarykey' => $this->primaryKey,
            'init' => $this->init,
            'foreignkeys' => $this->foreignKeys
        );
        foreach ($this->columns as $column)
        {
            $array[] = $column->toArray();
        }
        return $array;
    }
}
// Copyright  2009  Thomas Meyer  http://www.yanaframework.net
?>