<?php
/**
 * YANA library
 *
 *
 * Software:        Yana PHP-Framework 3.0.0
 * Author:          Dipl. Inf. Thomas Meyer
 * Homepage:        http://www.yanaframework.net
 * Copyright:       (c) 2009  Thomas Meyer
 * License:         GNU GPL  http://www.gnu.org/licenses/
 *
 * This program: can be redistributed and/or modified under the
 * terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see http://www.gnu.org/licenses/.
 *
 * You may use this software for free. In case you like it, or want to
 * use it for a commercial purpose, please consider making a donation.
 * http://yanaframework.net/donate.php
 *
 * This notice MAY NOT be removed.
 *
 *
 * @package  yana
 * @license  http://www.gnu.org/licenses/gpl.txt
 */

/**#@+
 * @ignore
 */
if (!defined('DB_FETCHMODE_ASSOC')) {
    define('DB_FETCHMODE_ASSOC', 2);
}
/**#@-*/

/**
 * Database query builder
 *
 * This class is a query builder that can be used to build SQL statements.
 *
 * Note: this class does NOT untaint input data for you.
 * It also does NOT automatically resolve foreign keys.
 * This is mentioned here for security reasons.
 *
 * Note that there are some special features that you may find usefull:
 *
 * First, this class is able to detect and resolve table inheritance.
 * Note that this feature is turned on by default. You may turn it of
 * using the function useInheritance(), if you want. Usually this is not
 * necessary.
 *
 * Second, this class allows you to navigate from one table to another along
 * definded foreign keys and also into columns of type array. This is done by
 * using the function setKey(). These keys form something similar to an
 * address within the database that consists of table.row.column. For example
 * foo.2.bar means select bar from foo where id = 2.
 * If the column is a foreign key, you may add another column inside the
 * referenced table, which itself may also be another foreign key.
 * If the column is an array, you may add an index inside that array, which you
 * want to return. You will find more details on that inside the developer's
 * cookbook in the manual.
 *
 * @access      public
 * @package     yana
 * @subpackage  database
 * @since       2.9 RC1
 */
class DbQuery extends Object
{
    /**#@+
     * @access  protected
     * @ignore
     */

    /** @var string */ var $id             = null;
    /** @var int    */ var $type           = YANA_DB_UNKNOWN;
    /** @var int    */ var $expectedResult = YANA_DB_UNKNOWN;

    /** @var string */ var $table          = null;
    /** @var string */ var $row            = '*';
    /** @var array  */ var $column         = array();

    /** @var array  */ var $profile        = array();
    /** @var array  */ var $rowId          = array();
    /** @var array  */ var $where          = array();
    /** @var array  */ var $orderBy        = null;
    /** @var bool   */ var $desc           = false;
    /** @var int    */ var $limit          = 0;
    /** @var int    */ var $offset         = 0;

    /** @var object */ var $db             = null;
    /** @var array  */ var $joins          = array();

    /** @var array  */ var $values         = null;

    /** @var string */ var $arrayAddress   = null;
    /** @var bool   */ var $useInheritance = true;
    /** @var array  */ var $parentTables   = array();
    /** @var array  */ var $parentByColumn = array();
    /** @var array  */ var $queue          = array();

    /**#@-*/

    /**
     * create a new instance
     *
     * This creates and initializes a new instance of this class.
     *
     * The argument $database can be an instance of class DbStream or
     * any derived sub-class (e.g. FileDb).
     *
     * @param  object  $database
     */
    function DbQuery($database)
    {
        /* error - data type */
        if (is_string($database)) {

            global $YANA;
            if (isset($YANA)) {
                $database = $YANA->connect($database);
                if (is_object($database)) {
                    $this->db =& $database;
                }
            }

        } elseif (!is_object($database)) {

            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'Instance of DbStream', gettype($database)), E_USER_WARNING);

        /* error - invalid base class */
        } elseif (!is_a($database, 'DbStream') && !is_subclass_of($database, 'DbStream')) {

            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'Instance of DbStream', get_class($database)), E_USER_WARNING);

        } else {

            $this->db =& $database;

        }
    }

    /**
     * reset query
     *
     * Resets all properties of the query object, except
     * for the database connection and the properties
     * "table", "type", "useInheritance".
     *
     * This function allows you to "recycle" a query object
     * and reuse it without creating another one. This can
     * help to improve the performance of your application.
     *
     * @access  public
     * @since   2.9.4
     */
    function resetQuery()
    {
        $this->id             = null;
        $this->row            = '*';
        $this->column         = array();
        $this->profile        = array();
        $this->rowId          = array();
        $this->where          = array();
        $this->orderBy        = null;
        $this->desc           = false;
        $this->limit          = 0;
        $this->offset         = 0;
        $this->joins          = array();
        $this->values         = null;
        $this->arrayAddress   = null;
        $this->parentTables   = array();
        $this->parentByColumn = array();
        $this->queue          = array();
    }

    /**
     * select the kind of statement
     *
     * The argument type can be one of the following:
     * <ul>
     *  <li> YANA_DB_UNKNOWN = to reset type </li>
     *  <li> YANA_DB_SELECT = Select column from table ... </li>
     *  <li> YANA_DB_UPDATE = Update table ... </li>
     *  <li> YANA_DB_INSERT = Insert into table ... </li>
     *  <li> YANA_DB_DELETE = Delete from table where ... </li>
     *  <li> YANA_DB_EXISTS = Select 1 from ... where ... </li>
     *  <li> YANA_DB_LENGTH = Select count(*) from ...  </li>
     * </ul>
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * @access  public
     * @param   int  $type
     * @return  bool
     */
    function setType($type)
    {
        $this->id = null;

        switch ($type)
        {
            case YANA_DB_INSERT:
                if ($this->row === '*' && $this->expectedResult === YANA_DB_TABLE) {
                    if (@isset($this->db->structure) && $this->db->structure->isAuto($this->table, $this->db->structure->getPrimaryKey($this->table))) {
                        $this->expectedResult = YANA_DB_ROW;
                    }
                }
                $this->type = $type;
                return true;
            break;

            case YANA_DB_LENGTH:
                if (is_array($this->column) && count($this->column) > 1) {
                    trigger_error("Cannot use query type 'length' with multiple columns.", E_USER_WARNING);
                    return false;
                }
            case YANA_DB_UNKNOWN:
            case YANA_DB_SELECT:
            case YANA_DB_UPDATE:
            case YANA_DB_DELETE:
            case YANA_DB_EXISTS:
                $this->type = $type;
                return true;
            break;

            default:
                trigger_error("Invalid argument 1 in " . __CLASS__ . "->" . __FUNCTION__ . "(). " .
                "The selected statement type is unknown.", E_USER_WARNING);
                return false;
            break;
        }
    }

    /**
     * get the currently selected type of statement
     *
     * Returns currently selected constant.
     *
     * @access  public
     * @return  int
     */
    function getType()
    {
        if (is_int($this->type)) {
            return $this->type;
        } else {
            /* error: no type has been selected, yet */
            return YANA_DB_UNKNOWN;
        }
    }

    /**
     * find out which kind of result is expected
     *
     * Returns currently selected constant.
     *
     * <ul>
     *  <li> YANA_DB_UNKNOWN - no input </li>
     *  <li> YANA_DB_TABLE   - table only </li>
     *  <li> YANA_DB_ROW     - table + row </li>
     *  <li> YANA_DB_COLUMN  - table + column </li>
     *  <li> YANA_DB_CELL    - table + row + column </li>
     * </ul>
     *
     * Note: YANA_DB_CELL means to refer to exactly 1 column.
     * When retrieving multiple columns from a row,
     * use YANA_DB_ROW instead.
     *
     * @access  public
     * @return  int
     * @since   2.9.3
     * @ignore
     */
    function getExpectedResult()
    {
        if (is_int($this->expectedResult)) {
            return $this->expectedResult;
        } else {
            /* error: no type has been selected, yet */
            return YANA_DB_UNKNOWN;
        }
    }

    /**
     * activate / deactivate automatic handling of inheritance
     *
     * The query builder is able to detect if one table inherits
     * from another and if so, it will auto-join both tables.
     * In this case, selecting a row from the offspring table will
     * also return all entries of the corresponding row in the
     * parent table.
     *
     * However: while this usually comes in handy, there are some
     * rare situations where you won't want this to be done.
     * E.g. when copying rows from one table to another.
     *
     * This function allows you to enable or disable this feature.
     * It is enabled by default.
     *
     * Note: you have to set this before you set the table property.
     * Otherwise it will have no effect.
     *
     * The function will return the previous state.
     * If no new state is provided, it will just return the current
     * setting without changing anything.
     *
     * @access  public
     * @param   bool  $state  true = on, false = off
     * @return  bool
     * @since   2.9.3
     */
    function useInheritance($state = null)
    {
        $useInheritance = $this->useInheritance;

        if (!is_null($state)) {
            if ($state) {
                $this->useInheritance = true;
            } else {
                $this->useInheritance = false;
            }
        }

        return $useInheritance;
    }

    /**
     * set parent table
     *
     * The query builder will set this automatically, to indicate,
     * that one table inherits from another.
     *
     * @access  private
     * @param   string  $table
     * @param   string  $parentTable
     * @return  bool
     * @since   2.9.6
     */
    function setParent($table, $parentTable)
    {
        assert('is_string($table); // Wrong argument type argument 1. String expected');
        assert('is_string($parentTable); // Wrong argument type argument 2. String expected');

        /**
         * add columns
         */
        if (@isset($this->db->structure)) {
            $columns = $this->db->structure->getColumns($parentTable);
            if (empty($columns)) {
                return false;
            } else {
                assert('!isset($column); // Cannot redeclare var $column');
                foreach ($columns as $column)
                {
                    $column = strtoupper($column);
                    if (!isset($this->parentByColumn[$column]) && !$this->db->structure->isColumn($this->table, $column)) {
                        $this->parentByColumn[$column] = $parentTable;
                    }
                }
                unset($column);
            }
        }
        /**
         * add table
         */
        $table = strtoupper($table);
        $this->parentTables[$table] = $parentTable;
        return true;
    }

    /**
     * get parent table by column name
     *
     * This function provides information on entity inheritance
     * within the database's data structure.
     *
     * If the table extends another table, and the column
     * is inherited one of the parent tables, then this function
     * will return the name of the parent table, where the
     * column was defined or re-defined.
     *
     * It will return bool(false) if there is no such parent.
     *
     * @access  public
     * @param   string  $column  name of a column
     * @since   2.9.6
     */
    function getParentByColumn($column)
    {
        assert('is_string($column); // Wrong argument type argument 1. String expected');

        $column = strtoupper($column);
        if (isset($this->parentByColumn[$column])) {
            return $this->parentByColumn[$column];
        } else {
            return false;
        }
    }

    /**
     * get the parent of a table
     *
     * This function provides information on entity inheritance
     * within the database's data structure.
     *
     * If $table extends another table, then this
     * will return the name of the parent table as a string.
     *
     * It will return bool(false) if there is no such parent.
     *
     * If the argument $table is empty, or not provided, the
     * currently selected table (see {link DbQuery::setTable()})
     * is used instead.
     *
     * @access  public
     * @param   string  $table  name of a table
     * @since   2.9.6
     */
    function getParent($table = "")
    {
        assert('is_string($table); // Wrong argument type argument 1. String expected');

        if ($table === "") {
            $table = $this->table;
        }
        $table = strtoupper($table);
        if (isset($this->parentTables[$table])) {
            return $this->parentTables[$table];
        } else {
            return false;
        }
    }

    /**
     * recursively detect the parent of a table
     *
     * @access  private
     * @param   string  $table
     * @since   2.9.6
     */
    function detectInheritance($table)
    {
        assert('is_string($table); // Wrong argument type argument 1. String expected');
        
        if (@isset($this->db->structure)) {
            $table = strtoupper($table);
    
            $parents = array($table); /* to detect circular refrences */
    
            /**
             * recursively detect parents
             *
             * Inheritance occurs, when the primary key also is a foreign key and
             * the parent table is does not have itself as one of its descendants.
             */
            $primaryKey = strtoupper($this->db->structure->getPrimaryKey($table));
            while ($this->db->structure->isForeignKey($table, $primaryKey))
            {
                $foreignTable = $this->db->structure->getTableByForeignKey($table, $primaryKey);
                $fTableKey = strtoupper($foreignTable);
                if (!in_array($fTableKey, $parents)) {
                    $foreignKey = strtoupper($this->db->structure->getPrimaryKey($foreignTable));
                    $this->setJoin($foreignTable, $primaryKey, $foreignKey);
                    $this->setParent($table, $foreignTable);
                    $table = $fTableKey;
                    $primaryKey = $foreignKey;
                } else {
                    break;
                }
            }
        }
    }

    /**
     * set source table
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * @access  public
     * @param   string  $table
     * @return  bool
     */
    function setTable($table)
    {
        $this->id = null;

        /**
         * 1) wrong argument type
         */
        if (!is_string($table)) {

            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($table)), E_USER_WARNING);
            return false;

        /**
         * 2) invalid argument, not a table
         */
        } elseif (!is_null($this->db->structure) && !$this->db->structure->isTable($table)) {

            trigger_error("The table '$table' is unknown.", E_USER_WARNING);
            return false;

        /**
         * 3) argument is valid
         */
        } else {

            /**
             * 3.1) Auto-attach profile check to where clause
             *      if profile constraint is present
             */
            global $YANA;
            if (isset($this->db->structure) && isset($YANA)) {
                $profile = $this->db->structure->getProfile($table);
                if (empty($profile)) {
                    $this->profile = null;
                } else {
                    $this->profile = array($profile, $YANA->session->getId(), '=');
                }
            }

            /**
             * 3.2) update type of expected result
             */
            if ($this->expectedResult === YANA_DB_UNKNOWN) {
                $this->expectedResult = YANA_DB_TABLE;
            }

            /**
             * 3.3) assign table
             */
            $this->table = strtolower($table);

            /**
             * 3.4) inheritance
             *
             * Details:
             * If one table inherits from another -
             * that is if the primary key is also a foreign key -
             * then these are to be joined automatically.
             */
            if ($this->useInheritance) {
                $this->detectInheritance($table);
            }

            return true;
        }
    }

    /**
     * get the currently selected table
     *
     * Returns the lower-cased name of the currently
     * selected table, or bool(false) if none has been
     * selected yet.
     *
     * @access  public
     * @return  bool(false)|string
     */
    function getTable()
    {
        if (is_string($this->table)) {
            return $this->table;
        } else {
            /* error: no table has been selected, yet */
            return false;
        }
    }

    /**
     * set source column
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * The second argument applies to columns of type 'array' only.
     * In such case you may provide the array key inside the value
     * of the column that you wish to get.
     * If it is a multidimensional array, you may traverse deper
     * dimensions by linking keys with a dot '.' - for example:
     * "foo.bar" gets $result['foo']['bar'].
     *
     * Note: this will not check if the key that you provided is
     * a valid key or if it really points to a value. If it is not,
     * the resultset will be empty.
     *
     * An E_USER_WARNING is issued if the second argument is
     * provided but the targeted column is not of type 'array'.
     *
     * @access  public
     * @param   string  $column
     * @param   string  $arrayAddress
     * @return  bool
     * @name    DbQuery::setColumn()
     * @see     DbQuery::setColumns()
     */
    function setColumn($column = '*', $arrayAddress = null)
    {
        $this->id = null;

        /**
         * 1) wrong argument type
         */
        if (!is_string($column)) {

            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($column)), E_USER_WARNING);
            return false;

        } elseif (!is_null($arrayAddress) && !is_string($arrayAddress)) {

            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 2, 'String', gettype($arrayAddress)), E_USER_WARNING);
            return false;

        /**
         * 2) wrong order of commands, need to set up table first
         */
        } elseif (empty($this->table)) {

            trigger_error("Cannot set column - need to set table first!", E_USER_WARNING);
            return false;

        } else {
            
            /**
             * 3) select all columns
             */
            if ($column === '*' || $column === '') {

                /**
                 * 3.1) set column
                 */
                if (is_null($arrayAddress)) {
                    $this->column = array();
                    if ($this->row === '*') {
                        if ($this->expectedResult !== YANA_DB_ROW) {
                            $this->expectedResult = YANA_DB_TABLE;
                        }
                    } else {
                        $this->expectedResult = YANA_DB_ROW;
                    }
                    return true;
    
                /**
                 * 3.2) error - cannot set array address on a table
                 */
                } else {
                    trigger_error("Argument 2 is invalid in function " . __CLASS__ . "->" . __FUNCTION__ . ".\n\t\t" .
                    "Array address cannot be used in combination with the column selector '*'.", E_USER_WARNING);
                    return false;
                }


            /**
             * 4) select one specific column
             */
            } else {

                /**
                 * 4.1) extract table, where provided
                 */
                assert('!isset($table); // Cannot redeclare var $table');
                if (strpos($column, '.')) {
                    list($table, $column) = explode('.', $column);
                    if (!$this->setTable($table)) {
                        return false;
                    }
                    unset($table);
                } elseif (($table = $this->getParentByColumn($column)) !== false) {
                    if (!$this->setTable($table)) {
                        return false;
                    }
                    unset($table);
                }

    
                /**
                 * 4.2) invalid argument, not a column
                 */
                if (!is_null($this->db->structure) && !$this->db->structure->isColumn($this->table, $column)) {
        
                    trigger_error("The column '$column' is not found in table '".$this->table."'.", E_USER_WARNING);
                    return false;
                }

                /**
                 * 4.3) set array address
                 */
                if (is_string($arrayAddress)) {
                    /**
                     * 4.3.a) error - not a column of type array
                     */
                    if (!is_null($this->db->structure) && $this->db->structure->getType($this->table, $column) !== 'array') {
                        trigger_error("Argument 2 is invalid in function " . __CLASS__ . "->" . __FUNCTION__ . ".\n\t\t" .
                        "Array address can only be used on columns of type array. Found column of type '" .
                        $this->db->structure->getType($this->table, $column) . "' instead.", E_USER_WARNING);
                        return false;

                    /**
                     * 4.3.b) column is valid
                     */
                    } else {
                        $this->arrayAddress = $arrayAddress;
                    }
                }
                /**
                 * 4.4) set column
                 */
                $this->column = array(array($this->table, strtolower($column)));
                if ($this->row === '*') {
                    if ($this->expectedResult === YANA_DB_ROW) {
                        $this->expectedResult = YANA_DB_CELL;
                    } else {
                        $this->expectedResult = YANA_DB_COLUMN;
                    }
                } else {
                    $this->expectedResult = YANA_DB_CELL;
                }
                return true;
    
            }
        }
    }

    /**
     * set source columns
     *
     * This sets the list of columns to retrieve, like in
     * SELECT col1, col2, ... colN FROM ...
     *
     * Note that this applies only to select statements,
     * not insert, update or delete.
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * Note that, depending on the number of columns you wish to
     * retrieve, the datatype of the result may differ.
     *
     * Getting 1 column from 1 row will just return the
     * value of that cell, e.g. int(1). Getting multiple columns
     * from 1 row will return an array containing the values,
     * e.g. array('col1'=>1, 'col2'=>2, 'col3'=>3).
     *
     * Getting 1 column from multiple rows will return an
     * one-dimensional array of these values.
     * Getting multiple columns from multiple rows will
     * return a two-dimensional array of rows, where each
     * row is an associative array containing the values
     * of the selected columns.
     *
     * Examples:
     * <code>
     * // select 1 column
     * $dbq->setColumns(array('foo'));
     * // same as:
     * $dbq->setColumn('foo');
     *
     * // select multiple columns
     * $dbq->setColumns(array('foo1', 'foo2'));
     *
     * // select multiple columns from different tables
     * // 1) join with table2
     * $dbq->setJoin('table2');
     * // 2) select columns from current table and table2
     * $dbq->setColumns(array('foo1', 'table2.foo2'));
     * </code>
     *
     * @access  public
     * @param   array  $columns  list of columns
     * @return  bool
     * @since   2.9.6
     * @name    DbQuery::setColumns()
     * @see     DbQuery::setColumn()
     */
    function setColumns($columns = array())
    {
        $this->id = null;

        /**
         * 1) wrong argument type
         */
        if (!is_array($columns)) {

            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'Array', gettype($columns)), E_USER_WARNING);
            return false;

        /**
         * 1.1) wrong argument type
         */
        } elseif (count($columns) === 1) {
            
            return $this->setColumn($columns[0]);

        /**
         * 1.2) wrong order of commands, need to set up table first
         */
        } elseif ($this->type === YANA_DB_LENGTH && !empty($columns)) {

            trigger_error("Cannot select multiple columns for query type 'length'.", E_USER_WARNING);
            return false;

        /**
         * 2) wrong order of commands, need to set up table first
         */
        } elseif (empty($this->table)) {

            trigger_error("Cannot set columns - need to set table first!", E_USER_WARNING);
            return false;


        /**
         * 3) select all columns
         */
        } elseif (empty($columns)) {

            $this->column = array();
            if ($this->row === '*') {
                if ($this->expectedResult !== YANA_DB_ROW) {
                    $this->expectedResult = YANA_DB_TABLE;
                }
            } else {
                $this->expectedResult = YANA_DB_ROW;
            }
            return true;

        /**
         * 4) error - no database connection available
         */
        } elseif (is_null($this->db->structure)) {

            return false;

        /**
         * 5) select specific columns
         */
        } else {

            assert('!isset($column); // Cannot redeclare var $column');
            assert('!isset($i); // Cannot redeclare var $i');
            assert('!isset($table); // Cannot redeclare var $table');
            foreach ($columns as $i => $column)
            {
                if (!is_string($column)) {
                    trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'Array of Strings', gettype($column) . "[$i]"), E_USER_WARNING);
                    return false;

                } else {
                    if (strpos($column, '.')) {
                        list($table, $column) = explode('.', $column);
                    } elseif (($table = $this->getParentByColumn($column)) === false) {
                        $table = $this->table;
                    }
                    if (!$this->db->structure->isColumn($table, $column)) {
                        trigger_error("The column '$column' is not found in table '".$table."'.", E_USER_WARNING);
                        return false;
                    }

                }
                $columns[$i] = array($table, strtolower($column));

            }
            unset($table, $i, $column);
            $this->column = $columns;
            if ($this->row === '*') {
                $this->expectedResult = YANA_DB_TABLE;
            } else {
                $this->expectedResult = YANA_DB_ROW;
            }
            return true;

        }
    }

    /**
     * get the currently selected column
     *
     * Returns the lower-cased name of the currently
     * selected column.
     *
     * If none has been selected, '*' is returned.
     *
     * Version info: the argument $i became available in 2.9.6.
     * When multiple columns are selected, use this argument to
     * choose the index of the column you want. Where 0 is the
     * the first column, 1 is the second aso.
     * If the argument $i is not provided, the function returns
     * the first column.
     *
     * See {link DbQuery::getColumns()} to get a list of all
     * selected columns.
     *
     * @access  public
     * @param   int     $i  index of column to get
     * @return  string
     * @name    DbQuery::getColumn()
     * @see     DbQuery::getColumns()
     */
    function getColumn($i = 0)
    {
        if (is_array($this->column)) {
            if (isset($this->column[$i])) {
                if (is_array($this->column[$i])) {
                    if (isset($this->column[$i][1]) && is_string($this->column[$i][1])) {
                        return $this->column[$i][1];
                    }

                } elseif (is_string($this->column[$i])) {
                    return $this->column[$i];

                }
            }

        } elseif (is_string($this->column)) {
            return $this->column;

        }

        /* error: invalid property */
        return '*';
    }

    /**
     * get the list of all selected columns
     *
     * Returns the lower-cased names of the currently
     * selected columns as a numeric array of strings.
     *
     * If none has been selected, an empty array is returned.
     *
     * @access  public
     * @return  array
     * @since   2.9.6
     * @name    DbQuery::getColumns()
     * @see     DbQuery::getColumn()
     */
    function getColumns()
    {
        if (is_array($this->column)) {
            return $this->column;

        /**
         * catchable error: column is string (can be converted to array)
         */
        } elseif (is_string($this->column)) {
            return array(array($this->table, $this->column));

        /**
         * uncatchable error: column has unexpected type
         */
        } else {
            return array();
        }
    }

    /**
     * get the currently selected array address
     *
     * Returns the currently address as a string, or
     * bool(false) if none has been selected yet.
     *
     * @access  public
     * @return  bool(false)|string
     * @since   2.9.3
     */
    function getArrayAddress()
    {
        if (is_string($this->arrayAddress)) {
            return $this->arrayAddress;
        } else {
            /* error: no address has been selected, yet */
            return false;
        }
    }

    /**
     * set source row
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * Note: does not check if row exists.
     *
     * Currently you may only request 1 row or all.
     * To search for all rows, use the wildcard '*'.
     *
     * @access  public
     * @param   scalar  $row
     * @return  bool
     */
    function setRow($row)
    {
        assert('is_scalar($row); // Wrong argument type for argument 1. Scalar expected.');
        $this->id = null;

        /**
         * 1) wrong argument type
         */
        if (!is_scalar($row)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'Scalar', gettype($row)), E_USER_WARNING);
            return false;

        /**
         * 2) wrong order of commands, need to set up table first
         */
        } elseif (empty($this->table)) {

            trigger_error("Cannot set row - need to set table first!", E_USER_WARNING);
            return false;

        /**
         * 3) select all rows
         *
         * Note: this intentionally also takes int(0) as an alias for '*'.
         */
        } elseif (empty($row) || $row === '*' || $row === '?') {

            /**
             * 3.1) reset row id
             */
            $this->rowId = null;
            /**
             * 3.2) set row
             */
            $this->row = '*';
            /**
             * 3.3) update type of expected result
             */
            if ($row === '?' || ($this->type === YANA_DB_INSERT && $this->db->structure->isAuto($this->table, $this->db->structure->getPrimaryKey($this->table)))) {
                if (empty($this->column)) {
                    $this->expectedResult = YANA_DB_ROW;
                } else {
                    $this->expectedResult = YANA_DB_CELL;
                }
            } elseif (empty($this->column)) {
                $this->expectedResult = YANA_DB_TABLE;
            } else {
                $this->expectedResult = YANA_DB_COLUMN;
            }
            return true;

        /**
         * 4) select one specific row
         */
        } else {

            $row = (string) $row;
            $row = strtoupper($row);

            /**
             * 4.1) update row id
             */
            if (isset($this->db->structure)) {
                $this->rowId = array(array($this->table, $this->db->structure->getPrimaryKey($this->table)), $row, '=');
            }
            /**
             * 4.2) set row
             */
            $this->row = strtolower($row);
            /**
             * 4.3) update type of expected result
             */
            if (empty($this->column)) {
                $this->expectedResult = YANA_DB_ROW;
            } else {
                $this->expectedResult = YANA_DB_CELL;
            }
            return true;

        }
    }

    /**
     * get the currently selected row
     *
     * Returns the lower-cased name of the currently
     * selected column, or bool(false) if none has been
     * selected yet.
     *
     * If none has been selected, '*' is returned.
     *
     * @access  public
     * @return  string
     */
    function getRow()
    {
        if (is_string($this->row)) {
            return strtolower($this->row);
        } else {
            /* error: no row has been selected, yet */
            return '*';
        }
    }

    /**
     * resolve key address to determine table, column and row
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * @access  public
     * @param   string  $key
     * @return  bool
     */
    function setKey($key)
    {
        /**
         * 1) input is invalid
         */
        if (!is_string($key)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($key)), E_USER_WARNING);
            return false;
        } elseif (!preg_match("/^[\w\d-_]+(\.([\w\d-_]*|\*|\?)){0,}(\.\*)?$/i", $key)) {
            trigger_error("Syntax error. The key '${key}' is not valid.", E_USER_WARNING);
            return false;
        } else {
            $key = preg_replace("/\.(\*)?$/", '', $key);
            $array = explode(".", $key);
        }

        if (empty($array)) {

            return false;

        /**
         * 2) input is valid
         *
         * 2.1) resolve foreign keys to get true adress
         */
        } elseif (count($array) > 3) {
            assert('!isset($isArray); /* cannot redeclare variable $isArray */');
            $isArray = ($this->db->structure->getType($array[0], $array[2]) === 'array');

            assert('!isset($a); /* cannot redeclare variable $a */');
            assert('!isset($foreignTable); /* cannot redeclare variable $foreignTable */');
            while (!$isArray && count($array) > 3 && $this->db->structure->isForeignKey($array[0], $array[2]))
            {
                $a = $this->db->get($array[0] . "." . $array[1] . "." . $array[2]);
                if (empty($a)) {
                    trigger_error('Operation aborted due to invalid foreign key.', E_USER_NOTICE);
                    $this->db->handleError("SQL ERROR: Unable to resolve foreign key '${key}'. " .
                    "This may mean one of the associated tables contains inconsistent data. " .
                    "Check if foreign key association has been broken.");
                    return false;
                } else {
                    $foreignTable = $this->db->structure->getTableByForeignKey($array[0], $array[2]);
                    @array_shift($array);
                    @array_shift($array);
                    @array_shift($array);
                    array_unshift($array, $a);
                    array_unshift($array, $foreignTable);
                    $isArray = ($this->db->structure->getType($array[0], $array[2]) === 'array');
                }
            }
            unset($a, $foreignTable);
            if ($isArray) {
                $this->arrayAddress = implode('.', array_slice($array, 3));
                $array = array_slice($array, 0, 3);
            } else {
                /* intentionally left blank */
            }
            if (!$isArray && count($array) > 3) {
                trigger_error("There is no database object that corresponds to the key '${key}'.", E_USER_WARNING);
                return false;
            }
            unset($isArray);
        } else {
            /* intentionally left blank */
        }

        /**
         * 2.2) set new values
         */

        /**
         * 2.2.1) select table
         */
        if (!$this->setTable($array[0])) {
            return false;
        }

        /**
         * 2.2.2) select row (if any)
         */
        if (empty($array[1])) {
            $this->setRow('*');
        } elseif (!$this->setRow($array[1])) {
            return false;
        } elseif ($array[1] === '?') {
            if (is_object($this->db->structure)) {
                $this->setOrderBy($this->db->structure->getPrimaryKey($array[0]), true);
                $this->setLimit(1);
            } else {
                trigger_error("Unable to set wildcard '?' in key '${key}'. Need to set database structure first.", E_USER_NOTICE);
            }
        }

        /**
         * 2.2.3) select column (if any)
         */
        if (empty($array[2])) {
            $this->setColumn('*');
        } elseif (!$this->setColumn($array[2])) {
            return false;
        }

        return true;

    }

    /**
     * set column to sort the resultset by
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * @access  public
     * @param   string|array  $orderBy  column name / list of column names
     * @param   bool          $desc     sort descending (true=yes, false=no)
     * @return  bool
     */
    function setOrderBy($orderBy, $desc = false)
    {
        assert('is_string($orderBy) || is_array($orderBy); // Wrong argument type for argument 1. String expected.');
        $this->id = null;
        /**
         * 1) check input
         */
        if (!is_string($orderBy) && !is_array($orderBy)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($orderBy)), E_USER_WARNING);
            return false;
        } elseif (!is_bool($desc)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 2, 'Boolean', gettype($desc)), E_USER_WARNING);
            return false;
        }

        /**
         * 2) set order by
         *
         * 2.1) reset
         */
        if (empty($orderBy)){
            $this->orderBy = array();

        /**
         * 2.2) set to value
         */
        } else {
            /* settype to ARRAY */
            settype($orderBy, "array");

            /* check and lower-case all values */
            assert('!isset($column); // Cannot redeclare var $column');
            assert('!isset($i); // Cannot redeclare var $i');
            assert('!isset($table); // Cannot redeclare var $table');
            foreach($orderBy as $i => $column)
            {
                /**
                 * 2.2.1) get base table
                 */
                if (strpos($column, '.')) {
                    list($table, $column) = explode('.', $column);
                } else {
                    $table = $this->table;
                }

                /**
                 * 2.2.2) check if column exists
                 */
                if (!$this->db->structure->isColumn($table, $column)) {

                    trigger_error("The column '$column' is not found in table '" . $table . "'.", E_USER_WARNING);
                    return false;

                } else {

                    $orderBy[$i] = array($table, strtolower($column));

                }
            }

            $this->orderBy = $orderBy;
        } /* end foreach */
        unset($i, $column, $table);

        /**
         * 3) set direction
         */
        if ($desc) {
            if (empty($this->orderBy)) {
                $this->orderBy = array($this->db->structure->getPrimaryKey($this->table));
            }
            $this->desc = true;
        } else {
            $this->desc = false;
        }
        return true;
    }

    /**
     * get the list of columns the resultset is ordered by
     *
     * Returns a lower-cased list of column names.
     * If none has been set yet, then the list is empty.
     *
     * @access  public
     * @return  array
     */
    function getOrderBy()
    {
        if (is_array($this->orderBy)) {
            return $this->orderBy;
        } else {
            /* error: unexpected value */
            return array();
        }
    }

    /**
     * check if resultset is sorted in descending order
     *
     * Returns bool(true) if sorted descending and bool(false) otherwise.
     *
     * @access  public
     * @return  bool
     */
    function isDescending()
    {
        $this->id = null;
        if (is_bool($this->desc)) {
            return $this->desc;
        } else {
            /* error: unexpected value */
            return false;
        }
    }

    /**
     * set where clause (filter)
     *
     * The syntax is as follows:
     * column1=value1,column2=value2,...,columnN=valueN
     *
     * The alternative is as follows:
     * array(array(0=>column1,1=>value1,2=>operator1),...)
     * Where "operator" can be one of the following:
     * '=', 'REGEXP', 'LIKE', '<', '>', '!=', '<=', '>='
     *
     * To unset the where clause, call this function without
     * providing a parameter.
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * @access  public
     * @param   string|array  $where
     * @return  bool
     */
    function setWhere($where = null)
    {
        $this->id = null;
        global $YANA;

        if (is_null($where)) {

            $this->where = array();
            return true;

        } elseif (!is_array($where) && !is_string($where)) {

            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($where)), E_USER_WARNING);
            return false;

        }

        /**
         * 1) check if the request is a table scan
         *
         * Reason: When scanning a whole table you may
         * search for primary keys - otherwise not.
         */
        if ($this->row === '*' || is_null($this->row) || $this->row === '?') {
            $isTableScan = true;
        } else {
            $isTableScan = false;
        }

        /**
         * 2) convert string to array
         *
         * This applies only if the input is not
         * already an array.
         */
        if (empty($where)) {
            $where = array();
        } elseif (is_string($where)) {
            if (!preg_match("/^((, ?)?([\w\d-_\.])+=([\w\d-_\.\@ \&;])+)*$/i", $where)) {
                if (isset($YANA)) {
                    $YANA->report( new Log("SQL-Statement rejected du to invalid where clause.", $where) );
                }
                trigger_error("The where clause you provided contains syntax errors.\n\t\tMake sure it does only contain alpha-numeric characters and matches the following syntax: fieldname1=value1,fieldname2=value2", E_USER_WARNING);
                return false;
            }

            $where = preg_split("/, ?/", $where);
            /* settype to ARRAY */
            $where = (array) $where;
            assert('!isset($a); /* cannot redeclare variable $a */');
            assert('!isset($b); /* cannot redeclare variable $b */');
            while (list($a, $b) = each($where))
            {
                if (is_string($b)) {
                    $where[$a] = explode("=", $b);
                    /**
                     * 2.1) is foreign key
                     */
                    if ($this->db->structure->isForeignKey($this->table, $where[$a][0])) {
                        $where[$a][0] = array($this->table, strtolower($where[$a][0]));
                        $where[$a][1] = strtoupper($where[$a][1]);
                        $where[$a][2] = '=';
    
                    /**
                     * 2.2) is primary key
                     */
                    } elseif ($this->db->structure->isPrimaryKey($this->table, $where[$a][0])) {
                        $where[$a][0] = array($this->table, strtolower($where[$a][0]));
                        $where[$a][1] = strtoupper($where[$a][1]);
                        $where[$a][2] = '=';
                        $this->rowId = $where[$a];
                        unset($where[$a]);
                        $where = array_values($where);
                        reset($where);
                        continue;

                    /**
                     * 2.3) is other
                     */
                    } else {
                        switch ($this->db->structure->getType($this->table, $where[$a][0]))
                        {
                            case 'image':
                            case 'file':
                                trigger_error("Invalid where clause. Cannot search in 'blob' columns.", E_USER_NOTICE);
                                continue 2;
                            break;
                            case 'time':
                                trigger_error("Invalid where clause. Cannot use operators '=', 'like' in 'time' columns.", E_USER_NOTICE);
                                continue 2;
                            break;
                            case 'array':
                            case 'string':
                            case 'select':
                            case 'text':
                            case 'set':
                                $where[$a][1] = '%' . $where[$a][1] . '%';
                                $where[$a][2] = 'like';
                            break;
                            default:
                                $where[$a][2] = '=';
                            break;
                        }
                        $where[$a][0] = array($this->table, strtolower($where[$a][0]));
                    }
                }
            } /* end foreach */
            reset($where);
            unset($a, $b);
        }

        assert('is_array($where); // unexpected result: $where should have been converted to an array - but it is not.');

        /**
         * 3) check the array
         */
        if (count($where) > 0) {

            /**
             * 3.a) adding a single condition
             */
            if (count($where) === 3 && isset($where[2]) && is_string($where[2])) {
                assert('isset($where[0]); /* unexpected argument: where clause is invalid at index 0 */');
                assert('isset($where[1]); /* unexpected argument: where clause is invalid at index 1 */');
                $where = array($where);
            }

            assert('!isset($isValid); /* cannot redeclare variable $isValid */');
            assert('!isset($i); /* cannot redeclare variable $i */');
            assert('!isset($clause); /* cannot redeclare variable $clause */');
            $isValid = true;
            foreach ($where as $i => $clause)
            {
                if (empty($clause)) {
                    continue;
                } elseif (is_array($clause) && count($clause) >= 2 && count($clause) <= 3 && !empty($clause[0])) {

                    assert('!isset($table); // Cannot redeclare variable $table');
                    assert('!isset($column); // Cannot redeclare variable $column');

                    /**
                     * 3.1) handle reference column
                     */
                    if (is_array($clause[0]) && count($clause[0]) === 2) {
                        $table = strtolower(array_shift($clause[0]));
                        $column = strtolower(array_shift($clause[0]));
                    } elseif (is_string($clause[0])) {
                        $table = $this->table;
                        $column = strtolower( $clause[0] );
                    } else {

                        $isValid = false;
                        trigger_error("Missing column name in where clause.", E_USER_WARNING);
                        break;
                    }

                    /**
                     * 3.1.1) check if table - column pair is valid
                     */
                    if ($this->db->structure->isStrict()) {
                        assert('is_string($column); // Unexpected result: $column. String expected.');
                        if ($table !== $this->table && !$this->db->structure->isTable($table)) {

                            $isValid = false;
                            trigger_error("Invalid where clause. The name '${table}' is not a table.", E_USER_WARNING);
                            break;

                        } elseif (!$this->db->structure->isColumn($table, $column)) {

                            $isValid = false;
                            trigger_error("Invalid where clause. The name '${column}' is not a column in table '${table}'.", E_USER_WARNING);
                            break;

                        } elseif (!$isTableScan && $this->db->structure->isPrimaryKey($table, $column)) {

                            $isValid = false;
                            trigger_error("Invalid where clause. You are trying to search for a primary key.\n\t\t" .
                            "This is not allowed, since it might cause results wether to be ambigious or empty.\n\t\t" .
                            "Try setting 'USE_STRICT' to 'false' if you want to turn this message off and continue without checking.", E_USER_WARNING);
                            break;

                        }
                    }
                    $where[$i][0] = array($table, $column);
                    unset($table, $column);

                    /**
                     * 3.2) handle compared value / cell
                     */
                    if (is_array($clause[1]) && count($clause[1]) === 2) {
                        $table = strtolower(array_shift($clause[1]));
                        $column = strtolower(array_shift($clause[1]));

                        /**
                         * 3.2.1) check if table - column pair is valid
                         */                        
                        if ($this->db->structure->isStrict()) {
                            if ($table !== $this->table && !$this->db->structure->isTable($table)) {

                                $isValid = false;
                                trigger_error("Invalid where clause. The name '${table}' is not a table.", E_USER_WARNING);
                                break;
    
                            } elseif (!$this->db->structure->isColumn($table, $column)) {

                                $isValid = false;
                                trigger_error("Invalid where clause. The name '${column}' is not a column in table '${table}'.", E_USER_WARNING);
                                break;
    
                            }
                        }
                        $where[$i][1] = array($table, $column);
                        unset($table, $column);
                    } elseif (!is_null($clause[1])) {
                        $where[$i][1] = strtolower( (string) $clause[1] );
                    } else {
                        $where[$i][1] = null;
                    }

                    /**
                     * 3.3) handle operator
                     */
                    if (isset($clause[2])) {
                        /* map aliases */
                        if ($clause[2] === '==') {
                            $where[$i][2] = '=';
                        } else {
                            $operator = strtolower((string) $clause[2]);

                            /**
                             * 3.3.1) check if operator is valid
                             */
                            if (is_null($where[$i][1]) && !in_array($operator, array('=', '!='))) {
        
                                $isValid = false;
                                trigger_error("Invalid where clause.\n\t\t" .
                                "The operator '${operator}' is not supported when comparing a column with NULL.", E_USER_WARNING);
                                break;
        
                            } elseif (!in_array($operator, array('=', 'like', 'regexp', '<', '>', '!=', '<=', '>='))) {
        
                                $isValid = false;
                                trigger_error("Invalid where clause.\n\t\t" .
                                "The operator '${operator}' is not supported.", E_USER_WARNING);
                                break;
        
                            }
                            $where[$i][2] = $operator;
                        }
                    } else {
                        $where[$i][2] = '=';
                    }

                } else {
                    trigger_error("Invalid where clause.\n\t\t" .
                    "Malformed argument '" . print_r($clause, true) . "'.", E_USER_WARNING);
                    $isValid = false;
                    break;
                }
            } /* end foreach */

            if (!$isValid) {
                if (isset($YANA)) {
                    $YANA->report( new Log("SQL-Statement rejected due to invalid where clause.", $clause) );
                }
                trigger_error("SQL-Statement rejected due to invalid where clause: '" . print_r($clause, true) . "'.", E_USER_WARNING);
                return false;
            } else {
                $this->where = $where;
                return true;
            }
            unset($i, $clause, $isValid);

        } else {
            $this->where = array();
            return true;
        }
    }

    /**
     * get the currently set where clause
     *
     * Returns the current where clause.
     *
     * @access  public
     * @return  array
     */
    function getWhere()
    {
        if (is_array($this->where)) {
            $where = $this->where;
            if (is_array($this->profile)) {
                $where[] = $this->profile;
            }
            if (is_array($this->rowId)) {
                $where[] = $this->rowId;
            }
            if ($this->type === YANA_DB_EXISTS && !empty($this->column)) {
                assert('!isset($column); // Cannot redeclare var $column');
                foreach ($this->getColumns() as $column)
                {
                    $where[] = array($column, null, '!=');
                }
                unset($column);
            }
            return $where;
        } else {
            return array();
        }
    }

    /**
     * join the resultsets for two tables
     *
     * This will join the currently selected table with another.
     *
     * If $table is not provided, this will reset the list of joined tables.
     * If $key1 is not provided, the function will automatically search for
     * a suitable foreign key, that refers to $table.
     * If $key2 is not provided, the function will automatically look up
     * the primary of $table and use it instead.
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * @access  public
     * @param   string $table  name of another table to join the current table with (when omitted will remove all previously set joins)
     * @param   string $key1   name of the foreign key in current table (when omitted the API will look up the key in the structure file)
     * @param   string $key2   name of the key in foreign table that is referenced (may be omitted if it is the primary key)
     * @return  bool
     */
    function setJoin($table = null, $key1 = null, $key2 = null)
    {
        $this->id = null;

        if (!is_string($table) && !is_null($table)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($table)), E_USER_WARNING);
            return false;
        }        
        if (!is_string($key1) && !is_null($key1)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 2, 'String', gettype($key1)), E_USER_WARNING);
            return false;
        }        
        if (!is_string($key2) && !is_null($key2)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 3, 'String', gettype($key2)), E_USER_WARNING);
            return false;
        }

        /* 1 check input */
        /* 1.1 argument 1 */
        if (is_null($table)) {

            /* if no arguments are provided, reset the list of joined tables */

            $this->joins = array();
            return true;

        } else {
            /* lower-case input */
            $table = strtolower($table);
        }

        /* 1.2 argument 2 */
        if (is_null($key1)) {

            /* if no key is provided, take the first association available */

            assert('!isset($foreignKeys); /* Cannot redeclare variable $foreignKeys */');
            $foreignKeys = $this->db->structure->getForeignKeys($this->table);

            if (empty($foreignKeys)) {
                trigger_error("Cannot join tables '".$this->table."' and '${table}'. " .
                "No foreign key constraint has been found.", E_USER_NOTICE);
                return false;
            }

            assert('!isset($foreignTable); /* Cannot redeclare variable $foreignTable */');
            assert('!isset($foreignKey); /* Cannot redeclare variable $foreignKey */');
            foreach ($foreignKeys as $foreignKey => $foreignTable)
            {
                if (strcasecmp($table, $foreignTable) === 0) {
                    $key1 = strtolower($foreignKey);
                    $key2 = $this->db->structure->getPrimaryKey($foreignTable);
                    break;
                }
            }
            unset($foreignKey, $foreignKeys, $foreignTable);

        } else  {
            /* lower-case input */
            $key1 = strtolower($key1);
        }
        /* 1.3 argument 3 */
        if (is_null($key2)) {

            /* if the target column is not provided, take the primary key instead */
            $key2 = $this->db->structure->getPrimaryKey($table);

        } else {
            /* lower-case input */
            $key2 = strtolower($key2);
        }

        /* 1.4.a) strict input checking */
        if ($this->db->structure->isStrict()) {

            /**
             * error - no such table
             */
            if (!$this->db->structure->isTable($table)) {
                trigger_error("Cannot join tables '" . $this->table . "' and '${table}'. " .
                "Table does not exist in structure file.", E_USER_WARNING);
                return false;

            /**
             * error - no such column in current table
             */
            } elseif (!is_null($key1) && !$this->db->structure->isColumn($this->table, $key1)) {
                trigger_error("Cannot join tables '" . $this->table . "' and '${table}'. " .
                "Field '${key1}' does not exist in table '" . $this->table . "'.", E_USER_WARNING);
                return false;

            /**
             * error - no such column in referenced table
             */
            } elseif (!is_null($key2) && !$this->db->structure->isColumn($table, $key2)) {
                trigger_error("Cannot join tables '" . $this->table . "' and '${table}'. " .
                "Field '${key2}' does not exist in table '${table}'.", E_USER_WARNING);
                return false;

            /**
             * test passed - proceed
             */
            } else {
                /* intentionally left blank */
            }

        /* 1.4.b) loose input checking */
        } else {

            for ($i = 0; $i < func_num_args(); $i++)
            {
                if (!preg_match("/^[\"\'`]?[\w\d-_]+[\"\'`]?$|^\*$/si", func_get_arg($i))) {
                    trigger_error("Argument ${i} is not a valid identifier.", E_USER_WARNING);
                }
            }

        }

        /* 2. reset old association */
        if (isset($this->joins[$table])) {
            unset($this->joins[$table]);
        }

        /* 3. create new association */
        $this->joins[$table] = array($key1, $key2);

        return true;

    }

    /**
     * get a list of the tables, the query has joined together
     *
     * If $table is not provided, all tables are returned.
     *
     * @access  public
     * @param   string  $table
     * @return  array
     */
    function getJoin($table = null)
    {
        if (is_null($table)) {

            return $this->joins;

        } elseif (!is_string($table)) {

            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($table)), E_USER_WARNING);
            return false;

        } elseif (!$this->db->structure->isTable($table)) {

            trigger_error("The table '$table' is unknown.", E_USER_WARNING);
            return false;

        } else {

            $table = strtolower($table);

        }

        if (!isset($this->joins[$table])) {

            return false;

        } else {

            return $this->joins[$table];

        }
    }

    /**
     * set value(s) for current query
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * @access  public
     * @param   mixed  &$values
     * @return  bool
     */
    function setValues(&$values)
    {
        $this->id = null;
        /**
         * 1.a) lowercase array keys (as expected by function DbStructure::checkRow())
         */
        if (is_array($values)) {
            $values = Hashtable::changeCase($values, CASE_LOWER);

        /**
         * 1.b) error - wrong argument type
         */
        } elseif (!is_scalar($values)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'Scalar or array', gettype($values)), E_USER_WARNING);
            return false;

        /**
         * 1.c) error - table not set
         */
        } elseif (empty($this->table)) {
            trigger_error("Cannot set values - need to set table first!", E_USER_WARNING);
            return false;
        }

        /**
         * 1.d) handle table inheritance
         */
        if ($this->getParent()) {
            assert('!isset($columns); // Cannot redeclare var $columns');
            $columns = $this->getColumns();
            if (empty($columns)) {
                assert('!isset($columnName); // Cannot redeclare var $columnName');
                assert('!isset($parent); // Cannot redeclare var $parent');
                foreach (array_keys($values) as $columnName)
                {
                    if (false !== ($parent = $this->getParentByColumn($columnName))) {
                        $this->appendValue($parent, $columnName, $values[$columnName]);
                        unset($values[$columnName]);
                    }
                }
                unset($parent, $columnName);
            } else {
                assert('!isset($column); // Cannot redeclare var $column');
                foreach ($columns as $column)
                {
                    assert('is_array($column); // Invalid property "column". Two-dimensional array expected.');
                    assert('!isset($parent); // Cannot redeclare var $parent');
                    if (false !== ($parent = $this->getParentByColumn($column[1]))) {
                        $this->appendValue($parent, $column[1], $values);
                        return true;
                    }
                    unset($parent);
                }
                unset($column);
            }
            unset($columns);
        }

        /**
         * 2.a) inserting a row
         */
        assert('!isset($primaryKey); /* Cannot redeclare var $primaryKey */');
        $primaryKey = $this->db->structure->getPrimaryKey($this->table);

        if ($this->type === YANA_DB_INSERT) {
            $isInsert = true;

            /* invalid argument type */
            if (!is_array($values)) {
                trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'Array', gettype($values)), E_USER_WARNING);
                return false;
            }

            /**
             * 2.a.1) copy primary key from row property or vice versa
             */
            if (!isset($values[$primaryKey])) {
                if ($this->db->structure->isAutonumber($this->table, $primaryKey)) {
                    /* ignore - is to be inserted automatically by database */
                } elseif ($this->row !== '*') {
                    $values[$primaryKey] = $this->db->structure->untaintInput($this->table, $primaryKey, $this->row, YANA_ESCAPE_SLASHED);
                } else {
                    trigger_error("Cannot insert a row without a primary key. Operation aborted.", E_USER_WARNING);
                    return false;
                }
            } elseif ($this->row === '*') {
                $this->setRow($values[$primaryKey]);
            } elseif (strcasecmp($this->row, $values[$primaryKey]) === 0) {
                /* both the same - nothing to do */
            } else {
                trigger_error("Cannot set values. The primary key is ambigious.\n\t\t" .
                "The primary key has been set via " . __CLASS__ . "->setRow() or " . __CLASS__ . "->setKey() to '" . $this->row . "'.\n\t\t" .
                "However, the primary key provided with " . __CLASS__ . "->setValues() is '" . $values[$primaryKey] . "'.", E_USER_WARNING);
                return false;
            }

        /**
         * 2.b) updating a row / cell
         */
        } elseif ($this->type === YANA_DB_UPDATE) {
            $isInsert = false;

            /**
             * 2.b.1) copy primary key to row property
             */
            if ($this->expectedResult === YANA_DB_TABLE && isset($values[$primaryKey])) {
                $this->setRow($values[$primaryKey]);
            }

        /**
         * 2.c) error - type of operation not set
         */
        } elseif ($this->type === YANA_DB_UNKNOWN) {
            trigger_error("Cannot set values - need to set type of operation first!", E_USER_WARNING);
            return false;

        /**
         * 2.d) error - invalid type of operation
         */
        } else {
            trigger_error("Cannot set values. This is only available with the operations 'insert' and 'update'.", E_USER_WARNING);
            return false;

        }

        /* upper-case primary key */
        if ($this->expectedResult === YANA_DB_ROW && isset($values[$primaryKey])) {
            $values[$primaryKey] = strtoupper($values[$primaryKey]);
        }
        unset($primaryKey);

        /**
         * 3) error - row is invalid
         */
        if (!$this->db->structure->checkRow($this->table, $values, $isInsert)) {
            trigger_error("Cannot set values. Row is invalid.", E_USER_WARNING);
            return false;

        /**
         * 4) error - access denied
         */
        } elseif (!$this->_checkProfile($values)) {
            trigger_error("Cannot set values. Profile constraint mismatch.", E_USER_WARNING);
            return false;


        /**
         * 6) input is valid - update values
         */
        } else {
            $this->values =& $values;
            return true;
        }
    }

    /**
     * get the list of values
     *
     * If none are available, NULL (not bool(false)!) is returned.
     *
     * @access  public
     * @return  mixed
     */
    function &getValues()
    {
        return $this->values;
    }

    /**
     * get the parent of a table
     *
     * Returns the name of the parent table as a string, if the table inherits from another,
     * or bool(false) if there is no parent.
     *
     * @access  private
     * @param   string  $table
     * @param   string  $column
     * @param   mixed   $value
     * @return  bool
     * @since   2.9.6
     */
    function appendValue($table, $column, $value)
    {
        assert('is_string($table); // Wrong argument type argument 1. String expected');
        assert('is_string($column); // Wrong argument type argument 2. String expected');

        if (!isset($this->queue[$table])) {
            $this->queue[$table] = array();
        }
        
        /**
         * 1) append value to queue
         */
        if (!isset($this->queue[$table][$column])) {
            $this->queue[$table][$column] = $value;
            return true;
            
        /**
         * 2) error - duplicate key
         */
        } else {
            trigger_error("Invalid query. Duplicate value for column $table.$column.", E_USER_WARNING);
            return false;
        }
    }

    /**
     * get the currently selected limit
     *
     * Note: This setting will not be part of the sql statement
     * produced by {link DbQuery::toString()}.
     * Use the API's $limit and $offset parameter instead when sending
     * the query.
     *
     * This restriction does not apply if you use
     * {link DbQuery::sendQuery()}.
     *
     * Note: For security reasons all delete queries will automatically
     * be limited to 1 row at a time.
     *
     * @access  public
     * @return  int
     * @since   2.9.3
     */
    function getLimit()
    {
        if ($this->type === YANA_DB_DELETE) {
            return 1;
        } else {
            return (int) $this->limit;
        }
    }

    /**
     * set a limit for this query
     *
     * Note: This setting will not be part of the sql statement
     * produced by {link DbQuery::toString()}.
     * Use the API's $limit and $offset parameter instead when sending
     * the query.
     *
     * This restriction does not apply if you use
     * {link DbQuery::sendQuery()}.
     *
     * Note: For security reasons all delete queries will automatically
     * be limited to 1 row at a time.
     *
     * @access  public
     * @param   int  $limit
     * @return  bool
     * @since   2.9.3
     */
    function setLimit($limit)
    {
        assert('is_int($limit); // Wrong argument type for argument 1. Integer expected.');
        $this->id = null;
        if (is_int($limit) && $limit >= 0) {
            $this->limit = $limit;
            return true;
        } else {
            return false;
        }
    }

    /**
     * get the currently selected offset
     *
     * Note: This setting will not be part of the sql statement
     * produced by {link DbQuery::toString()}. Use the API's $limit and
     * $offset parameter instead when sending the query.
     *
     * This restriction does not apply if you use {link DbQuery::sendQuery()}.
     *
     * Note: For security reasons all delete queries will automatically
     * have an offset of 0.
     *
     * @access  public
     * @return  int
     * @since   2.9.3
     */
    function getOffset()
    {
        if ($this->type === YANA_DB_DELETE) {
            return 0;
        } else {
            return (int) $this->offset;
        }
    }

    /**
     * set an offset for this query
     *
     * Note: This setting will not be part of the sql statement
     * produced by {link DbQuery::toString()}. Use the API's $limit and
     * $offset parameter instead when sending the query.
     *
     * This restriction does not apply if you use {link DbQuery::sendQuery()}.
     *
     * Note: For security reasons all delete queries will automatically
     * have an offset of 0.
     *
     * @access  public
     * @param   int  $offset
     * @return  bool
     * @since   2.9.3
     */
    function setOffset($offset)
    {
        assert('is_int($offset); // Wrong argument type for argument 1. Integer expected.');
        $this->id = null;
        if (is_int($offset) && $offset >= 0) {
            $this->offset = $offset;
            return true;
        } else {
            return false;
        }
    }

    /**
     * check profile constraint
     *
     * @access  private
     * @param   mixed   &$value
     * @return  bool
     * @since   2.9.3
     * @ignore
     */
    function _checkProfile(&$value)
    {
        global $YANA;
        $profile = $this->db->structure->getProfile($this->table);

        $table = $this->getTable();
        $row = $this->getRow();

        if (empty($profile)) {
            return true;
        } else {

            $profile = strtolower($profile);
            switch ($this->getType())
            {
                /**
                 * 1) If user does not have permission to write new entries,
                 *    issue an error message and abort.
                 */
                case YANA_DB_INSERT:
                    if (isset($YANA)) {
                        if ($YANA->session->checkPermission() !== true) {
                            $YANA->report(new Alert("INSUFFICIENT_RIGHTS"));
                            return false;
                        } else {
                            $value[$profile] = $YANA->session->getId();
                            return true;
                        }
                    }
                break;
                /**
                 * 2) If user does not have permission to update the entry,
                 *    issue an error message and abort.
                 */
                case YANA_DB_UPDATE:
                    if ($row === '*') {
                        trigger_error("SQL security restriction. Cannot update a table (only rows and cells).", E_USER_WARNING);
                        return false;
                    }                
                    /**
                     * build query: select profile_id from table where id = "foo"
                     */
                    assert('!isset($dbQuery); /* Cannot redeclare variable $dbQuery */');
                    $dbQuery = new DbQuery($this->db);
                    $dbQuery->type = YANA_DB_SELECT;
                    $dbQuery->expectedResult = YANA_DB_CELL;
                    $dbQuery->table = $table;
                    $dbQuery->column = array(array($table, $profile));
                    $dbQuery->row = $row;
                    $dbQuery->limit = 1;
                    $result = $dbQuery->sendQuery();
                    unset($dbQuery);
                    /**
                     * handle result
                     */
                    if ($this->db->isError($result)) {
                        trigger_error("Unable to update entry ${table}.${row}.\n\t\t".$result->getMessage(), E_USER_WARNING);
                        return false;
                    } else {
                        $resultRow = $result->fetchRow(DB_FETCHMODE_ASSOC, 0);
                        assert('is_array($resultRow); /* unexpected result: $resultRow */');
                        $profileId = array_pop($resultRow);
                    }
                    if (isset($YANA) && $YANA->session->checkPermission($profileId) !== true) {
                        $YANA->report(new Alert("INSUFFICIENT_RIGHTS"));
                        return false;
                    } else {
                        switch ($this->getExpectedResult())
                        {
                            case YANA_DB_ROW:
                                if (isset($value[$profile]) && $value[$profile] != $profileId) {
                                    trigger_error("Security restriction. The profile id of an entry may not be changed.", E_USER_WARNING);
                                    if (isset($YANA)) {
                                        $YANA->report(new Log("Security restriction. The profile id of an entry may not be changed."));
                                    }
                                    return false;
                                } else {
                                    return true;
                                }
                            break;
                            case YANA_DB_CELL:
                                if (strcasecmp($this->getColumn(), $profile) === 0) {
                                    trigger_error("Security restriction. The profile id of an entry may not be changed.", E_USER_WARNING);
                                    if (isset($YANA)) {
                                        $YANA->report(new Log("Security restriction. The profile id of an entry may not be changed."));
                                    }    
                                    return false;
                                } else {
                                    return true;
                                }
                            break;
                            default:
                                return false;
                            break;
                        }
                    }
                break;
                /**
                 * 3) If user does not have permission to delete the entry,
                 *    issue an error message and abort.
                 */
                case YANA_DB_DELETE:
                    if ($row === '*') {
                        trigger_error("SQL security restriction. Cannot delete a table (only rows).", E_USER_WARNING);
                        return false;
                    }                
                    /**
                     * build query: select profile_id from table where id = "foo"
                     */
                    assert('!isset($dbQuery); /* Cannot redeclare variable $dbQuery */');
                    $dbQuery = new DbQuery($this->db);
                    $dbQuery->type = YANA_DB_SELECT;
                    $dbQuery->expectedResult = YANA_DB_CELL;
                    $dbQuery->table = $table;
                    $dbQuery->column = array(array($table, $profile));
                    $dbQuery->row = $row;
                    $dbQuery->limit = 1;
                    $result = $dbQuery->sendQuery();
                    unset($dbQuery);
                    /**
                     * handle result
                     */
                    if ($this->db->isError($result)) {
                        trigger_error("Unable to delete entry ${table}.${row}.\n\t\t".$result->getMessage(), E_USER_WARNING);
                        return false;
                    } else {
                        $resultRow = $result->fetchRow(DB_FETCHMODE_ASSOC, 0);
                        assert('is_array($resultRow); /* unexpected result: $resultRow */');
                        $profileId = array_pop($resultRow);
                    }
                    if (isset($YANA)) {
                        if ($YANA->session->checkPermission($profileId) !== true) {
                            $YANA->report(new Alert("INSUFFICIENT_RIGHTS"));
                            return false;
                        } else {
                            return true;
                        }
                    }
                break;
                /**
                 * 4) syntax error
                 */
                default:
                    trigger_error("Syntax error. The input '${table}.${column}.${row}' is not a valid key address.", E_USER_WARNING);
                    return false;
                break;
            } /* end switch */
            return true;
        }
    }

    /**
     * get unique id
     *
     * @access  public
     * @return  string
     * @since   2.9.3
     * @ignore
     */
    function toId()
    {
        if (!isset($this->id)) {
            $this->id = serialize(array($this->type, $this->table, $this->column, $this->row,
            $this->where, $this->orderBy, $this->desc, $this->joins, $this->values));
        }
        return $this->id;
    }

    /**
     * send query to server
     *
     * This sends the query to the database and returns a result-object.
     *
     * @access  public
     * @return  object
     * @since   2.9.3
     * @ignore
     */
    function sendQuery()
    {
        if (count($this->queue) === 0) {
            return $this->db->query($this);

        } else {
            $result = $this->db->query($this);
            if ($this->db->isError($result)) {
                return $result;
            } else {
                $row = $this->getRow();
                $prevSetting = $this->useInheritance(false);
                $result = null;
                assert('!isset($table); // Cannot redeclare var $table');
                assert('!isset($values); // Cannot redeclare var $values');
                foreach ($this->queue as $table => $values)
                {
                    if (!$this->setTable($table)) {
                        $this->db->handleError("No such table", $table);
                        break;
                    } elseif (!$this->setRow($row)) {
                        $this->db->handleError("Invalid row", $row);
                        break;
                    } elseif (!$this->setValues($values)) {
                        $this->db->handleError("Invalid values", $values);
                        break;
                    } else {
                        $result = $this->db->query($this);
                        if ($this->db->isError($result)) {
                            break;
                        }
                    }
                }
                unset($table, $values);
                $this->useInheritance($prevSetting);
                $this->queue = array();
                return $result;
            }

        }
    }

    /**
     * build a SQL-query
     *
     * @access  public
     * @return  string
     */
    function toString()
    {
        /* prepare where clause */
        $where = $this->getWhere();

        if (is_array($where) && count($where) > 0) {

            assert('!isset($_where); // Cannot redeclare $_where');
            $_where = "";
            /* build filter */
            if (count($where) > 0) {
                assert('!isset($clause); // Cannot redeclare $clause');
                foreach ($where as $clause)
                {
                    if (empty($clause)) {
                        continue;
                    } else {
                        if (is_array($clause[0])) {
                            $clause[0] = $this->db->quoteId(YANA_DATABASE_PREFIX.$clause[0][0]) . '.' . $clause[0][1];
                        }
                        if (is_array($clause[1])) {
                            $clause[1] = $this->db->quoteId(YANA_DATABASE_PREFIX.$clause[1][0]) . '.' . $clause[1][1];
                        } elseif (is_string($clause[1])) {
                            $clause[1] = $this->db->quote($clause[1]);
                        }
                        if (is_null($clause[1])) {
                            if ($clause[2] == '=') {
                                $_where .= (($_where !== '') ? 'AND ' : '') . $clause[0] . ' is null ';
                            } elseif ($clause[2] == '!=') {
                                $_where .= (($_where !== '') ? 'AND ' : '') . $clause[0] . ' is not null ';
                            } else {
                                trigger_error("The invalid operator '" . $clause[2] . "' in your where clause has been ignored.", E_USER_NOTICE);
                                continue;
                            }
                        } else {
                            $_where .= (($_where != '') ? 'AND ' : '') . $clause[0] . ' ' . $clause[2] . ' ' . $clause[1] . ' ';
                        }
                    }
                }
                unset($clause);
                if (!empty($_where)) {
                    $_where = 'WHERE ' . $_where;
                }
            }
            $where = $_where;
            unset($_where);

        } else {

            $where = "";

        }

        /* select a statement */
        switch ($this->type)
        {
            case YANA_DB_SELECT:
                $stmt = "SELECT %COLUMN% FROM %TABLE% %WHERE% %ORDERBY% %DESC%";
            break;

            case YANA_DB_EXISTS:
                $stmt = "SELECT 1 FROM %TABLE% %WHERE%";
            break;

            case YANA_DB_LENGTH:
                $stmt = "SELECT count(%COLUMN%) FROM %TABLE% %WHERE%";
            break;

            case YANA_DB_UPDATE:
                $stmt = "UPDATE %TABLE% SET %SET% %WHERE%";

                /**
                 * replace %SET%
                 *
                 * Note: this is done here, since all other types
                 * of statements do not have this token.
                 */
                assert('!isset($set); // Cannot redeclare $set');
                $set = "";
                if ($this->expectedResult === YANA_DB_ROW) {
                    if (is_array($this->values)) {
                        assert('!isset($column); // Cannot redeclare $column');
                        assert('!isset($value);  // Cannot redeclare $value');
                        foreach ($this->values as $column => $value)
                        {
                            if ($set !== '') {
                                $set .= ', ';
                            }
                            if (is_array($value)) {
                                $set .= $column . ' = ' . $this->db->quote(SML::encode($value));
                            } elseif (!is_null($value)) {
                                $set .= $column . ' = ' . $this->db->quote($value);
                            }
                        }
                        unset($column, $value);
                    } else {
                        trigger_error("Cannot build update statement. No valid values provided.", E_USER_WARNING);
                        return "";
                    }
                } elseif ($this->expectedResult === YANA_DB_CELL) {
                    if (is_array($this->values)) {
                        $set = $this->getColumn() . ' = ' . $this->db->quote(SML::encode($this->values));
                    } else {
                        $set = $this->getColumn() . ' = ' . $this->db->quote($this->values);
                    }
                } else {
                    trigger_error("Cannot build update statement. No row or cell selected for update.", E_USER_WARNING);
                    return "";
                }
                $stmt = str_replace('%SET%', $set, $stmt);
                unset($set);

            break;

            case YANA_DB_INSERT:
                $stmt = "INSERT INTO %TABLE% (%KEYS%) VALUES (%VALUES%)";

                /**
                 * replace %KEYS% and %VALUES%
                 *
                 * Note: this is done here, since all other types
                 * of statements do not have these token.
                 */
                if (is_array($this->values) && count($this->values) > 0) {

                    assert('!isset($keys);   // Cannot redeclare $keys');
                    assert('!isset($values); // Cannot redeclare $values');
                    $keys = implode(', ', array_keys($this->values));
                    $values = '';
                    assert('!isset($value); // Cannot redeclare $values');
                    foreach ($this->values as $value)
                    {
                        $values .= (($values !== '') ? ', ' : '' );
                        if (is_array($value)) {
                            $values .= $this->db->quote(SML::encode($value));
                        } else {
                            $values .= $this->db->quote($value);
                        }
                    }
                    unset($value);
                    $stmt = str_replace('%KEYS%', $keys, $stmt);
                    $stmt = str_replace('%VALUES%', $values, $stmt);
                    unset($keys, $values);

                } else {
                    return "";
                }

            break;

            case YANA_DB_DELETE:
                $stmt = "DELETE FROM %TABLE% %WHERE% %ORDERBY% %DESC%";
            break;

            default:
                return "";
            break;

        }

        /* 1. replace %TABLE% */
        $table = $this->getTable();
        if (is_string($table)) {
            $table = $this->db->quoteId(YANA_DATABASE_PREFIX.$table);

            if (!empty($this->joins) && ($this->type === YANA_DB_SELECT || $this->type === YANA_DB_EXISTS || $this->type === YANA_DB_LENGTH)) {

                assert('!isset($_where);    /* cannot redeclare variable $_where    */');
                $_where = array();

                assert('!isset($tableName); /* cannot redeclare variable $tableName */');
                assert('!isset($join);      /* cannot redeclare variable $join      */');
                foreach ($this->joins as $tableName => $join)
                {
                    /* collect tables */
                    if (empty($join)) {
                        $table .= "NATURAL JOIN " . $this->db->quoteId(YANA_DATABASE_PREFIX.$tableName) . " ";
                    } else {
                        $table .= ', ' . $this->db->quoteId(YANA_DATABASE_PREFIX.$tableName) . ' ';
                    }
                    $_where[] = array($this->table . '.' . $join[0], $tableName . '.' . $join[1]);
                } /* end foreach */
                unset($tableName, $join);

                if (count($_where) > 0) {
                    assert('!isset($clause); // Cannot redeclare $clause');
                    foreach ($_where as $clause)
                    {
                        if ($where === '') {
                            $where = 'WHERE ';
                        } else {
                            $where .= 'AND ';
                        }
                        $where .= YANA_DATABASE_PREFIX.$clause[0] . ' = ' . YANA_DATABASE_PREFIX.$clause[1] . ' ';

                    }
                    unset($clause);
                }
                unset($_where);


            }

            $stmt = str_replace('%TABLE%', $table, $stmt);

        } else {

            return false;

        }

        /* 2. replace %COLUMN% */
        if ($this->getColumn() === '*') {
            $stmt = str_replace('%COLUMN%', '*', $stmt);

        } else {
            assert('!isset($column); // Cannot redeclare $column');
            $column = "";

            assert('!isset($item); // Cannot redeclare var $item');
            foreach ($this->getColumns() as $item)
            {
                if (is_array($item)) {
                    if ($column !== "") {
                        $column .= ', ';
                    }
                    $column .= $this->db->quoteId(YANA_DATABASE_PREFIX.$item[0]) . '.' . $item[1];
                }
            }
            unset($item);

            $stmt = str_replace('%COLUMN%', $column, $stmt);
            unset($column);
        }

        /* 3. replace %WHERE% */
        assert('is_string($where); // Unexpected value $where');
        if (!empty($where)) {
            $stmt = str_replace('%WHERE%', trim($where), $stmt);
        } else {
            $stmt = str_replace(' %WHERE%', '', $stmt);
        }
        unset($where);


        /* 4. replace %ORDERBY% */
        assert('!isset($orderBy); // Cannot redeclare $orderBy');
        $orderBy = $this->getOrderBy();
        if (is_array($orderBy) && !empty($orderBy)) {
            assert('!isset($_orderBy); // Cannot redeclare var $_orderBy');
            $_orderBy = 'ORDER BY ';
            assert('!isset($i); // Cannot redeclare var $i');
            assert('!isset($element); // Cannot redeclare var $element');
            foreach ($orderBy as $i => $element)
            {
                if (is_array($element)) {
                    $_orderBy .= $element[0] . '.' . $element[1];
                } else {
                    $_orderBy .= $this->table . '.' . $element;
                }
                if (++$i < count($orderBy)) {
                    $_orderBy .= ', ';
                }
            } /* end foreach */
            unset($i, $element); /* clean up garbage */
            $stmt = str_replace('%ORDERBY%', $_orderBy, $stmt);
            unset($_orderBy); /* clean up garbage */
        } else {
            $stmt = str_replace(' %ORDERBY%', '', $stmt);
        }
        unset($orderBy);

        /* 6. replace %DESC% */
        if ($this->isDescending()) {
            $stmt = str_replace('%DESC%', 'DESC', $stmt);
        } else {
            $stmt = str_replace(' %DESC%', '', $stmt);
        }

        return $stmt;

    }

    /**
     * self test
     *
     * This function exists only for debugging purposes.
     * Do not uncomment in productive environments.
     *
     * @access  public
     * @static
     *
     * @ignore
     */
    function test($nr = 1)
    {
        switch ($nr)
        {
            case 1:
                $query = new DbQuery(new FileDb("ChecK"));
                $profileId = $GLOBALS['YANA']->session->getId();

                // test set/get table
                $query->setTable('T');
                if ($query->getTable() !== 't') {
                    return '"table": test failed';
                }

                // test set/get column
                $query->setColumn('tValue');
                if ($query->getColumn() !== 'tvalue') {
                    return '"column": test failed';
                }
                $query->setColumn('*');
                if ($query->getColumn() !== '*') {
                    return '"default column": test failed';
                }

                // test set/get row
                $query->setRow('AbC');
                if ($query->getRow() !== 'abc') {
                    return '"row": test failed';
                }
                $query->setRow('*');
                if ($query->getRow() !== '*') {
                    return '"row": test failed';
                }

                // test set/get type
                if ($query->getType() !== YANA_DB_UNKNOWN) {
                    return '"default type": test failed';
                }
                $query->setType(YANA_DB_SELECT);
                if ($query->getType() !== YANA_DB_SELECT) {
                    return '"set type": test failed';
                }

                // test set/get order by / desc
                $test = $query->getOrderBy();
                if (!is_array($test) || !count($test) === 0) {
                    return '"default order by": test failed';
                }
                $query->setOrderBy('t.TvAlUe', true);
                if ($query->getOrderBy() !== array(array('t', 'tvalue'))) {
                    return '"set order by": test failed';
                }
                if ($query->isDescending() !== true) {
                    return '"descending order": test failed';
                }
                $query->setOrderBy(array('tId', 'tvaLue'));
                if ($query->getOrderBy() !== array(array('t', 'tid'), array('t', 'tvalue'))) {
                    return '"order by": test failed';
                }
                if ($query->isDescending() !== false) {
                    return '"descending order": test failed';
                }

                // test set/get where
                $test = $query->getWhere();
                if (!is_array($test) || !count($test) === 0) {
                    return '"default where": test failed';
                }
                $query->setWhere();
                $test = $query->getWhere();
                if (!is_array($test) || !count($test) === 0) {
                    return '"reset where": test failed';
                }
                $query->setWhere('tId=aBc, tValue=0');
                $test = array(
                    array( array('t', 'tvalue'),     '0',   '='),
                    array( 'profile_id', $GLOBALS['YANA']->session->getId(), '='),
                    array( array('t', 'tid'),        'ABC', '=')
                );
                if ($query->getWhere() != $test) {
                    return '"set where (string)": test failed';
                }
                $query->setWhere(array('Tvalue', 1, 'lIke'));
                $test = array(
                    array( array('t', 'tvalue'), '1',    'like'),
                    array( 'profile_id', $GLOBALS['YANA']->session->getId(), '='),
                    array( array('t', 'tid'),     'ABC', '=')
                );
                if ($query->getWhere() != $test) {
                    return '"set where (array)": test failed';
                }

                // test set/get join
                if ($query->getJoin() !== array()) {
                    return '"join": test failed';
                }
                $query->setJoin('Ft', 'fTid', 'FtID');
                if ($query->getJoin() !== array('ft' => array('ftid', 'ftid'))) {
                    return '"join": test failed';
                }
                $query->setJoin('fT', 'fTid');
                if ($query->getJoin() !== array('ft' => array('ftid', 'ftid'))) {
                    return '"join": test failed';
                }
                $query->setJoin();
                if ($query->getJoin() !== array()) {
                    return '"join": test failed';
                }
                $query->setJoin('FT');
                if ($query->getJoin() !== array('ft' => array('ftid', 'ftid'))) {
                    return '"join": test failed';
                }

                // test set/get values
                if ($query->getValues() !== null) {
                    return '"default values": test failed';
                }
                $query->setColumn('tvalue');
                $query->setRow(1);
                $value = true;
                if (@$query->setValues($value) !== false) { // supposed to fail
                    return '"values type error": test failed';
                }
                $query->setType(YANA_DB_UPDATE);
                if ($query->setValues($value) !== true) {
                    return '"set values (cell)": test failed';
                }
                if ($query->getValues() !== true) {
                    return '"get values": test failed';
                }
                $query->setColumn('*');
                $value = array('tID' => 'AbC', 'tValue' => 0);
                if ($query->setValues($value) !== true) {
                    return '"set values (row)": test failed';
                }
                if ($query->getValues() !== array('tid' => 'ABC', 'tvalue' => 0)) {
                    return 'get values (row)": test failed';
                }
                if ($query->getValues() !== $value) {
                    return '"get values": test failed';
                }

                // test conversion to SELECT stmt
                $query->setType(YANA_DB_SELECT);
                if ($query->toString() != 'SELECT * FROM t, ft  WHERE t.tvalue like "1" AND profile_id = "' . $profileId . '" AND t.tid = "1" AND t.ftid = ft.ftid ORDER BY t.tid, t.tvalue') {
                    return '"toString" - select statement: test failed';
                }

                // test conversion to EXISTS stmt
                $query->setType(YANA_DB_EXISTS);
                if ($query->toString() !== 'SELECT 1 FROM t, ft  WHERE t.tvalue like "1" AND profile_id = "' . $profileId . '" AND t.tid = "1" AND t.ftid = ft.ftid') {
                    return '"toString" - exists statement: test failed';
                }

                // test conversion to LENGTH stmt
                $query->setType(YANA_DB_LENGTH);
                if ($query->toString() !== 'SELECT count(*) FROM t, ft  WHERE t.tvalue like "1" AND profile_id = "' . $profileId . '" AND t.tid = "1" AND t.ftid = ft.ftid') {                    
                    return '"toString" - length statement: test failed';
                }

                // test conversion to UPDATE stmt
                $query->setType(YANA_DB_UPDATE);
                if ($query->toString() !== 'UPDATE t SET tid = "ABC", tvalue = "0" WHERE t.tvalue like "1" AND profile_id = "' . $profileId . '" AND t.tid = "1"') {
                    return '"toString" - update statement: test failed';
                }

                // test conversion to INSERT stmt
                $query->setType(YANA_DB_INSERT);
                if ($query->toString() !== 'INSERT INTO t (tid, tvalue) VALUES ("ABC", "0")') {
                    return '"toString" - insert statement: test failed';
                }

                // test conversion to DELETE stmt
                $query->setType(YANA_DB_DELETE);
                if ($query->toString() !== 'DELETE FROM t WHERE t.tvalue like "1" AND profile_id = "' . $profileId . '" AND t.tid = "1" ORDER BY t.tid, t.tvalue') {
                    return '"toString" - delete statement: test failed';
                }

                // test set key
                if (@$query->setKey('t.1.tb.1') !== false) { // supposed to fail
                    return '"set key (failure)": test failed';
                }
                $query->setKey('T.1.Tb');
                if ($query->getTable() !== 't' || $query->getRow() !== '1' || $query->getColumn() !== 'tb') {
                    return '"set key (cell)": test failed';
                }
                $query->setKey('T.*.TB');
                if ($query->getTable() !== 't' || $query->getRow() !== '*' || $query->getColumn() !== 'tb') {
                    return '"set key (column)": test failed';
                }
                $query->setKey('T.*');
                if ($query->getTable() !== 't' || $query->getRow() !== '*' || $query->getColumn() !== '*') {
                    return '"set key (table + wildcard)": test failed';
                }
                $query->setKey('T.tb');
                if ($query->getTable() !== 't' || $query->getRow() !== 'tb' || $query->getColumn() !== '*') {
                    return '"set key (row)": test failed';
                }
                $query->setKey('T');
                if ($query->getTable() !== 't' || $query->getRow() !== '*' || $query->getColumn() !== '*') {
                    return '"set key (table)": test failed';
                }

                $query->setJoin();
                $query->setType(YANA_DB_SELECT);
                $query->setKey('i');
                if ($query->toString() !== 'SELECT * FROM i, t  WHERE t.tvalue like "1" AND i.iid = t.tid ORDER BY t.tid, t.tvalue') {
                    return '"toString" - automated table inheritance: test failed';
                }

                $query->setWhere();
                $query->setColumns(array('iid', 't.ta'));
                if ($query->toString() !== 'SELECT i.iid, t.ta FROM i, t  WHERE i.iid = t.tid ORDER BY t.tid, t.tvalue') {
                    return '"toString" - select multiple columns: test failed';
                }

                return true;

            break;
        }
    } /* end of function */
}
// Copyright  2009  Thomas Meyer  http://www.yanaframework.net
?>