<?php
/**
 * YANA library
 *
 *
 * Software:        Yana PHP-Framework 3.0.0
 * Author:          Dipl. Inf. Thomas Meyer
 * Homepage:        http://www.yanaframework.net
 * Copyright:       (c) 2009  Thomas Meyer
 * License:         GNU GPL  http://www.gnu.org/licenses/
 *
 * This program: can be redistributed and/or modified under the
 * terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see http://www.gnu.org/licenses/.
 *
 * You may use this software for free. In case you like it, or want to
 * use it for a commercial purpose, please consider making a donation.
 * http://yanaframework.net/donate.php
 *
 * This notice MAY NOT be removed.
 *
 *
 * @package  yana
 * @license  http://www.gnu.org/licenses/gpl.txt
 */


/**
 * Connection to a database server
 *
 * This class provides methods to establish or test
 * connections with database servers.
 *
 * @access      public
 * @name        DbServer
 * @package     yana
 * @subpackage  database
 */
class DbServer extends Object
{

    /**#@+
     * @ignore
     * @access  private
     */

    /** @var mixed */ var $database = null;
    /** @var array */ var $dsn      = array();
    /** @var array */ var $options  = array();

    /**#@-*/

    /**
     * create a new instance
     *
     * The $dsn parameter became available in version 2.8
     * It is an array containing the following information (all entries are optional):
     *
     * <ul>
     *     <li><pre> string  ( DBMS )     default =  mysql      name of the php-dbms api to be used
     *                                            e.g. mysql, mysqli, db2, ... </pre></li>
     *     <li><pre> string  ( HOST )     default =  localhost  adress of the host e.g. localhost,
     *                                            123.456.789.0,
     *                                            COMPUTER-NAME\DB-INSTANCE (windows+MS-SQL)</pre></li>
     *     <li><pre> integer ( PORT )     default =  0          port number </pre></li>
     *     <li><pre> string  ( USERNAME ) default =  root </pre></li>
     *     <li><pre> string  ( PASSWORD ) default =  n/a </pre></li>
     *     <li><pre> string  ( DATABASE ) default =  yana       name of the database </pre></li>
     * </ul>
     *
     * The default settings may be changed by editing file config/system.config in key DEFAULT.DATABASE.DSN.
     * The parameter $dsn has the following fall-back behaviour: $dsn -> global user settings -> yana default settings
     *
     * {@internal
     *
     * Note on PEAR-DB connection parameters
     *
     * <ul>
     * <li> phptype:    Database backend used in PHP (i.e. mysql  , odbc etc.) </li>
     * <li> dbsyntax:   Database used with regards to SQL syntax etc. When using ODBC as the phptype, set this to the DBMS type the ODBC driver is connecting to. Examples: access, db2, mssql, navision, solid, etc. </li>
     * <li> protocol:   Communication protocol to use ( i.e. tcp, unix etc.) </li>
     * <li> hostspec:   Host specification (hostname[:port]) </li>
     * <li> database:   Database to use on the DBMS server </li>
     * <li> username:   User name for login </li>
     * <li> password:   Password for login </li>
     * <li> proto_opts: Maybe used with protocol </li>
     * <li> option:     Additional connection options in URI query string format. options get separated by & </li>
     * </ul>
     *
     * Note on PEAR-DB connection options
     *
     * <ul>
     * <li> autofree:       should results be freed automatically when there are no more rows? </li>
     * <li> debug:          debug level </li>
     * <li> persistent:     should the connection be persistent? </li>
     * <li> portability:    portability mode constant. These constants are bitwised, so they can be combined using | and removed using ^. See the examples below and the "Intro - Portability" for more information. </li>
     * <li> seqname_format: the sprintf() format string used on sequence names. This format is applied to sequence names passed to createSequence(), nextID() and dropSequence(). </li>
     * <li> ssl:            use ssl to connect? </li>
     * </ul>
     *
     * }}
     *
     * @name   DbServer::__consruct()
     * @param  array  $dsn  for a description of the $dsn parameter see the text above
     */
    function DbServer($dsn = null)
    {
        /* load PEAR-DB */
        @include_once("DB.php");

        if (!is_array($dsn) && !is_null($dsn)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'Array', gettype($dsn)), E_USER_WARNING);
            return false;
        }

        global $YANA;

        /**
         * 1 retrieve default settings for fall back
         */

        /* if Yana instance available */
        if (isset($YANA)) {
            $require_odbc = $YANA->getDefault('database.require_odbc');
            $this->options = $YANA->getDefault('database.options');
            assert('is_array($require_odbc);');
            assert('is_array($this->options);');
            $system_dsn = array('DBMS'     => YANA_DATABASE_DBMS,
                                'HOST'     => YANA_DATABASE_HOST,
                                'PORT'     => YANA_DATABASE_PORT,
                                'USERNAME' => YANA_DATABASE_USER,
                                'PASSWORD' => YANA_DATABASE_PASSWORD,
                                'DATABASE' => YANA_DATABASE_NAME);
            $this->dsn = Hashtable::merge($YANA->getDefault('database.dsn'), $system_dsn);
            assert('is_array($this->dsn);');
            unset($system_dsn);

        /* if Yana instance NOT available, BUT can read settings from constant values */
        } elseif (defined('YANA_DATABASE_DBMS')) {
            $this->dsn = array('DBMS'     => YANA_DATABASE_DBMS,
                               'HOST'     => YANA_DATABASE_HOST,
                               'PORT'     => YANA_DATABASE_PORT,
                               'USERNAME' => YANA_DATABASE_USER,
                               'PASSWORD' => YANA_DATABASE_PASSWORD,
                               'DATABASE' => YANA_DATABASE_NAME,
                               'USE_ODBC' => false);
            $require_odbc = array();
            $this->options = array();

        /* if no settings can be read at all */
        } else {
            $require_odbc = array();
            $this->options = array();
        }

        /**
         * 1.2 auto-detect MySQL port for Server2Go application
         */
        if (isset($_ENV["S2G_MYSQL_PORT"]) && empty($this->dsn["PORT"]) && strpos(YANA_DATABASE_DBMS, 'mysql') !== false) {
            $this->dsn["PORT"] = $_ENV["S2G_MYSQL_PORT"];
        }

        /**
         * 1.3 there are some static options that always have to be there and can't be changed
         */
        if (defined('DB_PORTABILITY_ALL')) {
            $this->options['portability'] = DB_PORTABILITY_ALL;
        }    

        /**
         * 2 process settings provided by the user
         */
        if (is_array($dsn)) {
            $dsn = Hashtable::changeCase($dsn, CASE_UPPER);
            $this->dsn = Hashtable::merge($this->dsn, $dsn);
        }
        assert('is_array($this->dsn);');
        $dsn = array();

        /**
         * 3 create PEAR-DB compatible dsn-array
         */
        /* 3.1 determine if odbc is required to connect to this dbms */
        if (!empty($this->dsn['DBMS'])) {
            if (@$this->dsn['USE_ODBC'] == true || in_array(strtolower($this->dsn['DBMS']),$require_odbc)) {
                $dsn['phptype']  = 'ODBC';
                $dsn['dbsyntax'] = $this->dsn['DBMS'];
            } else {
                $dsn['phptype']  = $this->dsn['DBMS'];
            }
        }
        /* 3.2 now for the database host */
        if (!empty($this->dsn['HOST'])) {
            $dsn['hostspec'] = $this->dsn['HOST'];
            if (is_int($this->dsn['PORT']) && $this->dsn['PORT'] > 0) {
                $dsn['hostspec'] .= ':'.$this->dsn['PORT'];
            }
        }
        /* 3.3 database name */
        if (!empty($this->dsn['DATABASE'])) {
            $dsn['database'] = $this->dsn['DATABASE'];
        }
        /* 3.4 collect login information */
        if (!empty($this->dsn['USERNAME'])) {
            $dsn['username'] = $this->dsn['USERNAME'];
        }
        if (!empty($this->dsn['PASSWORD'])) {
            $dsn['password'] = $this->dsn['PASSWORD'];
        }

        /**
         * 4 if PEAR-DB is available try to connect using the collected settings
         */
        @include_once("DB.php");
        if (class_exists("DB")) {

            assert('is_array($dsn);');
            /**
             * {@internal
             *
             * Note:
             * Might be a bug - might be a "feature".
             * PEAR-DB does not accept foreign error handlers to be used - at least not
             * while establishing a database connection.
             * Restoring the error handler and then reseting it somehow makes things right
             * for PEAR.
             *
             * The returned object is an instance of DB_common or an instance of DB_Error.
             *
             * }}
             */
            restore_error_handler();
            $this->database = DB::connect($dsn);
            ErrorUtility::setErrorReporting(YANA_ERROR_REPORTING);

            /* error handling */
            if (DB::isError($this->database) || !is_object($this->database)) {
                $err_msg = "DATABASE NOT AVAILABLE: Unable to establish connection with database server.\n\t\tOpen administration panel and choose \"database administration\" to edit settings";
                /* Don't output passwords */                
                if (empty($dsn['password'])) {
                    $dsn['password'] = 'NO';
                } else {
                    $dsn['password'] = 'YES';
                }
                if (isset($YANA)) {
                    trigger_error($err_msg.': '.$this->database->getMessage() . "\nUsing DSN:\n" . print_r($dsn, true), E_USER_NOTICE);
                    $YANA->report( new Log($err_msg, $this->database->getMessage() . "\nUsing DSN:\n" . print_r($dsn, true) ) );
                } else {
                    trigger_error($err_msg.': '.$this->database->getMessage() . "\nUsing DSN:\n" . print_r($dsn, true) , E_USER_WARNING);
                }
            }
        } else {
            /* error handling */
            if (isset($YANA)) {
                $YANA->report( new Error('PEARDB_FAILURE') );
            } else {
                trigger_error("Unable to open connection to database using PEAR-DB. Might not be installed.", E_USER_ERROR);
            }
            $this->database = null;
        }
    }

    /**
     * get a PEAR-DB connection object
     *
     * This function returns an open database connection via PEAR-DB.
     * The returned values are:
     *
     * <ul>
     *   <li><pre> null               = if PEAR-DB was not found </pre></li>
     *   <li><pre> (boolean) false    = if the connection failed </pre></li>
     *   <li><pre> (object) DB_common = if the connection was established successfully </pre></li>
     * </ul>
     *
     * @access  public
     * @name    DbServer::get()
     * @return  mixed
     */
    function &get()
    {
        if (!class_exists("DB")) {
            $database = null;
            return $database;
        } elseif (DB::isError($this->database)) {
            trigger_error($this->database->getMessage(), E_USER_ERROR);
            $this->database = null;
            return $this->database;
        } else {
            return $this->database;
        }
    }

    /**
     * alias of DbServer::get()
     *
     * See {@link DbServer::get()} for details.
     *
     * @access  public
     * @name    DbServer::getConnection()
     * @see     DbServer::get()
     * @return  array
     * @since   2.9.8
     */
    function getConnection()
    {
        if (is_array($this->dsn)) {
            return $this->dsn;

        } else {
            return false;
        }
    }

    /**
     * get the DSN
     *
     * This function returns an associative array containing
     * information on the current connection:
     *
     * <ul>
     *   <li>USE_ODBC: true, if ODBC is used to connect to the database</li>
     *   <li>DBMS: name of used database system</li>
     *   <li>HOST: host name, e.g. localhost</li>
     *   <li>PORT: port number of database server (may be empty)</li>
     *   <li>USERNAME</li>
     *   <li>PASSWORD</li>
     *   <li>DATABASE: name of database to connect to</li>
     * </ul>
     *
     * Returns bool(false) on error.
     *
     * @access  public
     * @name    DbServer::getDsn()
     * @return  array
     * @since   2.9.8
     */
    function getDsn()
    {
        if (is_array($this->dsn)) {
            return $this->dsn;

        } else {
            return false;
        }
    }

    /**
     * test if a connection is available
     *
     * Returns bool(true) if a connection to a db-server could be established via the provided parameters,
     * and bool(false) otherwise. See the constructor method for details on the $dsn parameter
     *
     * @access  public
     * @static
     * @name    DbServer::isAvailable()
     * @param   array  $dsn
     * @return  bool
     *
     * @see     DbServer::__consruct()
     */
    function isAvailable($dsn)
    {
        if (!is_array($dsn)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'Array', gettype($dsn)), E_USER_WARNING);
            return false;
        }

        /* load PEAR-DB */
        @include_once("DB.php");
        if (!class_exists("DB")) {
            return false;
        } else {
            $db =& new DbServer($dsn);
            /*
             * NOTE: the constructor already created a log-entry if the connection failed,
             * so there's no need to do this again
             */
            if ($db->database === false || DB::isError($db->database)) {
                return false;
            } else {
                return true;
            }
        }
    }

}
// Copyright  2009  Thomas Meyer  http://www.yanaframework.net
?>