<?php
/**
 * YANA library
 *
 *
 * Software:        Yana PHP-Framework 3.0.0
 * Author:          Dipl. Inf. Thomas Meyer
 * Homepage:        http://www.yanaframework.net
 * Copyright:       (c) 2009  Thomas Meyer
 * License:         GNU GPL  http://www.gnu.org/licenses/
 *
 * This program: can be redistributed and/or modified under the
 * terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see http://www.gnu.org/licenses/.
 *
 * You may use this software for free. In case you like it, or want to
 * use it for a commercial purpose, please consider making a donation.
 * http://yanaframework.net/donate.php
 *
 * This notice MAY NOT be removed.
 *
 *
 * @package  yana
 * @license  http://www.gnu.org/licenses/gpl.txt
 */

/**
 * Database trigger handling
 *
 * {@internal
 * This class provides functions for handling of triggers and constraints.
 * Calls the function {link DbStructure::getFields()}.
 * }}
 *
 * This is a utility class that provides a collection of default
 * triggers, that you might find usefull when writing database structure
 * files.
 *
 * A "trigger" here means, a custom fragment of PHP-Code, which
 * is associated with some database event.
 * So the code gets executed every time the database event
 * is triggered.
 *
 * So a trigger gives you the chance to "hook" directly into the
 * database API and ensure that a certain code is always executed.
 *
 * @access      public
 * @package     yana
 * @subpackage  database
 * @since       2.9
 * @ignore
 */
class DbTrigger extends Object
{
    /**#@+
     * @access  private
     */

    /** @var object */ var $structure   = null;

    /**#@-*/

    /**
     * create a new instance
     *
     * This creates and initializes a new
     * instance of this class.
     *
     * @param  object  &$structure
     * @ignore
     */
    function DbTrigger(&$structure)
    {
        assert('is_object($structure) && is_a($structure, "DbStructure"); // Wrong argument type for argument 1. Instance of DbStructure expected');
        $this->structure =& $structure;

    }

    /**
     * constraint
     *
     * Evaluates a constraint and returns bool(true) on success or bool(false) on error or if the constraint fails.
     *
     * @access  public
     * @param   int     $operation  currently one of YANA_DB_SELECT, YANA_DB_UPDATE, YANA_DB_INSERT, YANA_DB_DELETE
     * @param   string  $table
     * @param   string  $field      (optional)
     * @param   mixed   $value      (optional)
     * @return  bool
     * @ignore
     */
    function constraint($operation, $table, $field = null, $value = null)
    {
        switch ($operation)
        {
            case YANA_DB_SELECT:
            case YANA_DB_UPDATE:
            case YANA_DB_INSERT:
            case YANA_DB_DELETE:
                return $this->_eval(true, $operation, $table, $field, $value);
            break;
            default:
                trigger_error("Invalid input for argument 1 in ".__CLASS__."->".__FUNCTION__."().\n\t\t" .
                "Expected one of: YANA_DB_SELECT, YANA_DB_UPDATE, YANA_DB_INSERT, YANA_DB_DELETE.", E_USER_WARNING);
                return false;
            break;
        } /* end switch */
    }

    /**
     * trigger
     *
     * Evaluates a trigger and returns bool(true) on success or bool(false) on error.
     *
     * @access  public
     * @param   int     $prefix     currently one of YANA_TRIGGER_BEFORE, YANA_TRIGGER_AFTER
     * @param   int     $operation  currently one of YANA_DB_SELECT, YANA_DB_UPDATE, YANA_DB_INSERT, YANA_DB_DELETE
     * @param   string  $table
     * @param   string  $field
     * @param   mixed   &$value
     * @return  bool
     * @ignore
     */
    function trigger($prefix, $operation, $table, $field, &$value)
    {
        /**
         * 1) argument 1
         */
        switch ($prefix)
        {
            case YANA_TRIGGER_BEFORE:
            case YANA_TRIGGER_AFTER:
            break;
            default:
                trigger_error("Invalid input for argument 1 in ".__CLASS__."->".__FUNCTION__."().\n\t\t" .
                "Expected one of: YANA_TRIGGER_BEFORE, YANA_TRIGGER_AFTER.", E_USER_WARNING);
                return false;
            break;
        } /* end switch */

        /**
         * 2) argument 2
         */
        switch ($operation)
        {
            case YANA_DB_SELECT:
            case YANA_DB_UPDATE:
            case YANA_DB_INSERT:
            case YANA_DB_DELETE:
            break;
            default:
                trigger_error("Invalid input for argument 2 in ".__CLASS__."->".__FUNCTION__."().\n\t\t" .
                "Expected one of: YANA_DB_SELECT, YANA_DB_UPDATE, YANA_DB_INSERT, YANA_DB_DELETE.", E_USER_WARNING);
                return false;
            break;
        } /* end switch */

        /**
         * 3) get result
         */
        return $this->_eval(false, $prefix | $operation, $table, $field, $value);

    }

    /**
     * evaluate triggers and constraints
     *
     * Use $fieldname parameter to separate
     * between a constraint check and triggers.
     *
     * @access  private
     * @param   bool    $isConstraint
     * @param   int     $op
     * @param   string  $table   name of table
     * @param   string  $field   name of column
     * @param   mixed   &$value  changed value
     * @return  bool
     * @ignore
     */
    function _eval($isConstraint, $op, $table, $field, &$value)
    {
        global $YANA;
        $constraintArgs = '$OPERATION,$PERMISSION,$TABLE,$FIELD,$ID,$VALUE';
        $triggerArgs    = '$OPERATION,$PERMISSION,$TABLE,$FIELD,$ID,&$VALUE';
        if (isset($YANA)) {
            $profileId  = $YANA->getVar('ID');
        } elseif (isset($_REQUEST['id'])) {
            $profileId  = $_REQUEST['id'];
        } else {
            $profileId  = '';
        }

        if ($field === '*') {
            $field = null;
        }

        if (!is_bool($isConstraint)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, "1 in ".__CLASS__."->".__FUNCTION__."()", 'Bool', gettype($isConstraint)), E_USER_WARNING);
            return false;
        } else {
            if ($isConstraint) {
                $fieldName = 'CONSTRAINT';
            } else {
                $fieldName = 'TRIGGER';
            }
        }
        if (!is_int($op)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, "2 in ".__CLASS__."->".__FUNCTION__."()", 'Integer', gettype($op)), E_USER_WARNING);
            return false;
        } else {
            /* add events here */
            if ($op & YANA_TRIGGER_BEFORE) {
                $opName = 'BEFORE_';
            } elseif ($op & YANA_TRIGGER_AFTER) {
                $opName = 'AFTER_';
            } else {
                $opName = '';
            }
            /* add operation types here */
            if ($op & YANA_DB_SELECT) {
                $opName .= 'SELECT';
            } elseif ($op & YANA_DB_INSERT) {
                $opName .= 'INSERT';
            } elseif ($op & YANA_DB_UPDATE) {
                $opName .= 'UPDATE';
            } elseif ($op & YANA_DB_DELETE) {
                $opName .= 'DELETE';
            } else {
                $opName .= 'SELECT';
            }
        }
        if (!is_string($table)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, "3 in ".__CLASS__."->".__FUNCTION__."()", 'String', gettype($table)), E_USER_WARNING);
            return false;
        } else {
            $table = strtoupper($table);
        }
        if (!is_null($field) && !is_string($field)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, "4 in ".__CLASS__."->".__FUNCTION__."()", 'String', gettype($field)), E_USER_WARNING);
            return false;
        } else {
            /* settype to STRING */
            $field = (string) $field;
            $field = strtoupper($field);
        }

        /* get list of fields
         * DbStructure->getFields($fieldName, $opName, $table, $columns = array(), $as_assoc = false)
         */
        if (empty($field)) {
            $list_of_fields = $this->structure->getFields($fieldName, $opName, $table, array(), true);
        } else {
            $list_of_fields = $this->structure->getFields($fieldName, $opName, $table, array($field), true);
        }
        assert('is_array($list_of_fields);');

        /* 1.1 prepare value */
        if (is_string($value)) {
            $value = stripslashes($value);
        } elseif (is_array($value) && (is_null($field) || empty($field))) {
            $value = Hashtable::changeCase($value, CASE_UPPER);
        }

        /* 1.2 prepare permission */
        if (isset($YANA)) {
            $arg_permission = $YANA->getVar("PERMISSION");
        } else {
            $arg_permission = 0;
        }

        /* 2 evaluate element */
        $test = true;
        foreach ($list_of_fields as $column_name => $element)
        {
            if ($element === '') {
                continue;
            } elseif (!is_scalar($element)) {
                continue;
            } else {
                if ($isConstraint) {
                    if (is_bool($element)) {
                        $function = create_function($constraintArgs, "return ".(($element)?'true':'false').";");
                    } elseif (trim($element) == '') {
                        $function = create_function($constraintArgs, 'return false;');
                    } else {
                        if (!preg_match(YANA_DB_CONSTRAINT_SYNTAX, $element)) {
                            trigger_error("ERROR: in structure file '".$this->structure->getPath()."', syntax error in constraint: '".trim($element)."'.", E_USER_WARNING);
                            return false;
                        }
                        $function = create_function($constraintArgs, "if ($element) return true; else return false;");
                    }
                } elseif (is_string($element)) {
                    $function = create_function($triggerArgs, $element);
                } else {
                    $function = create_function($triggerArgs, 'return true;');
                }
            }
            /* handling row element */
            if (empty($field) && is_string($column_name)) {
                $column_name = strtoupper($column_name);
                if (!isset($value[$column_name])) {
                    $value[$column_name] = null;
                }
                $col_value =& $value[$column_name];
                if ($op & YANA_DB_SELECT || $op & YANA_DB_DELETE) {
                    $test = $function($opName, $arg_permission, $table, $column_name, $profileId, $col_value);
                } elseif ($op & YANA_DB_INSERT || $op & YANA_DB_UPDATE) {
                    if (!is_null($col_value)) {
                        $test = $function($opName, $arg_permission, $table, $column_name, $profileId, $col_value);
                    }
                }
                /* handling database, table and field element */
            } else {
                $test = $function($opName, $arg_permission, $table, $field, $profileId, $value);
            }
            if (is_bool($test) && $test === false) {
                if (isset($YANA)) {
                    $YANA->report( new Log("DB WARNING: Operation '${opName}' stopped by '${element}' on table '${table}'".((!empty($field))?" at field '${field}'":"").((!is_null($value))?" with value '".print_r($value,true)."'":"")."."));
                    return false;
                }
            }
        } /* end foreach */

        if ($test) {
            return true;
        } else {
            return false;
        }
    }

}
// Copyright  2009  Thomas Meyer  http://www.yanaframework.net
?>