<?php
/**
 * YANA library
 *
 *
 * Software:        Yana PHP-Framework 3.0.0
 * Author:          Dipl. Inf. Thomas Meyer
 * Homepage:        http://www.yanaframework.net
 * Copyright:       (c) 2009  Thomas Meyer
 * License:         GNU GPL  http://www.gnu.org/licenses/
 *
 * This program: can be redistributed and/or modified under the
 * terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see http://www.gnu.org/licenses/.
 *
 * You may use this software for free. In case you like it, or want to
 * use it for a commercial purpose, please consider making a donation.
 * http://yanaframework.net/donate.php
 *
 * This notice MAY NOT be removed.
 *
 *
 * @package  yana
 * @license  http://www.gnu.org/licenses/gpl.txt
 */

/**
 * checked reading and writing of file sources
 *
 * This class adds functionality to write changes
 * on the object to the corresponding file.
 *
 * @access      public
 * @package     yana
 * @subpackage  file_system
 */
class File extends FileReadonly
{

    /**
     * insert new content
     *
     * @access  public
     * @param   scalar  $text
     */
    function insert($text)
    {
        if (is_scalar($text)) {
            if (!isset($this->content)) {
                $this->content = array();
            }
            assert('is_array($this->content);');
            array_push($this->content, $text);
        } else {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'Scalar value', gettype($text)), E_USER_WARNING);
        }
    }

    /**
     * write file to system
     *
     * This function will return bool(true) on success.
     * It issues an E_USER_NOTICE and returns bool(false) on error.
     *
     * @access  public
     * @return  bool
     */
    function write()
    {
        assert('empty($this->content) || is_array($this->content) || is_string($this->content); /* Content attribute has illegal value. Array or string expected. */');
        /* settype to ARRAY */
        $this->content = (array) $this->content;

        if (!$this->isWriteable()) {
            trigger_error("Unable to write to file '".$this->getPath()."'. The file is not writeable.", E_USER_NOTICE);
            return false;
        } elseif ($this->getLastModified() != filemtime($this->getPath())) {
            trigger_error("Unable to write to file '".$this->getPath()."'.\n\t\tThe file has been changed by some third party recently. Your cached copy is out of date.", E_USER_NOTICE);
            return false;
        } elseif ($this->getMd5() !== $this->getMd5($this->getPath())) {
            trigger_error("Unable to write to file '".$this->getPath()."'.\n\t\tThe file on the filesystem has been changed. Your cached copy is whether outdated or invalid.", E_USER_NOTICE);
            return false;
        } else {
            $handle = fopen($this->getPath(), "w+");
            flock($handle, LOCK_EX);
            fwrite($handle, $this->getFileContent());
            flock($handle, LOCK_UN);
            fclose($handle);
            $this->md5 = $this->getMd5($this->getPath());
            if (method_exists($this, 'resetStats')) {
                $this->resetStats();
            }
            return true;
        } /* end if */
    }

    /**
     * get size of this file
     *
     * Returns the size of the file in bytes (from cached value).
     * If an error occurs, bool(false) is returned.
     *
     * @access  public
     * @return  int
     * @since   2.8.5
     */
    function getFilesize()
    {
        if (isset($this->attributes['FILESIZE'])) {
            return $this->attributes['FILESIZE'];
        } else {
            return false;
        }
    }

    /**
     * delete this file
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * @access  public
     * @return  bool
     */
    function delete()
    {
        if (@unlink($this->getPath())) {
            return true;
        } else {
            trigger_error("Unable to delete file '".$this->getPath()."'.".((!is_writeable($this->getPath()))?"\n\t\tThe file is not writeable.":''), E_USER_NOTICE);
            return false;
        }
    }

    /**
     * failSafe writing of data
     *
     * Automatically restarts writing if the file-resource
     * is temporarily not available after waiting for 0.5 seconds.
     *
     * The process is aborted if it failed 3 times.
     *
     * @access  public
     * @return  bool
     */
    function failSafeWrite()
    {
        for ($i = 0; $i < 3; $i++)
        {
            if ($this->write()) {
              return true;
              break;
            } else {
              sleep(0.7);
            }
        }
        return false;
    }

    /**
     * create the current file if it does not exist
     *
     * @access  public
     * @return  bool
     */
    function create()
    {
        if (!$this->exists() && !empty($this->path)) {
            if (touch($this->path) !== false) {
                chmod($this->path, 0777);
                $this->resetStats();
                return true;
            } else {
                return false;
            }
        } else {
            return false;
        }
    }

    /**
     * return bool(true) if file is writeable
     *
     * A file is "writeable" if:
     * <ul>
     *  <li> the file exists </li>
     *  <li> it is accessible </li>
     *  <li> the script has permission to write data to the file </li>
     * </ul>
     *
     * @access  public
     * @return  bool
     * @since   2.9 RC3
     */
    function isWriteable()
    {
        return ( $this->exists() && is_writeable($this->getPath()) );
    }

    /**
     * get the number of contents inside the file
     *
     * @access  public
     * @return  int
     */
    function length()
    {
        if (!isset($this->content)) {
            return 0;
        } else {
            assert('is_array($this->content);');
            return count($this->content);
        }
    }

    /**
     * remove an entry from the file
     *
     * @access  public
     * @param   string  $key  (optional)
     * @return  bool
     */
    function remove($key = "")
    {
        /* settype to STRING */
        $key = (string) $key;

        if ($key == "" || $key == "*") {
            $this->content = array();
            return true;
        } else {
            if (is_numeric($key)) {
                if (!isset($this->content[$key])) {
                    return false;
                } else {
                    array_splice($this->content, $key, 1);
                    return true;
                }
            } else {
                return Hashtable::remove($this->content, $key);
            }
        } /* end if */

    }

    /**
     * copy the file to some destination
     *
     * This will create a copy of this file on the filesystem.
     * Bool(true) will be returned on success and bool(false)
     * on error.
     *
     * Possible errors are:
     * <ul>
     *   <li>  the destination directory does not exist (and $isRecursive is not provided, or set to false)  </li>
     *   <li>  another file with the same name does already exist (and $overwrite is set to false)  </li>
     *   <li>  the destination file is not accessible for some other reason  </li>
     * </ul>
     *
     * If the directory the desination file would be place in
     * does not exist, and $isRecursive is set to true,
     * it will automatically, recursively create the missing directories.
     * Newly created directories will be set to access restriction 0766.
     * Note that this is unlike the default behaviour, where there is no restriction at all.
     *
     * Note: instead of a file name in parameter $destFile, you may also provide the name of a directory.
     * In this case the file the current name of the file will be used.
     * The destination directory needs to be terminated by a forward-slash '/'.
     *
     * The $destFile parameter must not be any longer than 512 characters.
     * It must not contain any but alphanumeric characters.
     *
     * If you copy a PHP file, note that for security reasons the $mode parameter defaults
     * to 0766, which means the file will not be executable in a UNIX environment (while it may
     * be executable on windows platforms) until you set the $mode parameter to 0777.
     *
     * @access   public
     * @param    string   $destFile     destination to copy the file to
     * @param    bool     $overwrite    setting this to false will prevent existing files from getting overwritten
     * @param    bool     $isRecursive  setting this to true will automatically, recursively create directories in the $destFile string, if required
     * @param    int      $mode         the access restriction that applies to the copied file, defaults to 0766
     * @return   bool
     *
     * @since    2.8.5
     */
    function copy($destFile, $overwrite = true, $isRecursive = false, $mode = 766)
    {
        if (!is_string($destFile)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($destFile)), E_USER_WARNING);
            return false;
        }
        if (!is_bool($overwrite)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 2, 'Boolean', gettype($overwrite)), E_USER_WARNING);
            return false;
        }
        if (!is_bool($isRecursive)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 4, 'Boolean', gettype($isRecursive)), E_USER_WARNING);
            return false;
        }
        if (!is_int($mode)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 5, 'Integer', gettype($mode)), E_USER_WARNING);
            return false;
        }

        /* validity checking */
        if (empty($destFile) || strlen($destFile) > 512 || !preg_match('/^[\w\d-_\.][\w\d-_\/\.]*$/',$destFile)) {
            trigger_error("Invalid filename '".$destFile."' in ".__CLASS__."->".__FUNCTION__."().", E_USER_WARNING);
            return false;
        }
        if (!preg_match('/^[1-7]{1}[0-7]{2}$/',"$mode")) {
            trigger_error("Invalid mode (access level) '".$mode."' in ".__CLASS__."->".__FUNCTION__."().\n\t\tThis argument needs to be an integer of 3 or 4 digits between 0-7.\n\t\tNote: when providing 4 digits, the first digit needs to be 0.", E_USER_WARNING);
            return false;
        }

        $destDir  = dirname($destFile) . '/';
        $destFile = preg_replace('/^.*\//s', '', $destFile);
        if (empty($destFile)) {
            $destFile = $this->getPath();
            $destFile = preg_replace('/^.*\//s', '', $destFile);
        }

        /* check if file already exists */
        if ($overwrite === false && file_exists($destDir . $destFile) === true) {
            trigger_error("Unable to copy to file '${destDir}${destFile}'. Another file with the same name does already exist.", E_USER_NOTICE);
            return false;
        } elseif ($overwrite === true && file_exists($destDir . $destFile) === true && is_writeable($destDir . $destFile) === false) {
            trigger_error("Unable to copy to file '${destDir}${destFile}'. The file does already exist and is not writeable.", E_USER_WARNING);
            return false;
        }

        assert('is_string($destDir); // Unexpected result: $destDir. String expected.');
        assert('is_string($destFile); // Unexpected result: $destFile. String expected.');

        /* recursively create directories */
        if (!empty($destDir) && !is_dir($destDir)) {
            if ($isRecursive === true) {
                assert('!isset($currentDir); // cannot redeclare variable $currentDir');
                $currentDir = '';
                assert('!isset($current); // cannot redeclare variable $dir');
                assert('!isset($dir); // cannot redeclare variable $dir');
                foreach (explode('/',$destDir) as $dir)
                {
                    if (!is_dir($currentDir . $dir)) {
                        $current = new Dir($currentDir . $dir);
                        if ($current->create($mode) === false) {
                            trigger_error("Unable to create directory '".$current->getPath()."'.", E_USER_NOTICE);
                            return false;
                        }
                    }
                    $currentDir .= $dir.'/';
                } /* end foreach */
                unset($dir,$current);
            } else {
                trigger_error("Unable to copy file '${destFile}'. The directory '${destDir}' does not exist.", E_USER_NOTICE);
                return false;
            } /* end if */
        }

        if (copy($this->getPath(), $destDir . $destFile)) {
            if (chmod($destDir . $destFile, decoct($mode)) === false) {
                trigger_error("Unable to set mode (access level) for file '${destDir}${destFile}'.", E_USER_NOTICE);
            }
            return true;
        } else {
            return false;
        }
    }

}
// Copyright  2009  Thomas Meyer  http://www.yanaframework.net
?>