<?php
/**
 * YANA library
 *
 *
 * Software:        Yana PHP-Framework 3.0.0
 * Author:          Dipl. Inf. Thomas Meyer
 * Homepage:        http://www.yanaframework.net
 * Copyright:       (c) 2009  Thomas Meyer
 * License:         GNU GPL  http://www.gnu.org/licenses/
 *
 * This program: can be redistributed and/or modified under the
 * terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see http://www.gnu.org/licenses/.
 *
 * You may use this software for free. In case you like it, or want to
 * use it for a commercial purpose, please consider making a donation.
 * http://yanaframework.net/donate.php
 *
 * This notice MAY NOT be removed.
 *
 *
 * @package  yana
 * @license  http://www.gnu.org/licenses/gpl.txt
 */

/**
 * Singleton Language
 *
 * This class may be used to dynamically load additional
 * language files at runtime.
 *
 * @access      public
 * @package     yana
 * @subpackage  core
 */
class Language extends Singleton
{

    /**#@+
     * @ignore
     * @access  private
     */

    /** @var String */ var $directory = "";
    /** @var Dir    */ var $dirStream = null;
    /** @var Dir    */ var $defaultDir = null;
    /** @var array  */ var $file_loaded = array();

    /**#@-*/

    /**
     * Singleton Constructor
     *
     * @param  string  $directory
     * @param  Dir     $dirStream
     * @param  string  $defaultDir
     *
     * @ignore
     */
    function Language($directory, $dirStream, $defaultDir)
    {
        $this->checkSingletonConstraint();
        assert('is_string($directory) && is_dir($directory);');
        assert('is_string($defaultDir) && is_dir($defaultDir);');
        /* settype to STRING */
        $directory  = (string) $directory;
        $defaultDir = (string) $defaultDir;

        $this->directory = $directory;
        $this->dirStream =& $dirStream;
        $this->dirStream->read();
        $this->defaultDir = $defaultDir;
    }
    
    /**
     * get name of selected language
     *
     * Returns the name of the currently selected
     * language as a string, or bool(false) on error.
     *
     * Example:
     * Returns 'en' for English, 'de' for German.
     *
     * Technically spoken, this is the name of the
     * sub-directory, where the current language's
     * files are stored. Check the directory
     * "languages/" for a complete list.
     *
     * @access  public
     * @return  string|bool(false)
     * @since   2.9.6
     */
    function getSelectedLanguage()
    {
        if (!empty($this->directory) && is_object($this->dirStream) && is_dir($this->directory)) {
            $language = substr($this->directory, strlen($this->dirStream->getPath()));
            $language = preg_replace('/\W/', '', $language);
            return $language;
        } else {
            return false;
        }
    }

    /**
     * read language strings from a file
     *
     * The file should be taken from 'languages/*.config',
     * without path and file extension.
     *
     * The file contents will automatically be stored in the global
     * registry and can be accessed using $YANA->getVar('LANGUAGE.SOME_VALUE').
     *
     * This function issues an E_USER_NOTICE if the file does not exist.
     * It returns bool(true) on success and bool(false) on error.
     *
     * @access  public
     * @param   string  $filename
     * @return  bool
     */
    function read($filename)
    {
        assert('is_string($filename);');
        /* settype to STRING */
        $filename = (string) $filename;
        global $YANA;

        $fileId = strtoupper($filename);

        if (!preg_match("/^[\w_-\d]+$/i", $filename)) {
            trigger_error("The provided language-file id '".$filename."' contains illegal characters. Be aware that only alphanumeric (a-z,0-9,-,_) characters are allowed.", E_USER_NOTICE);
            return false;
        }
        if (isset($this->file_loaded[$fileId]) && $this->file_loaded[$fileId] === true) {
            return true;
        }

        /* preload defaults */
        $defaultFile = $this->defaultDir.$filename.".config";
        if (file_exists($defaultFile)) {
            $defaultLanguage = SML::getFile($defaultFile, CASE_UPPER);
            $this->file_loaded[$fileId] = true;
        } else {
            $defaultLanguage = array();
            $this->file_loaded[$fileId] = false;
        }
        assert('is_array($defaultLanguage);');

        /* override defaults where available */
        $selectedFile = $this->directory.$filename.".config";
        if ($selectedFile !== $defaultFile) {
            if (file_exists($selectedFile)){
                $selectedLanguage = SML::getFile($selectedFile, CASE_UPPER);
                $this->file_loaded[$fileId] = true;
                assert('is_array($selectedLanguage);'); ;
                assert('is_array($defaultLanguage);');
            } else {
                /* No specific language file found, falling back to defaults. */
                trigger_error("No language-file '".$selectedFile."' found for id '".$filename."'. Falling back to default language.", E_USER_NOTICE);
            }
        } else {
            $selectedLanguage = array();
        }

        $LANGUAGE = Hashtable::merge($defaultLanguage, $selectedLanguage);
        unset($selectedLanguage,$defaultLanguage);
        assert('is_array($LANGUAGE); // Unexpected result type. Array expected.');

        if (!empty($LANGUAGE)) {
            if (isset($YANA)) {
                $YANA->merge("*", $LANGUAGE);
            }
            return true;
        } else {
            return false;
        }
    }

    /**
     * getContent
     *
     * @access  public
     * @return  array  numeric array of strings
     *
     * @ignore
     */
    function getContent()
    {
        $content = $this->dirStream->getContent();
        assert('is_array($content);');
        if (is_array($content)) {
            return $content;
        } else {
            return array();
        }
    }

    /**
     * setDirectory
     *
     * @access  public
     * @return  bool
     *
     * @ignore
     */
    function setDirectory($directory)
    {
        if (is_string($directory) && preg_match('/^[\w\d]+$/s', $directory) && is_dir($this->dirStream->getPath().$directory)) {
            $this->directory = $this->dirStream->getPath().$directory.'/';
            $this->read('default');
            return true;
        } else {
            trigger_error("Invalid language directory '".print_r($directory,true)."'.", E_USER_NOTICE);
            return false;
        }
    }

    /**
     * getContent
     *
     * @access  public
     * @return  string
     *
     * @ignore
     */
    function getDirectory()
    {
        assert('is_string($this->directory);');
        return $this->directory;
    }

    /**
     * getInfo
     *
     * @access  public
     * @param   string  $filename  relative path to config file
     * @return  array
     *
     * @ignore
     */
    function getInfo($filename)
    {
        assert('is_string($filename);');
        /* settype to STRING */
        $filename = (string) $filename;

        $file = ($this->dirStream->getPath()).$filename;
        if (is_file($file)) {
            $REF = SML::getFile($file, CASE_UPPER);
            assert('is_array($REF["LANGUAGE_INFO"]);');
            if (is_array($REF['LANGUAGE_INFO'])) {
                $REF['LANGUAGE_INFO']['LAST_CHANGE'] = filemtime($file);
                clearstatcache();
                return $REF["LANGUAGE_INFO"];
            } else {
                return false;
            }
        } else {
            return false;
        }
    }

}
// Copyright  2009  Thomas Meyer  http://www.yanaframework.net
?>