<?php
/**
 * YANA library
 *
 *
 * Software:        Yana PHP-Framework 3.0.0
 * Author:          Dipl. Inf. Thomas Meyer
 * Homepage:        http://www.yanaframework.net
 * Copyright:       (c) 2009  Thomas Meyer
 * License:         GNU GPL  http://www.gnu.org/licenses/
 *
 * This program: can be redistributed and/or modified under the
 * terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see http://www.gnu.org/licenses/.
 *
 * You may use this software for free. In case you like it, or want to
 * use it for a commercial purpose, please consider making a donation.
 * http://yanaframework.net/donate.php
 *
 * This notice MAY NOT be removed.
 *
 *
 * @package  yana
 * @license  http://www.gnu.org/licenses/gpl.txt
 */

/**
 * entity RSS feed
 *
 * This class represents a RSS feed.
 * It offers functionality to publish new feeds.
 *
 * Examples of usage:
 * <ol>
 *  <li> Creating and printing a RSS feed inside a plugin:
 *       <code>
 *  function rss_foobar($args)
 *  {
 *       // create rss feed
 *       $rss = new RSS();
 *       // provide some description for it
 *       $rss->title = 'foo';
 *       $rss->description = 'this is my RSS feed';
 *       // create 1st item
 *       $item = new RSSitem();
 *       $item->title = '1st entry';
 *       $item->link = 'http://some.url';
 *       $item->description = 'some text';
 *       $rss->addItem($item);
 *       // create 2nd item
 *       $item = new RSSitem();
 *       $item->title = '2nd entry';
 *       $item->link = 'http://some.other.url';
 *       $item->description = 'some text';
 *       $rss->addItem($item);
 *       // print rss feed
 *       print utf8_encode($rss->toString());
 *       // when you're done, terminate the program
 *       exit(0);
 *  }
 *  </code>
 *  </li>
 *  <li> Showing the RSS logo + link from inside a plugin:
 *       <code>$YANA->setVar('RSS_ACTION', 'rss_foobar')</code>
 *       Where 'rss_foobar' is the value for the "action" parameter of the
 *       URL. You don't need to call it 'rss_foobar' - just call it whatever
 *       you like.
 *       You are to implement the code for action 'rss_foobar' yourself.
 *       To find out how to do that, see the example above.
 *  </li>
 * </ol>
 *
 * @access      public
 * @package     yana
 * @subpackage  utilities
 */
class RSS extends Object
{
    /**#@+
     * this field is mandatory
     *
     * @access  public
     */

    /**
     * name of the channel, e.g. title of the website the channel refers to
     * @var string */ var $title       = "";
    /**
     * a link that points to the originating website of the feed
     * @var string */ var $link        = "";
    /**
     * a text to describe context and purpose of this channel
     * @var string */ var $description = "";

    /**#@-*/
    /**#@+
     * this field is optional
     *
     * @access  public
     */

    /**
     * language the channel is written in
     * @var string */ var $language       = "";
    /**
     * copyright notice of your choice
     * @var string */ var $copyright      = "";
    /**
     * mail of person responsible for editorial content
     * @var string */ var $managingEditor = "";
    /** 
     * mail of person responsible for technical issues
     * @var string */ var $webMaster      = "";
    /**
     * "time to live" - a number of minutes that indicates how long a channel can be cached before refreshing from the source
     * @var int    */ var $ttl            = 0;
    /**
     * *.jpg, *.gif, or *.png image to display with the feed
     * @var string */ var $image          = "";
    /**
     * one or more categories that the channel belongs to
     * @var array  */ var $category       = array();
    /**
     * a CSS stylesheet that formats the channel
     * @var string */ var $css            = "";
    /**
     * a XSL transformation that formats the channel
     * @var string */ var $xslt           = "";
    /**
     * list of one or more items of the channel
     * @var array  */ var $item          = array();

    /**#@-*/
    /**#@+
     * @ignore
     * @access  private
     */

    /**
     * date when the channel was published (auto-generated)
     * @var string */ var $pubDate       = "";
    /**
     * pubDate of most recent item (auto-generated)
     * @var string */ var $lastBuildDate = "";
    /**
     * a string to identify the class / program that created the file
     * @var string */ var $generator     = "";

    /**#@-*/
    
    /**
     * constructor
     *
     * creates and initializes a new instance of this class
     */
    function RSS()
    {
        global $YANA;
        $this->title     = '%LANGUAGE.PROGRAM_TITLE%';
        $this->generator = "Yana Framework ".YANA_VERSION;
        $this->item      = array();
        if (isset($_SESSION['language'])) {
            $this->language = $_SESSION['language'];
        } elseif (isset($YANA)) {
            $this->language = str_replace('/','',$YANA->getVar("LANGUAGE_INFO.DIRECTORY"));
        }
        if (isset($YANA)) {
            $this->link = $YANA->getVar("SETTINGS.WEBSITE_URL");
        }
    }

    /**
     * get string value
     *
     * This returns this object as a string in RSS 2.0 syntax.
     *
     * Returns bool(false) and issues an E_USER_NOTICE on error.
     *
     * @access  public
     * @return  bool(false)|string
     *
     * @name    RSS::toString()
     */
    function toString()
    {
        global $YANA;

        /* check mandatory fields */
        if (!is_string($this->title) || empty($this->title)) {
            trigger_error("Unable to create RSS. Mandatory field 'title' missing.", E_USER_NOTICE);
            return false;
        }
        if (!is_string($this->link) || empty($this->link)) {
            trigger_error("Unable to create RSS. Mandatory field 'link' missing.", E_USER_NOTICE);
            return false;
        }
        if (!is_string($this->description) || empty($this->description)) {
            trigger_error("Unable to create RSS. Mandatory field 'description' missing.", E_USER_NOTICE);
            return false;
        }

        /* check optional fields */
        if (!is_string($this->language) || !preg_match('/^(?:|\w{2}(?:-\w{2})?)$/s',$this->language)) {
            trigger_error("Unable to create RSS. Invalid 'language' value.\n\t\tUse following syntax: 'xx-xx', e.g. 'en', 'en-us', 'de', 'de-de', 'de-at' aso.", E_USER_NOTICE);
            return false;
        }
        if (!is_string($this->copyright)) {
            trigger_error("Unable to create RSS. Invalid 'copyright' value.", E_USER_NOTICE);
            return false;
        }
        if (!is_string($this->image)) {
            trigger_error("Unable to create RSS. Invalid 'image' value.", E_USER_NOTICE);
            return false;
        }
        if (!is_int($this->ttl)) {
            trigger_error("Unable to create RSS. Invalid 'ttl' value.", E_USER_NOTICE);
            return false;
        }
        if (!is_string($this->managingEditor)) {
            trigger_error("Unable to create RSS item. Field 'managingEditor' is invalid.", E_USER_NOTICE);
            return array();
        } else {
            $tmp = $this->managingEditor;
            $this->managingEditor = untaintInput($this->managingEditor, 'mail');
            if ($tmp != $this->managingEditor) {
                trigger_error("RSS item has been auto-converted.\n\t\tThe input value '".$tmp."' did not seem to be valid.\n\t\tThe value has been changed to '".$this->managingEditor."'.", E_USER_NOTICE);
            }
        }
        if (!is_string($this->webMaster)) {
            trigger_error("Unable to create RSS item. Field 'webMaster' is invalid.", E_USER_NOTICE);
            return array();
        } else {
            $tmp = $this->webMaster;
            $this->webMaster = untaintInput($this->webMaster, 'mail');
            if ($tmp != $this->webMaster) {
                trigger_error("RSS item has been auto-converted.\n\t\tThe input value '".$tmp."' did not seem to be valid.\n\t\tThe value has been changed to '".$this->webMaster."'.", E_USER_NOTICE);
            }
        }
        if (!is_string($this->css)) {
            trigger_error("Unable to create RSS. Invalid 'css' value.", E_USER_NOTICE);
            return false;
        }
        if (!is_string($this->xslt)) {
            trigger_error("Unable to create RSS. Invalid 'xslt' value.", E_USER_NOTICE);
            return false;
        }
        if (!is_array($this->item)) {
            trigger_error("Unable to create RSS. Invalid 'item' value.", E_USER_NOTICE);
            return false;
        }

        /* auto-fill fields, that can be generated */
        $this->pubDate = date('r',time());
        if (count($this->item) > 0) {
            $time = 0;
            foreach ($this->item as $item)
            {
                if (!empty($item['pubDate'])) {
                    $tmp = strtotime($item['pubDate']);
                    if (is_int($tmp) && $tmp > $time) {
                        $time = $tmp;
                    }
                }
            } /* end foreach */
            if ($time > 0) {
                $this->lastBuildDate = date('r',$time);
            } else {
                $this->lastBuildDate = "";
            }
        }

        /* publish data */
        if (isset($YANA)) {
            $src = $YANA->skin->getFile('RSS_TEMPLATE');
            assert('is_file($src);');
            $tpl = new SmartTemplate($src);
            $tpl->insert('rss', get_object_vars($this));
            return $tpl->get();
        } else {
            return XMLencode(get_object_vars($this), 'rss');
        }
    }

    /**
     * add RSS feed item to this channel
     *
     * The $item has to be a valid RSSitem object.
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * Note: this function does not check for duplicate guid's
     *
     * @access  public
     * @param   RSSitem  $item
     * @return  bool
     *
     * @name    RSS::toString()
     */
    function addItem($item)
    {
        if (!is_object($item) || (is_subclass_of($item,'RSSitem') === false && strcasecmp(get_class($item),'RSSitem') !== 0)) {
            trigger_error("Wrong argument type for argument 1. Instance of RSSitem expected, found '".get_class($item)."' instead.", E_USER_WARNING);
            return false;
        } else {
            if (!is_array($this->item)) {
                $this->item = array();
            }
            $item = $item->get();
            if (!is_array($item) || count($item) === 0) {
                return false;
            } else {
                array_push($this->item, $item);
                assert('is_array($this->item);');
                return true;
            }
        }
    }

    /**
     * publish a rss feed
     *
     * Adds the action identified by $action to the list of
     * rss-feeds to be offered to the user.
     *
     * The argument $name is optional and allows you to define
     * a custom name for your feed. If none is provided, the
     *
     *
     * Example of usage:
     *
     * Add this to your plugin code
     * <code>Microsummary::publish('create_rss');</code>
     * Where 'create_rss' is the name of the plugin function
     * that displays the rss feed.
     *
     * Returns bool(false) on error.
     *
     * @access  public
     * @static
     * @param   string  $action
     * @param   string  $name
     * @return  bool
     */
    function publish($action, $name = null)
    {
        global $YANA;
        if (!isset($YANA)) {
            return false;
        } elseif (!is_string($action)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($action)), E_USER_WARNING);
            return false;
        } elseif (!is_null($name) && !is_string($name)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 2, 'String', gettype($name)), E_USER_WARNING);
            return false;
        } else {
            $feeds = $YANA->getVar('RSS_ACTION');

            if (empty($feeds)) {
                $feeds = array();
            } else {
                settype($feeds, 'array');
            }

            if (is_null($name)) {
                $feeds[] = $action;
            } else {
                $feeds[$name] = $action;
            }

            $YANA->setVar('RSS_ACTION', $feeds);
            return true;
        }
    }
}
// Copyright  2009  Thomas Meyer  http://www.yanaframework.net
?>