<?php
/**
 * YANA library
 *
 *
 * Software:        Yana PHP-Framework 3.0.0
 * Author:          Dipl. Inf. Thomas Meyer
 * Homepage:        http://www.yanaframework.net
 * Copyright:       (c) 2009  Thomas Meyer
 * License:         GNU GPL  http://www.gnu.org/licenses/
 *
 * This program: can be redistributed and/or modified under the
 * terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see http://www.gnu.org/licenses/.
 *
 * You may use this software for free. In case you like it, or want to
 * use it for a commercial purpose, please consider making a donation.
 * http://yanaframework.net/donate.php
 *
 * This notice MAY NOT be removed.
 *
 *
 * @package  yana
 * @license  http://www.gnu.org/licenses/gpl.txt
 *
 * @ignore
 */

/**
 * utility SmartUtility
 *
 * This is a utility class. It encapsulates extensions to use with
 * the smarty temlate engine.
 *
 * This is also a global namespace for layout specific functions.
 * These functions implement recursive replacement of tokens.
 * This functionality is used for registry-files (sml/config)
 * of the framework.
 *
 * {@internal
 *
 * Additional smarty functions and modifiers are documented elsewhere.
 * These functions should be ignored in API-documentation.
 *
 * }}
 *
 * @static
 * @access      public
 * @package     yana
 * @subpackage  core
 * @ignore
 */
class SmartUtility extends Utility
{

    /**
     * replace each token within a text/template
     *
     * NOTE: this method is case-insensitive
     *
     * This means if your reference array contains
     * keys with the same name but different writing,
     * e.g. 'a' and 'A', the keys that last of the both
     * keys is used and the other is ignored.
     *
     * Also note that token that refer to a non-existing value
     * in the reference array are left alone and will not be replaced.
     * This is you can call this function multiple times with different
     * reference arrays to easily implement a simple fallback behaviour.
     *
     * @access  public
     * @param   string  $template  the text containing the token to be replaced
     * @param   array   &$array    a reference array containing the values
     * @return  string
     * @static
     */
    function fillTemplate($template, &$array)
    {
        assert('is_string($template); // Wrong argument type argument 1. String expected');
        assert('is_array($array); // Wrong argument type argument 1. String expected');

        if (is_string($template) && is_array($array) && count($array) > 0) {

            SmartUtility::_replace($template, $array);
            assert('is_string($template); // Unexpected result: $template is supposed to be a string.');
            return $template;

        } else {
            return $template;
        }
    }

    /**
     * replace each token within a text/template
     *
     * NOTE: this method is case-insensitive
     *
     * @access  private
     * @param   string   $template
     * @param   array    &$array
     * @param   $array   $reference  DO NOT SET
     * @return  string
     * @static
     */
    function _replace(&$template, &$array, $reference = null)
    {
        if (!is_string($template) || !is_array($array)) {
            return $template;
        }

        $ldim = YANA_UNI_DELIMITER_REGEXP;
        $rdim = YANA_UNI_DELIMITER_REGEXP;
        if (!is_array($reference)) {
            $reference = array_change_key_case($array, CASE_UPPER);
        }

        $match = array();
        if (preg_match_all("/$ldim([\w_\.][\w_\/\.]+)$rdim/s", $template, $match) > 0) {
            $match = $match[1];
            $match = array_unique($match);
            for ($i = 0; $i < count($match); $i++)
            {
                assert('is_array($reference); /* unexpected result: $reference should be an array */');
                $tmp = Hashtable::get($reference, strtoupper(current($match)));
                /* if $tmp is NULL, the reference $match is pointing to a non-existing value */
                if (is_null($tmp)) {
                    $tmp = current($match);
                    $test = false;
                } elseif (!is_scalar($tmp)) {
                    trigger_error("The token '".current($match)."' refers to a non-scalar value inside a string contetx. It's value will be converted to the string 'true'.", E_USER_NOTICE);
                    $tmp = "true";
                    $test = true;
                } else {
                    /* settype to STRING */
                    $tmp = (string) $tmp;
                    $test = true;
                }
                /**
                 * if the content string we got from the reference array contains token as well,
                 * we recursivle replace them.
                 */
                if (strpos($tmp, $ldim) !== false) {
                    assert('is_string($tmp); // Unexpected result: $tmp is supposed to be a string');
                    SmartUtility::_replace($tmp, $array, $reference);
                }
                if ($test === true) {
                    assert('is_string($tmp); // Unexpected result: $tmp is supposed to be a string');
                    $currentMatch = preg_quote(current($match), '/');
                    $template = preg_replace("/(<textarea [^>]*>[^<]*)${ldim}${currentMatch}${rdim}([^<]*<\/textarea>)/Usi", '${1}'.$tmp.'${2}', $template);
                    $template = preg_replace("/(<[^\!^>]+)${ldim}${currentMatch}${rdim}([^>]+>)/Usi", '${1}' . String::htmlSpecialChars($tmp, ENT_QUOTES, 'ISO-8859-1', false) . '${2}', $template);
                    $template = preg_replace("/${ldim}${currentMatch}${rdim}/i", SmartUtility::embeddedTags($tmp), $template);
                    $template = preg_replace("/(<textarea [^>]*>[^<]*)\[br\]([^<]*<\/textarea>)/Usi", '${1}' . "\n" . '${2}', $template);
                } /* end if */
                unset($currentMatch);
                next($match);
            } /* end for */
        } /* end if */
        assert('is_string($template); // Unexpected result: $template is supposed to be a string');
    }

    /**
     * toTimestamp
     *
     * converts input created by the smarty functions
     * "html_select_time" and "html_select_date" to a UNIX-Timestamp.
     *
     * Example:
     * <code>
     * // some input as provided by Smarty
     * $input = array(
     *     'FOO_hour' => 12,
     *     'FOO_minute' => 0,
     *     'FOO_second' => 0,
     *     'FOO_month' => 1,
     *     'FOO_day' => 1,
     *     'FOO_year' => 2000
     * );
     * // convert to unix timestamp
     * $date = SmartUtility::toTimestamp('FOO_', $input);
     * // now you may do something usefull with it ...
     * </code>
     *
     * @access  public
     * @static
     * @param   string  $prefix
     * @param   array   $ARGS
     * @return  int
     */
    function toTimestamp($prefix, $ARGS)
    {
        $prefix = (string) $prefix;
        return mktime((int) @$ARGS[$prefix.'hour'], (int) @$ARGS[$prefix.'minute'], (int) @$ARGS[$prefix.'second'], (int) @$ARGS[$prefix.'month'], (int) @$ARGS[$prefix.'day'], (int) @$ARGS[$prefix.'year']);
    }

    /**
     * smarty processor htmlPostProcessor
     *
     * @access  public
     * @static
     * @param   array   $params
     * @param   Smarty  &$smarty
     * @return  string
     * @ignore
     */
    function htmlPostProcessor($source, &$smarty)
    {
        if (!is_string($source)) {
            return $source;
        }        
        global $YANA;

        if (isset($YANA)) {
            $permission = $YANA->getVar('PERMISSION');
        } else {
            $permission = 0;
        }

        /* Create field for internal form validation */
        if (!is_int($permission) || $permission < 1) {
            $source = str_replace("</form>", "<span class=\"yana_button\"><input type=\"text\" name=\"yana_url\"/></span>\n</form>", $source);
        }

        /* Instance Management */
        if (isset($YANA) && $YANA->getVar("LICENCE.KEY")){
            $a = $YANA->getVar("LICENCE");
            $note = urldecode($GLOBALS['lcn_note']);
            $note = SmartUtility::fillTemplate($note, $a);
            $source = preg_replace("/<\/body>/i", $note."</body>", $source);
        } else {
            $id = base64_decode($GLOBALS['instance_id']);
            $source = preg_replace("/<\/body>/i", $id."</body>", $source);
        }

        return $source;
    }

    /**
     * smarty processor processLanguageInstructions
     *
     * @access  public
     * @static
     * @param   array   $params
     * @param   Smarty  &$smarty
     * @return  string
     * @ignore
     */
    function processLanguageInstructions($source, &$smarty)
    {
        if (!is_string($source)) {
            return $source;
        } else {
            $source = preg_replace("/".SmartUtility::_regExpToken("(\\\$LANGUAGE\..*)")."/Us", YANA_LEFT_DELIMITER."\\1|embeddedTags".YANA_RIGHT_DELIMITER, $source);
            return $source;
        }
    }

    /**
     * smarty processor htmlPreProcessor
     *
     * @access  public
     * @static
     * @param   string  $source
     * @param   Smarty  &$templateClass
     * @return  string
     * @ignore
     */
    function htmlPreProcessor($source, &$templateClass)
    {
        if (!is_string($source)) {
            return $source;
        }
        global $YANA;

        /* funcionality of 'include' function has been moved to 'import', so ignore the original */
        $source = preg_replace("/".SmartUtility::_regExpToken("include .*")."/Usi", "", $source);
        /* deactivate 'include_php' for security reasons */
        $source = preg_replace("/".SmartUtility::_regExpToken("include_php .*")."/Usi", "", $source);
        /* deactivate 'php' for security reasons */
        $source = preg_replace("/".SmartUtility::_regExpToken("php").".*".SmartUtility::_regExpToken("\/php")."/Usi", "", $source);
        /* reformat abbreviated %FOO% tags to [%$FOO%] */
        $source = preg_replace("/".toRegExp(YANA_UNI_DELIMITER)."(([A-Z]|\d|_|-|\.)+)".toRegExp(YANA_UNI_DELIMITER)."/", SmartUtility::_varToken("\\1"), $source);

        $basedir = $templateClass->get_template_vars('BASEDIR');
        /* settype to STRING */
        $basedir = (string) $basedir;
        $is_include = $templateClass->get_template_vars('FILE_IS_INCLUDE');
        if ($is_include) {
            $source = preg_replace("/^.*<body(.*)>(.*)<\/body>.*$/Usi", "<div\\1>\\2</div>", $source);
        }
            /* Create initialization for JavaScript/Ajax bridge */
            if (isset($YANA) && strpos($source, '<head') > -1) {
                assert('!isset($script); // Cannot redeclare var $script');
                $script = "\n        " . '<script type="text/javascript" language="javascript"><!--' . "\n" .
                '        var src="";' . "\n" .
                '        var php_self="[%$PHP_SELF%]";' . "\n" .
                "        //--></script>";
                $source = preg_replace('/<head[^>]*>/i', '${0}' . $script, $source);
                unset($script);
            }

        if (!empty($basedir)) {
            preg_match_all('/('.SmartUtility::_regExpToken(')import\s+(?:preparser(?:="true")?\s+|)file="(\S*)(".*').')/Ui', $source, $match2);
            for ($i = 0; $i < count($match2[0]); $i++)
            {
                if (preg_match('/\sliteral\s/i' ,$match2[0][$i])) {
                    $source = preg_replace('/'.toRegExp($match2[0][$i]).'/i', preg_replace("/\sliteral\s/i", " ", $match2[0][$i]), $source);
                } elseif (preg_match('/\spreparser\s/i' ,$match2[0][$i])) {
                    $source = preg_replace('/'.toRegExp($match2[0][$i]).'/i', preg_replace("/.*<body[^>]*>(.*)<\/body>.*/si", "\\1", implode("", file($basedir.$match2[2][$i]))), $source);
                } else {
                    $source = str_replace($match2[0][$i], $match2[1][$i].'import file="'.$basedir.$match2[2][$i].$match2[3][$i], $source);
                }
            }

            preg_match_all('/('.SmartUtility::_regExpToken(')insert\s+file="(\S*)(".*').')/Ui', $source, $match2);
            for ($i = 0; $i < count($match2[0]); $i++)
            {
                    $source = preg_replace('/'.toRegExp($match2[0][$i]).'/i', $match2[1][$i].'insert file="'.$basedir.$match2[2][$i].$match2[3][$i], $source);
            }

            preg_match_all('/ background\s*=\s*"(\S*)"/i', $source, $match2);
            for ($i = 0; $i < count($match2[1]); $i++)
            {
                if (!preg_match('/^https?:\/\/\S*/i', $match2[1][$i]) && !preg_match('/^'.SmartUtility::_regExpVarToken('PHP_SELF').'/i', $match2[1][$i])) {
                    $source = preg_replace('/ background\s*=\s*"'.toRegExp($match2[1][$i]).'"/i', ' background="'.$basedir.$match2[1][$i].'"', $source);
                }
            }
            preg_match_all('/ src\s*=\s*"(\S*)"/i', $source, $match2);
            for ($i = 0; $i < count($match2[1]); $i++)
            {
                if (!preg_match('/ src\s?=\s?"\S*"/i', $match2[1][$i]) && !preg_match('/^https?:\/\/\S*/i', $match2[1][$i]) && !preg_match('/^'.SmartUtility::_regExpVarToken('PHP_SELF').'/i', $match2[1][$i])) {
                    $source = preg_replace('/ src\s*=\s*"'.toRegExp($match2[1][$i]).'"/i', ' src="'.$basedir.$match2[1][$i].'"', $source);
                }
            }
            preg_match_all("/\.src\s*=\s*'(\S*)'/i", $source, $match2);
            for ($i = 0; $i < count($match2[1]); $i++)
            {
                if (!preg_match("/\.src\s*=\s*'\S*'/i", $match2[1][$i]) && !preg_match('/^https?:\/\/\S*/i', $match2[1][$i]) && !preg_match('/^'.SmartUtility::_regExpVarToken('PHP_SELF').'/i', $match2[1][$i])) {
                    $source = preg_replace("/\.src\s?=\s?'".toRegExp($match2[1][$i])."'/i", ".src='".$basedir.$match2[1][$i]."'", $source);
                }
            }
            preg_match_all("/ url\(\"([^\"]*)\"\)/i", $source, $match2);
            for ($i = 0; $i < count($match2[1]); $i++)
            {
                if (!preg_match("/ url\(\"[^\"]*\"\)/i", $match2[1][$i]) && !preg_match('/^https?:\/\/\S*/i', $match2[1][$i]) && !preg_match('/^'.SmartUtility::_regExpVarToken('PHP_SELF').'/i', $match2[1][$i])) {
                    $source = preg_replace("/ url\(\"".toRegExp($match2[1][$i])."\"\)/i", " url(\"".$basedir.$match2[1][$i]."\")", $source);
                }
            }
            preg_match_all("/ url\('([^']*)'\)/i", $source, $match2);
            for ($i = 0; $i < count($match2[1]); $i++)
            {
                if (!preg_match("/ url\('[^']*'\)/i", $match2[1][$i]) && !preg_match('/^https?:\/\/\S*/i', $match2[1][$i]) && !preg_match('/^'.SmartUtility::_regExpVarToken('PHP_SELF').'/i', $match2[1][$i])) {
                    $source = preg_replace("/ url\('".toRegExp($match2[1][$i])."'\)/i", " url('".$basedir.$match2[1][$i]."')", $source);
                }
            }
            preg_match_all('/ href\s*=\s*"([^"]*)"/i', $source, $match2);
            for ($i = 0; $i < count($match2[1]); $i++)
            {
                if (!preg_match('/ href\s*=\s*"[^"]*"/i', $match2[1][$i]) && !preg_match('/^javascript:/i', $match2[1][$i]) && !preg_match('/^\&\#109\;\&\#97\;\&\#105\;\&\#108\;\&\#116\;\&\#111\;\&\#58\;\S*/i', $match2[1][$i]) && !preg_match('/^mailto:\S*/i', $match2[1][$i]) && !preg_match('/^https?:\/\/\S*/i', $match2[1][$i]) && !preg_match('/^'.SmartUtility::_regExpVarToken('PHP_SELF').'/i', $match2[1][$i])) {
                    $source = preg_replace('/ href\s?=\s?"'.toRegExp($match2[1][$i]).'"/i', ' href="'.$basedir.$match2[1][$i].'"', $source);
                }
            }

        }

        return SmartUtility::scanForAt($source);
    }

    /**
     * smarty outputfilter outputfilter
     *
     * @access  public
     * @static
     * @param   string  $source
     * @param   Smarty  &$templateClass
     * @return  string
     * @ignore
     */
    function outputFilter($source, &$templateClass)
    {
        global $YANA;

        if (!is_string($source)) {
            return $source;
        }        
        $basedir = $templateClass->get_template_vars('BASEDIR');

        /* add meta data to header */
        if (strpos($source, '</head>') > -1) {

            /* Create initialization for JavaScript/Ajax bridge */
            if (isset($YANA)) {
                assert('!isset($script); // Cannot redeclare var $script');
                $script = "\n        " . '<script type="text/javascript" language="javascript"><!--' . "\n" .
                '        window.yanaProfileId="' . $YANA->getId() . '";' . "\n" .
                '        window.yanaSessionName="' . @session_name() . '";' . "\n" .
                '        window.yanaSessionId="' . @session_id() . '";' . "\n" .
                '        window.yanaLanguage="' . @$_SESSION['language'] . '";' . "\n" .
                "        //--></script>";
                $source = preg_replace('/<head[^>]*>/i', '${0}' . $script, $source);
                unset($script);
            }

            /* import rss */
            $feeds = $templateClass->get_template_vars('RSS_ACTION');
            if (!empty($feeds)) {
                settype($feeds, 'array');
                $title = $templateClass->get_template_vars("PROGRAM");
                $title = htmlentities($title['TITLE']);
                foreach ($feeds as $name => $action)
                {
                    if (is_int($name)) {
                        $name = $title;
                    } else {
                        $name = htmlentities($name);
                    }
                    $rssLink = '    ' .
                    '<link rel="alternate" type="application/rss+xml"' .
                    ' title="' . $name . '" href="' .
                    SmartUtility::url("action=$action") . '"/>' .
                    "\n    ";
                    $source = str_replace("</head>", $rssLink . "</head>", $source);
                }
                unset($name, $feeds, $action, $rssLink);
            }

            /**
             * import stylesheets
             */
            assert('!isset($styleList); // Cannot redeclare var $styleList');
            $styleList = $templateClass->get_template_vars('SYSTEM_STYLE');

            assert('is_null($styleList) || is_array($styleList);');
            if (is_array($styleList) && count($styleList) > 0) {
                $styleList = array_reverse($styleList, true);
                assert('!isset($stylesheet); /* cannot redeclare variable $stylesheet */');
                foreach ($styleList as $stylesheet)
                {
                    $source = preg_replace('/^(\s*<head[^>]*>)/m', '${1}'."\n".'        '.SmartUtility::css($stylesheet), $source);
                }
                unset($stylesheet);
            }
            unset($styleList);

            /**
             * import javascripts
             */
            assert('!isset($scriptList); // Cannot redeclare var $scriptList');
            $scriptList = $templateClass->get_template_vars('SYSTEM_SCRIPT');

            assert('is_null($scriptList) || is_array($scriptList);');
            if (is_array($scriptList) && count($scriptList) > 0) {
                assert('!isset($script); /* cannot redeclare variable $script */');
                foreach ($scriptList as $script)
                {
                    $source = preg_replace('/^(\s*<\/head>)/m', '        '.SmartUtility::script($script)."\n".'${1}', $source);
                }
                unset($script);
            }
            unset($scriptList);

            /* import microsummaries */
            assert('!isset($summaryList); // Cannot redeclare var $summaryList');
            $summaryList = $templateClass->get_template_vars('MICROSUMMARY');

            assert('is_null($summaryList) || is_array($summaryList);');
            if (is_array($summaryList) && count($summaryList) > 0) {
                assert('!isset($summary); /* cannot redeclare variable $summary */');
                foreach ($summaryList as $summary)
                {
                    $source = preg_replace('/^(\s*<\/head>)/m', '        ' . SmartUtility::microsummary($summary) . "\n" . '${1}', $source);
                }
                unset($summary);
            }
            unset($summaryList);

            /* remove empty comments */
            $source = preg_replace('/\s*<\!--\s*-->\s*/s', '', $source);
        } /* end if */

        return $source;
    }

    /**
     * smarty modifier replace token
     *
     * Replace a token within a provided text.
     *
     * Note that this function replaces ALL entities found.
     * If a token refers to a non-existing value it is removed.
     *
     * Example:
     * <code>
     * // assume the token %foo% is set to 'World'
     * $text = 'Hello %foo%.';
     * // prints 'Hello World.'
     * print String::replaceToken($string);
     * </code>
     *
     * @static
     * @access  public
     * @param   string  $string
     * @return  string
     */
    function replaceToken($string)
    {
        global $YANA;

        if (!is_string($string)) {
            return $string;

        } elseif (strpos($string, YANA_UNI_DELIMITER) === false) {
            return $string;

        } else {
            if (isset($YANA)) {
                $string = SmartUtility::fillTemplate($string, $YANA->registry->options);
            }
            if (strpos($string, YANA_UNI_DELIMITER) !== false) {
                return preg_replace('/' . YANA_UNI_DELIMITER_REGEXP . '([A-Z]|\.){1,}' . YANA_UNI_DELIMITER_REGEXP . '/U', '', $string);
            } else {
                return $string;
            }
        }

    }

    /**
     * smarty modifier date
     *
     * Create HTML from a unix timestamp.
     *
     * @access  public
     * @static
     * @param   string  $time
     * @return  string
     */
    function date($time)
    {
        /* settype to INTEGER */
        $time = (int) $time;
        global $YANA;

        if (empty($time)) {
            $time = time();
        }

        if (isset($YANA)) {
            $profileTimeFormat = $YANA->getVar("PROFILE.TIMEFORMAT");
            if (!is_numeric($profileTimeFormat)) {
                $profileTimeFormat = 0;
            }
            $timeformat = $YANA->getVar("DATE.".$profileTimeFormat);
            unset($profileTimeFormat);
        } else {
            $timeformat = "j.n.Y";
        }

        $script = "";
        if (isset($YANA)) {
            $ref = array('LANGUAGE' => $YANA->getVar('LANGUAGE'));
        } else {
            $ref = array('LANGUAGE' => array());
        }

        /**
         * 1) provide javascript
         */
        if ((is_array($timeformat) && isset($timeformat['JS'])) || is_string($timeformat)) {

            $script .= '<script type="text/javascript" language="JavaScript">';
            $script .= 'date=new Date('.$time."000);document.write(";
            if (is_string($timeformat)) {
                $script .= SmartUtility::fillTemplate($timeformat, $ref);
            } else {
                $script .= SmartUtility::fillTemplate($timeformat['JS'], $ref);
            }
            $script .= ");</script>";

        }

        /**
         * 2) provide textual representation for fall back
         */
        if (is_array($timeformat) && isset($timeformat['PHP'])) {

            $timeformat['PHP'] = SmartUtility::fillTemplate($timeformat['PHP'], $ref);
            $script .= '<span class="yana_noscript">'.date($timeformat['PHP'], $time).'</span>';

        }

        return $script;

    }

    /**
     * smarty modifier entities
     *
     * Calls the PHP function htmlentities().
     * See the PHP manual for details.
     *
     * @access  public
     * @static
     * @param   string  $string
     * @return  string
     * @ignore
     */
    function entities($string)
    {
        if (is_string($string)) {
            return htmlentities($string);
        } else {
            return $string;
        }
    }

    /**
     * smarty modifier urlEncode
     *
     * Encodes a string so it can safely be used as a parameter within an URL.
     *
     * @access  public
     * @static
     * @param   string   $string  url parameter
     * @return  string
     * @ignore
     */
    function urlEncode($string)
    {
        return urlencode($string);
    }

    /**
     * smarty modifier url
     *
     * Creates an URL to the script itself from a search-string fragment.
     *
     * @access  public
     * @static
     * @param   string   $string           url parameter list
     * @param   bool     $asString         decide wether entities in string should be encoded or not
     * @param   bool     $asAbsolutePath   decide wether function should return relative or absolut path
     * @return  string
     */
    function url($string, $asString = false, $asAbsolutePath = true)
    {
        assert('is_string($string); // Wrong argument type argument "string". String expected');

        /* settype to STRING */
        $string = (string) $string;
        global $YANA;

        /**
         * 1) encode URL
         */
        preg_match_all("/(&|^)(.*)=(.*)(&|$)/U", $string, $m);
        for ($i = 0; $i < count($m[0]); $i++)
        {
            $string = str_replace($m[0][$i], $m[1][$i] . urlencode($m[2][$i]) . "=" . urlencode($m[3][$i]) . $m[4][$i], $string);
        }

        $url = "";

        /**
         * 2) create absolute path on demand
         *
         * This includes the current protocol, domain name and script path
         */
        if ($asAbsolutePath === true) {
            if (isset($_SERVER['HTTPS']) && strcasecmp($_SERVER['HTTPS'], "on") === 0) {
                $url = "https://";
            } else {
                $url = "http://";
            }
            if (isset($_SERVER['HTTP_HOST'])) {
                $url .= $_SERVER['HTTP_HOST'];
            } else {
                $url .= $_SERVER['SERVER_NAME'];
            }
            $dirname = dirname($_SERVER['PHP_SELF']);
            if ($dirname !== DIRECTORY_SEPARATOR) {
                $url .= dirname($_SERVER['PHP_SELF']). '/';
            } else {
                $url .= '/';
            }
            unset($dirname);
        }

        /**
         * 3) build URL
         *
         * This adds the script name, current session id and current profile id.
         */
        if (isset($YANA)) {

            $url .= $YANA->getVar('PHP_SELF')."?".((@session_id() != "") ? @session_name() . "=" . @session_id() : "");

            /**
             * 3.1) add custom search-string fragment
             *
             * This encodes special characters found in the fragment,
             * depending on the $asString argument.
             */
            if ($asString === true) {
                $url .= "&id=" . $YANA->getId() . "&" . $string;
            } else {
                $url .= "&amp;id=" . $YANA->getId() . "&amp;" . htmlentities($string);
            }

        } else {

            $url .= $_SERVER['PHP_SELF'] . "?" . ((@session_id() != "") ? @session_name() . "=" . @session_id() : "");

        }

        return $url;

    }

    /**
     * smarty modifier href
     *
     * Alias of SmartUtility::url()
     *
     * @access  public
     * @static
     * @param   string  $time
     * @return  string
     */
    function href($string)
    {

        return '"' . SmartUtility::url($string, false, false) . '"';

    }

    /**
     * smarty modifier microsummary
     *
     * Create a link to a microsummary identified by the $target parameter.
     *
     * @access  public
     * @static
     * @param   string  $target
     * @return  string
     */
    function microsummary($target)
    {
        /* settype to STRING */
        $url = (string) $url;
        
        /**
         * error - invalid input
         */
        if (!preg_match("/^\w[\w\d-_]*$/si", $target)) {
            return "";

        } else {

            return '<link rel="microsummary" href="'.$_SERVER['PHP_SELF'].'?action=get_microsummary&amp;target='.$target.'" type="text/plain"/>';
        }
    }

    /**
     * smarty modifier css
     *
     * @access  public
     * @static
     * @param   string  $url
     * @return  string
     */
    function css($url)
    {
        /* settype to STRING */
        $url = (string) $url;

        /**
         * error - invalid input
         */
        if (!preg_match("/^[\w-_\.\/]+\.css$/si", $url)) {
            return "";

        } else {
            return '<link rel="stylesheet" type="text/css" href="'.$url.'"/>';

        }
    }

    /**
     * smarty modifier script
     *
     * @access  public
     * @static
     * @param   string  $time
     * @return  string
     */
    function script($string)
    {
        /* settype to STRING */
        $string = (string) $string;

        /**
         * error - invalid input
         */
        if (!preg_match("/^[\w-_\.\/]+\.js$/si", $string)) {
            return "";

        } else {
            return '<script type="text/javascript" language="javascript" src="'.$string.'"></script>';

        }
    }

    /**
     * smarty modifier embeddedTags
     *
     * @access  public
     * @static
     * @param   string  $source
     * @return  string
     */
    function embeddedTags($txt)
    {
        global $YANA;
        /**
         * error - invalid argument type 'txt' (string expected)
         */
        if (!is_string($txt)) {
            return $txt;

        /**
         * if not necessary -> skip the whole section for better performance
         */
        } elseif (($offset = strpos($txt, '[')) !== false && preg_match_all('/(?<=\[)\w+/', $txt, $machtedTags)) {

            /* remove duplicates */
            $machtedTags = array_unique($machtedTags[0]);

            $invalidResource = '<img title="the resource %RESOURCE% was blocked because it contained illegal characters" alt="invalid %RESOURCE%" border="0" src="%DATADIR%icon_x.gif"/>';

            foreach ($machtedTags as $tag)
            {
                switch ($tag)
                {
                    case 'br':
                    case 'wbr':
                        /* need to be replaced LAST - ignored here */
                    break;

                    /**
                     * apply default embedded tags
                     */
                    case 'b':
                    case 'i':
                    case 'u':
                    case 'emp':
                    case 'h':
                    case 'c':
                    case 'small':
                    case 'big':
                    case 'code':
                    case 'hide':
                        $txt = preg_replace("/\[$tag\](.*)(?:\[\/$tag\]|$)/Us", '<span class="embtag_tag_' . $tag . '">${1}</span>', $txt);
                    break;

                    /**
                     * handle tag [php]
                     */
                    case 'php':
                        if (YANA_EMBTAG_ALLOW_PHP) {
                            while (preg_match('/\[php\](.*)(\[\/php\]|$)/Us', $txt, $m))
                            {
                                /* This is to avoid double quoting , since highlight_string() will quote the string again. */
                                $m[1] = html_entity_decode($m[1]);
                                $m[1] = '<span class="embtag_tag_code">' . highlight_string("<?php " . $m[1] . " ?>", true) . '</span>';
                                $txt  = str_replace($m[0], $m[1], $txt);
                            }
                            unset($m);
                        }
                    break;

                    /**
                     * handle tag [mark]
                     */
                    case 'mark':
                        $txt = preg_replace("/\[$tag\](.*)(?:\[\/$tag\]|$)/Us", '<span class="embtag_tag_' . $tag . '">${1}</span>', $txt);
                        $txt = preg_replace(
                            "/\[mark=(\w+|\#[\da-fA-F]{3}|\#[\da-fA-F]{6})\](.*)(?:\[\/mark\]|$)/Us",
                            '<span class="embtag_tag_mark" style="background-color:${1}">${2}</span>',
                            $txt
                        );
                    break;

                    /**
                     * handle tag [color]
                     */
                    case 'color':
                        $txt = preg_replace("/\[$tag\](.*)(?:\[\/$tag\]|$)/Us", '<span class="embtag_tag_' . $tag . '">${1}</span>', $txt);
                        $txt = preg_replace(
                            "/\[color=(\w+|\#[\da-fA-F]{3}|\#[\da-fA-F]{6})\](.*)(?:\[\/color\]|$)/Us",
                            '<span class="embtag_tag_color" style="color:${1}">${2}</span>',
                            $txt
                        );
                    break;

                    /**
                     * handle tag [mail]
                     */
                    case 'mail':
                        $txt = preg_replace("/\[mail=mailto:(.*)\]/Us", '[mail=${1}]', $txt);
                        $txt = preg_replace("/\[mail]mailto:(.*)\[\/mail\]/Us", '[mail]${1}[/mail]', $txt);
                        /**
                         * may contain word-/line- breaks that need to be removed
                         */
                        while (preg_match('/\[mail=[^\[\]]*\[wbr\]/i', $txt))
                        {
                            $txt = preg_replace('/(\[mail=[^\[\]]*)\[wbr\]/i', '${1}', $txt);
                        }

                        while (preg_match("/\[mail=(.*)\](.*)\[\/mail\]/Usi", $txt, $matches1) || preg_match("/\[mail\](.*)\[\/mail\]/Ui", $txt, $matches2))
                        {
                            if (!empty($matches1)) {
                                $mailMatch =& $matches1[0];
                                $mailHref  =  $matches1[1];
                                $mailText  =  $matches1[2];
                            } elseif (!empty($matches2)) {
                                $mailMatch =& $matches2[0];
                                $mailHref  =  $matches2[1];
                                $mailText  =  $matches2[1];
                                $mailText  =  preg_replace('/ ?(\[wbr\]|\[br\])/','',$mailText);
                                if (strlen($mailText) > 40) {
                                    $mailText = substr($mailText,0,34) . "..." . substr($mailText,-3);
                                }
                            }
                            /**
                             * Count in the possibility of some nasty guy injecting white space characters
                             */
                            $mailHref = preg_replace('/[\x00-\x1f]/', '', $mailHref);
                            /**
                             * Although there should not be any tags - you never know ...
                             */
                            $mailHref = strip_tags($mailHref);
                            $mailHref = preg_replace('/ ?(\[wbr\]|\[br\])/', '', $mailHref);
                            $mailHref = preg_replace('/^mailto:/i', '', $mailHref);
                            $mailHref = untaintInput($mailHref, 'mail');
                            $mailHref = htmlentities($mailHref);
                            if (!empty($mailHref)) {
                                $txt = str_replace(
                                    $mailMatch,
                                    '<a href="&#109;&#97;&#105;&#108;&#116;&#111;&#58;' . $mailHref . '" target="_blank">' . $mailText . '</a>',
                                    $txt);
                            } else {
                                $txt = str_replace($mailMatch, str_replace('%RESOURCE%', 'mail', $invalidResource), $txt);
                            }
                        } /* end while */
                    break;

                    /**
                     * handle tag [img]
                     */
                    case 'img':
                        while (preg_match("/\[img\](.*)\[\/img\]/U", $txt, $matches))
                        {
                            if (preg_match("/^[\w\d-_\/]+\.(png|jpg|gif|jpeg)$/i", $matches[1], $ext)) {
                                $txt = str_replace($matches[0], '<img alt="" border="0" src="' . htmlentities(strip_tags(preg_replace("/\[wbr\]/i", "", $matches[1]))) . '" style="max-width: 320px; max-height: 240px" onload="javascript:if(this.width>320) { this.width=320; }; if(this.height>240) { this.height=240; };"/>', $txt);
                            } else {
                                $txt = str_replace($matches[0], str_replace('%RESOURCE%', 'image', $invalidResource), $txt);
                            }
                        } /* end foreach */
                    break;

                    /**
                     * handle tag [url]
                     *
                     * may contain word-/line- breaks that need to be removed
                     */
                    case 'url':
                        while (preg_match('/\[url=[^\[\]]*(?:\[wbr\]|\[br\])/si', $txt))
                        {
                            $txt = preg_replace('/(\[url=[^\[\]]*)(?:\[wbr\]|\[br\])/si', '${1}', $txt);
                        }
            
                        while (preg_match("/\[url=(.*)\](.*)\[\/url\]/Usi", $txt, $matches1) || preg_match("/\[url\](.*)\[\/url\]/Ui", $txt, $matches2))
                        {
                            if (!empty($matches1)) {
                                $uriMatch =& $matches1[0];
                                $uriHref  =  $matches1[1];
                                $uriText  =  $matches1[2];
                            } elseif (!empty($matches2)) {
                                $uriMatch =& $matches2[0];
                                $uriHref  =  $matches2[1];
                                $uriText  =  $matches2[1];
                                $uriText  =  preg_replace('/ ?(\[wbr\])/','',$uriText);
                                if (strlen($uriText) > 40) {
                                    $uriText = substr($uriText,0,34) . "..." . substr($uriText,-3);
                                }
                            }
                            /**
                             * Count in the possibility of some nasty guy injecting white space characters
                             */
                            $uriHref = preg_replace('/[\x00-\x1f]/','',$uriHref);
                            /**
                             * Although there should not be any tags - you never know ...
                             */
                            $uriHref = strip_tags($uriHref);
                            $uriHref = preg_replace('/ ?(\[wbr\]|\[br\])/', '', $uriHref);
                            $uriHref = untaintInput($uriHref, 'url');
                            $uriHref = htmlentities($uriHref);
                            if (!preg_match('/^(https?:\/\/|ftp:\/\/)/', $uriHref) && preg_match('/^[^:]+:/', $uriHref)) {
                                $uriHref = '';
                            } elseif (!preg_match('/^[^:]+:/', $uriHref)) {
                                $uriHref = 'http://'.$uriHref;
                            }
                            if (!empty($uriHref)) {
                                $txt = str_replace($uriMatch, '<a href="' . $uriHref . '" target="_blank">' . $uriText . '</a>', $txt);
                            } else {
                                $txt = str_replace($uriMatch, str_replace('%RESOURCE%', 'uri', $invalidResource), $txt);
                            }
                            unset($uriMatch);
                            unset($uriHref);
                            unset($uriText);
                        } /* end while */
                    break;

                    /**
                     * load and apply embedded tags from system configuration
                     */
                    default:
                        if (isset($YANA)) {
                            assert('!isset($userTag); // Cannot redeclare var $userTag');
                            assert('!isset($opt); // Cannot redeclare var $opt');
                            assert('!isset($regExp); // Cannot redeclare var $regExp');
                            assert('!isset($replace); // Cannot redeclare var $replace');
                            $userTag = $YANA->getVar('PROFILE.EMBTAG');
                            if (is_array($userTag)) {
                                foreach ($userTag as $tagName => $opt)
                                {
                                    $tagName = strtolower($tagName);
                                    if (is_array($opt)) {
                                        if (isset($opt[1])) {
                                            $regExp = '/' . $opt[1] . '/Us';
                                        } else {
                                            $regExp = "/\[$tagName\](.*)(?:\[\/$tagName\]|$)/Us";
                                        }
                                        if (isset($opt[2])) {
                                            $replace = htmlspecialchars_decode($opt[2]);
                                        } else {
                                            $replace = '<span class="embtag_tag_' . $tagName . '">$1</span>';
                                        }
                                        $txt = preg_replace($regExp, $replace, $txt);
    
                                    } else {
                                        trigger_error("Ignored an invalid embedded tag. String expected, found '" . gettype($opt) . "' instead.", E_USER_NOTICE);
                                        continue;
    
                                    } /* end if */
                                } /* end foreach */
                                unset($opt, $tagName, $regExp, $replace);
                            } /* end if */
                            unset($userTag);
                        }
                    break;
                } /* end switch */
            } /* end foreach */
            unset($tag);

            /**
             * handle tag [br] (line break)
             */
            $txt = str_replace('[br]', '<br />', $txt);

            /**
             * handle tag [wbr] (word break)
             */
            $txt = str_replace('[wbr]', '<span class="embtag_tag_wbr"> </span>', $txt);

        } /* end if*/

        assert('is_string($txt); // Unexpected result: $txt is supposed to be a string.');
        return $txt;
    }

    /**
     * smarty modifier scanForAt
     *
     * @access  public
     * @static
     * @param   string  $source
     * @return  string
     */
    function scanForAt($source)
    {
        if (!is_string($source)) {
            return $source;
        } else {
            preg_match_all("/[\w\.\-_]+@[\w\.\-_]+/", $source, $matches);

            for ($i = 0; $i < count($matches[0]); $i++)
            {
                $source = str_replace($matches[0][$i], String::htmlEntities($matches[0][$i]), $source);
            }

            return $source;
        }
    }

    /**
     * smarty modifier smilies
     *
     * @access  public
     * @static
     * @param   string  $txt
     * @return  string
     */
    function smilies($txt)
    {
        assert('empty($txt) || is_string($txt); // Wrong argument type. String expected found '.gettype($txt).' instead.');
        if (!is_string($txt)) {
            return $txt;
        }

        global $YANA;
        if (isset($YANA)) {
            $smilies = $YANA->getVar('SMILIES');
            if (empty($smilies)) {
                SmartUtility::loadSmilies();
                $smilies = $YANA->getVar("SMILIES");
            }
            $smilies_dir = $YANA->getVar('PROFILE.SMILEYDIR');
        } else {
            global $smilies, $smilies_dir;
        }

        /* default smilies */
        if (isset($smilies) && is_array($smilies)) {
            if (in_array("smile", $smilies))        {
                $txt = preg_replace("/:[\-oO]?\)/", '<img alt=":-)" border="0" hspace="2" src="'.$smilies_dir.'smile.gif"/>', $txt);
            }
            if (in_array("teufelchen", $smilies))   {
                $txt = preg_replace("/>:[\-oO]?\)/", '<img alt="&gt:-)" border="0" hspace="2" src="'.$smilies_dir.'teufelchen.gif"/>', $txt);
            }
            if (in_array("boese", $smilies))        {
                $txt = preg_replace("/:[\-oO]?\(/", '<img alt=":-(" border="0" hspace="2" src="'.$smilies_dir.'boese.gif"/>', $txt);
            }
            if (in_array("keinemeinung", $smilies)) {
                $txt = preg_replace("/:[\-oO]?\|/", '<img alt=":-|" border="0" hspace="2" src="'.$smilies_dir.'keinemeinung.gif"/>', $txt);
            }
            if (in_array("lachen", $smilies))       {
                $txt = preg_replace("/:[\-oO]?D/", '<img alt=":-D" border="0" hspace="2" src="'.$smilies_dir.'lachen.gif"/>', $txt);
            }
            if (in_array("grinsen", $smilies))      {
                $txt = preg_replace("/;[\-oO]\)/", '<img alt=";-)" border="0" hspace="2" src="'.$smilies_dir.'grinsen.gif"/>', $txt);
            }
            if (in_array("zunge", $smilies))        {
                $txt = preg_replace("/:[\-oO]?P/", '<img alt=":-P" border="0" hspace="2" src="'.$smilies_dir.'zunge.gif"/>', $txt);
            }
            if (in_array("traurig", $smilies))      {
                $txt = preg_replace("/:[\'\*]\(/", '<img alt=":*(" border="0" hspace="2" src="'.$smilies_dir.'traurig.gif"/>', $txt);
            }
        }

        /* if not necessary -> skip the whole section for better performance */
        if (strpos($txt, ':') !== false) {
            /* Emot-Codes */
            for ($j = 0; $j < count($smilies); $j++)
            {
                while (preg_match("/:".$smilies[$j].":(\s|\[wbr\]|\[br\]|<br \/>)*:".$smilies[$j].":/i", $txt))
                {
                    $txt = preg_replace("/:".$smilies[$j].":(\s|\[wbr\]|\[br\]|<br \/>)*:".$smilies[$j].":/i", ':'.$smilies[$j].':', $txt);
                }
                $txt = preg_replace("/:".addcslashes($smilies[$j], "+()[]{}.?*/\\$^").":/", '<img alt="" border="0" hspace="2" src="'.$smilies_dir.$smilies[$j].'.gif"/>', $txt);
            }
        }

        assert('is_string($txt);');
        return $txt;
    }

    /**
     * smarty function updateCheck
     *
     * This checks for updates and returns the result.
     * If the server is not reachable it returns a link instead.
     *
     * Note: since version look-ups can be very time consuming
     * (e.g. if the server is slow or temporarily unreachable)
     * this function caches the results for 8 hours before
     * searching for updates again.
     *
     * @static
     * @access  public
     * @param   array   $params
     * @param   Smarty  &$smarty
     * @return  int
     * @since   2.9.11
     */
    function updateCheck($params, &$smarty)
    {
        assert('isset($GLOBALS["YANA"]); // Global var $YANA not set');
        global $YANA;
        
        /* cache results */
        $tempDir = $YANA->getVar('TEMPDIR');
        $tempFile = "{$tempDir}updatecheck." . @$_SESSION['language'] . ".tmp";
        if (is_file($tempFile) && filemtime($tempFile) > time() - 28800 /* = 8h */) {
            return file_get_contents($tempFile);
        }

        /* create link to check for new version */
        $url = $YANA->getDefault('UPDATE_SERVER');
        $url = str_replace('%VERSION%', YANA_VERSION, $url);
        $url = str_replace('%IS_STABLE%', YANA_IS_STABLE, $url);
        $url = str_replace('%LANG%', @$_SESSION['language'], $url);
        $href = str_replace('%AS_NUMBER%', '', $url);
        $url = str_replace('%AS_NUMBER%', 'true', $url);
        $url = html_entity_decode($url);
        $link = '<a href="' . $href .  '" target="_blank">' . $YANA->getVar('LANGUAGE.INDEX_13') . '</a>';

        assert('!isset($urlInfo); // Cannot redeclare var $urlInfo');
        assert('!isset($errno); // Cannot redeclare var $errno');
        assert('!isset($errstr); // Cannot redeclare var $errstr');
        assert('!isset($latestVersion); // Cannot redeclare var $latestVersion');

        /**
         * 1) try using url_fopen
         */
        if (!ini_get('allow_url_fopen') == true) {
            $latestVersion = substr(file_get_contents($url), 0, 20);
            if (empty($latestVersion)) {

                file_put_contents($tempFile, $link);
                return $link;
            }

        /**
         * 2) try to connect using a socket
         *
         * this will only be done if step 1) failed
         */
        } elseif (($urlInfo = parse_url($url)) !== false && ($fsock = @fsockopen($urlInfo['host'], 80, $errno, $errstr, 30)) != false) {

            if (!empty($errno)) {
                trigger_error('Update-check failed to open connection to server. Reason: ' . $errstr, E_USER_NOTICE);
                @fclose($fsock);
                file_put_contents($tempFile, $link);
                return $link;
            }
            unset($errno, $errstr);

            /* send request header */
            @fputs($fsock, "GET " . (isset($urlInfo['path']) ? $urlInfo['path'] : '/') . (isset($urlInfo['query']) ? '?' . $urlInfo['query'] : '') . " HTTP/1.0\r\n");
            @fputs($fsock, "HOST: " . $urlInfo['host'] . "\r\n");
            @fputs($fsock, "Connection: close\r\n\r\n");

            /* read response */
            while (!@feof($fsock))
            {
                $latestVersion .= @fread($fsock, 1024);
            }
            @fclose($fsock);
            unset($fsock);

            /* cut off header data */
            $latestVersion = preg_replace('/^.*?\r\n\r\n(.*)$/si', '$1', $latestVersion);
            $latestVersion = substr($latestVersion, 0, 20);

            /* print reply */
            if (empty($latestVersion)) {
                file_put_contents($tempFile, $link);
                return $link;
            }
            unset($urlInfo);

        /**
         * 3) return link, users will have to click it
         *
         * this will only be done if steps 1) and 2) failed
         */
        } else {
            file_put_contents($tempFile, $link);
            return $link;
        }
        
        /**
         * Compare versions and return result;
         */
        if (version_compare(YANA_VERSION, $latestVersion) < 0) {
            $link = $YANA->getVar('LANGUAGE.INDEX_15') . ': ' . htmlentities($latestVersion) . ' <a href="' . $href .  '" target="_blank">' . $YANA->getVar('LANGUAGE.INDEX_16') . '</a>';
            file_put_contents($tempFile, $link);
            return $link;

        } else {
            $link = $YANA->getVar('LANGUAGE.INDEX_14') . ': ' . YANA_VERSION;
            file_put_contents($tempFile, $link);
            return $link;
        }

    }

    /**
     * smarty function sizeOf
     *
     * This does return the size of an array or string as an integer,
     * or false on error.
     *
     * If the parameter $assign is given, the result is assigned to
     * this var and this function returns nothing.
     *
     * Expected arguments:
     * <ul>
     * <li> mixed $value = scalar value or array to count </li>
     * <li> string $assign = (optional) name of template var to assign the result to </li>
     * </ul>
     *
     * @static
     * @access  public
     * @param   array   $params
     * @param   Smarty  &$smarty
     * @return  int
     * @since   2.9.7
     */
    function sizeOf($params, &$smarty)
    {
        if (isset($params['value'])) {
            $value = $params['value'];
            if (is_scalar($value)) {
                $result = strlen("$value");

            } elseif (is_array($value)) {
                $result = count($value);

            } else {
                $result = false;
            }
        } else {
            $result = 0;
        }
        if (isset($params['assign'])) {
            if (is_string($params['assign'])) {
                $smarty->assign($params['assign'], $result);
            } else {
                $smarty->trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 'assign', 'String in function '.__FUNCTION__.'()', gettype($params['value'])), E_USER_WARNING);
            }
        } else {
            return $result;
        }
    }

    /**
     * smarty function varDump
     *
     * This does a var dump of the inputed value.
     * Applies to debug-mode only.
     *
     * @static
     * @access  public
     * @param   array   $params
     * @param   Smarty  &$smarty
     * @return  string
     */
    function varDump($params, &$smarty)
    {
        if (is_scalar(@$params['var'])) {
            return '<pre style="text-align: left">' . gettype(@$params['var']) . '(' . htmlentities(var_export(@$params['var'], true)) . ')</pre>';
        } else {
            return '<pre style="text-align: left">' . htmlentities(var_export(@$params['var'], true)) . '</pre>';
        }
    }

    /**
     * smarty function rss
     *
     * @access  public
     * @static
     * @param   array   $params
     * @param   Smarty  &$smarty
     * @return  string
     */
    function rss($params, &$smarty)
    {
        assert('isset($GLOBALS["YANA"]); // Global var $YANA not set');
        if (isset($params['image'])) {
            $image = (string) $params['image'];
        } else {
            $image = $GLOBALS['YANA']->getVar('DATADIR') .'rss.gif';
        }
        $feeds = $GLOBALS['YANA']->getVar('RSS_ACTION');
        $title = $GLOBALS['YANA']->getVar('LANGUAGE.RSS_TITLE');
        $defaultName = $GLOBALS['YANA']->getVar('LANGUAGE.PROGRAM_TITLE');
        $result = "";
        if (is_array($feeds)) {
            foreach ($feeds as $name => $action)
            {
                if (is_int($name)) {
                    $name = $defaultName;
                }
                $result .= '<a title="' . $name . ': ' . $title . '" href="' . SmartUtility::url("action=${action}") . '">' .
                '<img alt="RSS" src="' . $image . '"/></a>';
            }
        }    
        return $result;
    }

    /**
     * smarty function printArray
     *
     * @access  public
     * @static
     * @param   array   $params
     * @param   Smarty  &$smarty
     * @return  string
     */
    function printArray($params, &$smarty)
    {
        if (!isset($params['value'])) {
            return "";
        } else {
            $array = $params['value'];
        }
        if (is_string($array)) {
            $array = SML::decode($array);
        }
        if (is_array($array)) {
            $array = htmlentities(SML::encode($array));
            $array = preg_replace('/(&lt;\w[^&]*&gt;)(.*?)(&lt;\/[^&]*&gt;)$/m', '<span style="color: #35a;">$1</span>$2<span style="color: #35a;">$3</span>' , $array);
            $array = preg_replace('/&lt;[^&]+&gt;\s*$/m', '<span style="color: #607; font-weight: bold;">$0</span>' , $array);
            $array = preg_replace('/' . YANA_UNI_DELIMITER . '[\w\.-_]+' . YANA_UNI_DELIMITER . '/m', '<span style="color: #080;">$0</span>' , $array);
            $array = preg_replace('/\[\/?[\w\=]+\]/m', '<span style="color: #800;">$0</span>' , $array);
            $array = preg_replace('/&amp;\w+;/m', '<span style="color: #880;">$0</span>' , $array);
            return "<pre>${array}</pre>";
        } else {
            return $array;
        }
    }

    /**
     * smarty function print unordered list
     *
     * Print an array using a tree menu.
     *
     * The following CSS classes are used:
     * <ul>
     *  <li> ul.gui_array_list </li>
     *  <li> ul.gui_array_list &gt; li.gui_array_list </li>
     *  <li> ul.gui_array_list &gt; li.gui_array_head </li>
     *  <li> ul.gui_array_list &gt; li.gui_array_list &gt; span.gui_array_key </li>
     *  <li> ul.gui_array_list &gt; li.gui_array_list &gt; span.gui_array_value </li>
     *  <li> ul.gui_array_list &gt; li.gui_array_head &gt; span.gui_array_key </li>
     *  <li> ul.gui_array_list &gt; li.gui_array_head &gt; span.gui_array_value </li>
     * </ul>
     *
     * If you set the argument 'keys_as_href' to true,
     * the tree menu will create links with the array
     * values as link text and the array keys as hrefs.
     * Links are only created if the value is scalar,
     * which means you may create sub-menues by using
     * another array as the value.
     *
     * @access  public
     * @static
     * @param   array   $params
     * @param   Smarty  &$smarty
     * @return  string
     */
    function printUnorderedList($params, &$smarty)
    {
        /* argument $value */
        if (!isset($params['value'])) {
            return "";
        } else {
            $array = $params['value'];
        }
        if (is_string($array)) {
            $array = SML::decode($array);
        }

        /* argument $keysAsHref */
        if (empty($params['keys_as_href'])) {
            $keysAsHref = 0;
        } else {
            $keysAsHref = 1;
        }

        /* argument $layout */
        if (empty($params['layout'])) {
            $layout = 1;
        } else {
            $layout = (int) $params['layout'];
        }

        /* implementation */
        if (is_array($array)) {
            switch ($layout)
            {
                case 1:
                    return SmartUtility::printUL1($array, $keysAsHref);
                break;
                case 2:
                    return SmartUtility::printUL2($array, $keysAsHref);
                break;
                case 3:
                    return SmartUtility::printUL3($array, $keysAsHref);
                break;
                default:
                    return SmartUtility::printUL1($array, $keysAsHref);
                break;
            }
        } else {
            return $array;
        }
    }

    /**
     * Alias of SmartUtility::printUL1()
     *
     * @see  SmartUtility::printUL1()
     * @ignore
     */
    function printUL($array, $keys = 0)
    {
        SmartUtility::printUL1($array, $keys);
    }

    /**
     * smarty modifier print unordered list
     *
     * This implements layout #1.
     *
     * called by SmartUtility::printUnorderedList()
     *
     * - currently not used as a modifier
     *
     * @access  public
     * @static
     * @param   array   $array
     * @param   int     $keys
     * @return  string
     * @ignore
     */
    function printUL1($array, $keys = 0)
    {
        $ul = '<ul class="gui_array_list">';
        foreach ($array as $key => $element)
        {
            /* print key */
            if (is_array($element)) {
                $ul .= '<li class="gui_array_head" onmouseover="this.className=\'gui_array_head_open\'" onmouseout="this.className=\'gui_array_head\'">';
                $ul .= '<span class="gui_array_key">';
                $ul .= htmlentities($key);
                $ul .= '</span>';
            } else {
                $ul .= '<li class="gui_array_list">';
                if ($keys === 2) {
                    /* intentionally left blank */
                } elseif ($keys == 1 && is_scalar($element)) {
                    $ul .= '<a href="'.htmlentities($key).'">';
                } else {
                    $ul .= '<span class="gui_array_key">';
                    $ul .= htmlentities($key).':';
                    $ul .= '</span>';
                }
            }

            /* print value */
            if (is_bool($element)) {
                $ul .= '<span class="gui_array_value">';
                if ($element) {
                    $ul .= '<img alt="true" title="true" src="%DATADIR%boolean_true.gif"/>';
                } else {
                    $ul .= '<img alt="false" title="false" src="%DATADIR%boolean_false.gif"/>';
                }
                $ul .= '</span>';

            } elseif (is_string($element)) {
                $ul .= '<span class="gui_array_value">'.SmartUtility::embeddedTags(htmlentities($element)).'</span>';

            } elseif (is_scalar($element)) {
                $ul .= '<span class="gui_array_value">'.htmlentities($element).'</span>';

            } elseif (is_array($element)) {
                $ul .= SmartUtility::printUL1($element, $keys);

            } elseif (is_object($element)) {
                if (is_subclass_of($element, 'Object')) {
                    $ul .= '<span class="gui_array_value">'.htmlentities($element->toString()).'</span>';
                } else {
                    $ul .= '<span class="gui_array_value">Instance of '.getclass($element).'</span>';
                }

            } else {
                $ul .= '<span class="gui_array_value">'.htmlentities(print_r($element, true)).'</span>';
            }

            /* close open 'a' tag */
            if ($keys == 1 && is_scalar($element)) {
                $ul .= '</a>';
            }
            $ul .= '</li>';
        }
        $ul .= '</ul>';
        return $ul;
    }

    /**
     * smarty modifier print unordered list
     *
     * This implements layout #2.
     *
     * called by SmartUtility::printUnorderedList()
     *
     * - currently not used as a modifier
     *
     * @access  public
     * @static
     * @param   array   $array
     * @param   bool    $keys
     * @param   bool    $isRoot
     * @return  string
     * @ignore
     */
    function printUL2($array, $keys = false, $isRoot = true)
    {
        $ul = '<ul class="menu root">';
        foreach ($array as $key => $element)
        {
            /* print key */
            if (is_array($element)) {
                $ul .= '<li class="menu">';
                $ul .= '<div class="menu_head" onclick="yanaMenu(this)">' . htmlentities($key) . '</div>';
            } else {
                $ul .= '<li class="entry">';
                if ($keys === 2) {
                    /* intentionally left blank */
                } elseif ($keys == 1 && is_scalar($element)) {
                    $ul .= '<a href="'.htmlentities($key).'">';
                } else {
                    $ul .= '<span class="gui_array_key">';
                    $ul .= htmlentities($key).':';
                    $ul .= '</span>';
                }
            }

            /* print value */
            if (is_array($element)) {
                $ul .= SmartUtility::printUL2($element, $keys, false);
            } else {
                if ($keys == 0) {
                    $ul .= '<span class="gui_array_value">';
                }
                if (is_bool($element)) {
                    if ($element) {
                        $ul .= '<img alt="true" title="true" src="%DATADIR%boolean_true.gif"/>';
                    } else {
                        $ul .= '<img alt="false" title="false" src="%DATADIR%boolean_false.gif"/>';
                    }
                } elseif (is_scalar($element)) {
                    $ul .= htmlentities($element);
                } elseif (is_object($element)) {
                    if (is_subclass_of($element, 'Object')) {
                        $ul .= htmlentities($element->toString());
                    } else {
                        $ul .= 'Instance of '.getclass($element);
                    }
                } else {
                    $ul .= htmlentities(print_r($element, true));
                }
                if ($keys == 0) {
                    $ul .= '</span>';
                }
            }

            /* close open 'a' tag */
            if ($keys == 1 && is_scalar($element)) {
                $ul .= '</a>';
            }
            $ul .= '</li>';
        }
        $ul .= '</ul>';
        return $ul;
    }

    /**
     * smarty modifier print unordered list
     *
     * This implements layout #3.
     *
     * called by SmartUtility::printUnorderedList()
     *
     * - currently not used as a modifier
     *
     * @access  public
     * @static
     * @param   array   $array
     * @param   bool    $keys
     * @return  string
     * @ignore
     */
    function printUL3($array, $keys = false)
    {
        $ul = '<ul class="hmenu">';
        foreach ($array as $key => $element)
        {
            /* print key */
            if (is_array($element)) {
                $ul .= '<li onmouseover="yanaHMenu(this,true)" onmouseout="yanaHMenu(this,false)" class="hmenu">';
                $ul .= '<div class="menu_head">' . htmlentities($key) . '</div>';
            } else {
                $ul .= '<li class="entry">';
                if ($keys === 2) {
                    /* intentionally left blank */
                } elseif ($keys == 1 && is_scalar($element)) {
                    $ul .= '<a href="'.htmlentities($key).'">';
                } else {
                    $ul .= '<span class="gui_array_key">';
                    $ul .= htmlentities($key).':';
                    $ul .= '</span>';
                }
            }

            /* print value */
            if (is_array($element)) {
                $ul .= SmartUtility::printUL3($element, $keys);
            } else {
                if ($keys == 0) {
                    $ul .= '<span class="gui_array_value">';
                }
                if (is_bool($element)) {
                    if ($element) {
                        $ul .= '<img alt="true" title="true" src="%DATADIR%boolean_true.gif"/>';
                    } else {
                        $ul .= '<img alt="false" title="false" src="%DATADIR%boolean_false.gif"/>';
                    }
                } elseif (is_scalar($element)) {
                    $ul .= htmlentities($element);
                } elseif (is_object($element)) {
                    if (is_subclass_of($element, 'Object')) {
                        $ul .= htmlentities($element->toString());
                    } else {
                        $ul .= 'Instance of '.getclass($element);
                    }
                } else {
                    $ul .= htmlentities(print_r($element, true));
                }
                if ($keys == 0) {
                    $ul .= '</span>';
                }
            }

            /* close open 'a' tag */
            if ($keys == 1 && is_scalar($element)) {
                $ul .= '</a>';
            }
            $ul .= '</li>';
        }
        $ul .= '</ul>';
        return $ul;
    }

    /**
     * smarty function loop through elements of an array
     *
     * This is pretty much the same as Smarty's "foreach" block-function,
     * excepts for the fact that it flattens multidimensional arrays
     * to one-dimension by including the '.' seperator to create compound
     * keys.
     *
     * Example:
     * <code>
     * $array = array('a' = 1, 'b' => array('foo', 'bar'), 'c' = 2);
     *
     * [%foreach from=$array item="key" key="element"%]
     * [%$key%] = [%$element%]
     * [%/foreach%]
     *
     * the code above will output:
     * a = 1
     * b = Array
     * c = 2
     *
     * Note that 'foreach' converts $element to a string,
     * even if it is was an array.
     *
     * [%loop from=$array item="key" key="element"%]
     * $key = $element
     * [%/loop%]
     *
     * the code above will output:
     * a = 1
     * b.0 = foo
     * b.1 = bar
     * c = 2
     *
     * Note that 'loop' does not convert $element to a string
     * if it is an array, but provides a list of it's items instead.
     *
     * You should also note the difference in the writing of the token.
     * While 'foreach' writes [%$key%] the function 'loop' uses just $key.
     * </code>
     *
     * @access  public
     * @static
     * @param   array   $params
     * @param   string  $source
     * @param   Smarty  &$smarty
     * @param   mixed   &$repeat
     * @return  string
     * @ignore
     */
    function loopArray($params, $template, &$smarty, &$repeat)
    {
        if (!is_array($params['from'])) {
            return "";
        }

        if (!$repeat) {
            return SmartUtility::loop(@$params['key'], @$params['item'], $params['from'], $template);
        }
    }

    /**
     * loop through array
     *
     * called by SmartUtility::loopArray()
     *
     * @access  public
     * @static
     * @param   string  $id
     * @param   string  $name
     * @param   array   &$array
     * @param   string  $prefix
     * @return  string
     * @ignore
     */
    function loop($key, $item, &$array, &$template, $prefix = "")
    {
        $list = '';
        assert('!isset($id);      // Cannot redeclare $id');
        assert('!isset($element); // Cannot redeclare $element');
        foreach ($array as $id => $element)
        {
            if (is_array($element)) {
                $list .= SmartUtility::loop($key, $item, $element, $template, $prefix.$id.'.');
            } elseif (is_scalar($element)) {
                $li = $template;
                $li = str_replace('$'.$key, htmlentities($prefix.$id), $li);
                if ($element === true) {
                    $li = str_replace('$'.$item, 'true', $li);
                } elseif ($element === false) {
                    $li = str_replace('$'.$item, 'false', $li);
                } else {
                    $li = str_replace('$'.$item, htmlentities($element), $li);
                }    
                $list .= $li;
            }

        } /* end foreach */
        unset($id, $element);
        return $list;
    }
    /**
     * smarty function dummy function (ignore)
     *
     * @access  public
     * @static
     * @param   array   $params
     * @param   string  $source
     * @param   Smarty  &$smarty
     * @param   mixed   &$repeat
     * @return  string
     * @ignore
     */
    function isDynamic($params, $template, &$smarty, &$repeat)
    {
        return $template;
    }

    /**
     * smarty function preview
     *
     * @static
     * @access  public
     * @name    SmartUtility::preview()
     * @param   array   $params
     * @param   Smarty  &$smarty
     * @return  string
     */
    function preview($params, &$smarty)
    {
        global $YANA;
        $document = new SmartTemplate();
        if (isset($params['width'])) {
            $document->insert('WIDTH', $params['width']);
        }
        if (isset($params['height'])) {
            $document->insert('HEIGHT', $params['height']);
        }
        $document->insert('ID', uniqid('yana'));
        $document->setPath($YANA->skin->getFile("gui_preview"));
        return $document->get();
    }

    /**
     * smarty function guiEmbeddedTags
     *
     * @static
     * @access  public
     * @name    SmartUtility::guiEmbeddedTags()
     * @param   array   $params
     * @param   Smarty  &$smarty
     * @return  string
     */
    function guiEmbeddedTags($params, &$smarty)
    {
        global $YANA;

        $listOfTags = array('b','i','u','h','emp','c','small','big','hide','code','img','url','mail','color','mark','smilies');

        /* Argument 'show' */
        if (isset($params['show']) && !is_string($params['show'])) {
            $smarty->trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 'show in '.__FUNCTION__.'()', 'String', gettype($params['show'])), E_USER_WARNING);
            return "";

        } elseif (isset($params['show']) && !preg_match('/^(\w+|\||-)(,(\w+|\||-))*$/is',$params['show'])) {
            $smarty->trigger_error("Argument 'show' contains illegal characters for function ".__FUNCTION__."().", E_USER_WARNING);
            return "";

        } elseif (!isset($params['show'])) {
            $show =& $listOfTags;

        } else {
            $show = explode(',', strtolower($params['show']));

        }

        /* Argument 'hide' */
        if (!empty($params['hide']) && !is_string($params['hide'])) {
            $smarty->trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 'hide in '.__FUNCTION__.'()', 'String', gettype($params['hide'])), E_USER_WARNING);
            return "";

        } elseif (!empty($params['hide']) && !preg_match('/^[\w,]+$/is',$params['hide'])) {
            $smarty->trigger_error("Argument 'hide' contains illegal characters for function ".__FUNCTION__."().", E_USER_WARNING);
            return "";

        } elseif (empty($params['hide'])) {
            $hide = array();

        } else {
            $hide = explode(',', strtolower($params['hide']));

        }

        assert('is_array($show);');
        assert('is_array($hide);');

        $tags = array_diff($show, $hide);

        /* create document */
        $document = new smartTemplate();
        $document->insert('TAGS', $tags);
        $document->insert('USER_DEFINED', $YANA->getVar('PROFILE.EMBTAG'));
        $document->insert('LANGUAGE', $YANA->getVar('LANGUAGE'));
        $document->setPath($YANA->skin->getFile("gui_embedded_tags"));

        return $document->get();
    }

    /**
     * smarty function smlLoad
     *
     * The function sml_load() provides the same interface as
     * the built-in smarty function config_load().
     * The difference is, that sml_load() works on SML files.
     *
     * In addition you may use the scope 'template_vars', which
     * is unique for sml_load(). This scope merges the values
     * with the current array of template vars instead of copying
     * them to the config vars of the template. This enables you
     * to handle loaded SML structure as common vars and provides
     * access to nested structures inside SML files, which is
     * a feature that is not available for normal smarty config files.
     *
     * @static
     * @access  public
     * @name    SmartUtility::smlLoad()
     * @param   array   $params
     * @param   Smarty  &$smarty
     * @return  string
     */
    function smlLoad($params, &$smarty)
    {

        /* input checking*/

        if (empty($params['file'])) {
            $smarty->trigger_error("Missing argument 'file' in function sml_load()", E_USER_WARNING);
            return "";
        } elseif (!is_string($params['file'])) {
            $smarty->trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 'file in sml_load()', 'String', gettype($params['file'])), E_USER_WARNING);
            return "";
        } elseif (!preg_match('/^[\w\d-_]+\.(?:sml|config)$/i', $params['file'])) {
            $smarty->trigger_error("Argument 'file' contains illegal characters for function sml_load().", E_USER_WARNING);
            return "";
        } elseif (!is_file($smarty->config_dir.$params['file'])) {
            $smarty->trigger_error("Argument 'file' is not a valid file path in function sml_load().", E_USER_WARNING);
            return "";
        } else {
            $in_file = $smarty->config_dir.$params['file'];
        } /* end if */

        if (empty($params['section']) || $params['section'] === '*') {
            $in_section = null;
        } elseif (!is_string($params['section'])) {
            $smarty->trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 'section in sml_load()', 'String', gettype($params['section'])), E_USER_WARNING);
            return "";
        } elseif (!preg_match('/^([\w\d-_\.]+)$/i', $params['section'])) {
            $smarty->trigger_error("Argument 'section' contains illegal characters for function sml_load().", E_USER_WARNING);
            return "";
        } else {
            $in_section = $params['section'];
        } /* end if */

        if (empty($params['scope'])) {
            if (empty($params['global'])) {
                $in_scope = 'local';
            } else {
                if (!is_boolean($params['global'])) {
                    $smarty->trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 'global in sml_load()', 'Boolean', gettype($params['global'])), E_USER_WARNING);
                    return "";
                } else {
                    $smarty->trigger_error("Argument 'global' is deprecated in function sml_load(). You should consider using scope='parent' (equals global=true) or scope='local' (equals global=false) instead.", E_USER_NOTICE);
                    if ($params['global'] === true) {
                        $in_scope = 'parent';
                    } elseif ($params['global'] === false) {
                        $in_scope = 'local';
                    } else {
                        $in_scope = 'local';
                    } /* end if */
                } /* end if */
            } /* end if */
        } elseif (!is_string($params['scope'])) {
            $smarty->trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 'scope in sml_load()', 'String', gettype($params['scope'])), E_USER_WARNING);
            return "";
        } else {
            switch(strtolower($params['scope'])) {
                case 'global':
                    $in_scope = 'global';
                break;

                case 'local':
                    $in_scope = 'local';
                break;

                case 'parent':
                    $in_scope = 'parent';
                break;

                case 'template_vars':
                    $in_scope = 'template_vars';
                break;

                default:
                    $smarty->trigger_error("Argument 'scope' has an illegal value for function sml_load(). Accepted values are: 'template_vars', 'global', 'local', 'parent'.", E_USER_WARNING);
                    return "";
                break;
            } /* end switch */
        } /* end if */

        unset($params);

        /* cache handling */
        $cacheFile = $smarty->compile_dir.'sml_load'.md5($in_file).'.php';
        if ($smarty->force_compile == true || !file_exists($cacheFile) || filemtime($in_file) > filemtime($cacheFile)) {
            $forceCompile = true;
        } else {
            $forceCompile = false;
        }
        clearstatcache();

        /* business logic */
        if ($forceCompile === true) {
            assert('!isset($array);');
            $array = SML::getFile($in_file, CASE_UPPER);
            $cache = new File($cacheFile);
            $cache->read();
            if (!$cache->exists()) {
                $cache->create();
            }
            $cache->insert('<?php $array = unserialize(\''.serialize($array).'\'); ?>');
            $cache->write();
            unset($cache);
        } else {
            include($cacheFile);
        }
        if (!empty($in_section)) {
            assert('is_array($array); /* unexpected result: $array should be an array */');
            $array = Hashtable::get($array, $in_section);
        }

        /* resolve scope argument */
        if ($in_scope == 'template_vars') {
            $i_max = -1;
        } elseif ($in_scope == 'local') {
            $i_max = 0;
        } elseif ($in_scope == 'parent') {
            $i_max = 1;
        } elseif ($in_scope == 'global') {
            $i_max = count($smarty->_config);
        } else {
            $i_max = UNDEFINED;
            $smarty->_config_overwrite(null);
            return "";
        }

        /* copy values to smarty config vars */
        if ($in_scope == 'template_vars') {
            $keys = array_keys($array);
            for ($i = 0; $i < count($keys); $i++)
            {
                $smarty->assign_by_ref($keys[$i], $array[$keys[$i]], true);
            }
        } else {
            for ($i = 0; $i <= $i_max; $i++)
            {
                if (!isset($smarty->_config[$i])) {
                    $smarty->_config[$i] = array('vars' => array(), 'files' => array());
                }
                if (empty($smarty->_config[$i]['vars'])) {
                    $smarty->_config[$i]['vars'] = $array;
                } else {
                    $smarty->_config[$i]['vars'] = Hashtable::merge($smarty->_config[$i]['vars'], $array);
                }
                $smarty->_config[$i]['files'][$in_file] = true;

            }
        }

        return "";
    }

    /**
     * smarty function guiSmilies
     *
     * @static
     * @access  public
     * @name    SmartUtility::guiSmilies()
     * @param   array   $params
     * @param   Smarty  &$smarty
     * @return  string
     *
     * @see     SmartUtility::loadSmilies()
     */
    function guiSmilies($params, &$smarty)
    {
        global $YANA;
        $table   = '<table summary="smilies" class="gui_generator_smilies"><tr>';
        if (isset($YANA)) {
            $title = $YANA->getVar("LANGUAGE.TITLE_SMILIES");
            $smilies_dir = $YANA->getVar("PROFILE.SMILEYDIR");
            $smilies = $YANA->getVar("SMILIES");
        } else {
            global $smilies, $smilies_dir;
            $title = "EmotIcons";
        }
        $params['width'] = (int) $params['width'];
        if ($params['width'] < 1) {
            $params['width'] = 1;
        }

        if (empty($smilies)) {
            SmartUtility::loadSmilies();
            if (isset($YANA)) {
                $smilies = $YANA->getVar("SMILIES");
            }
        }

        $dir = htmlspecialchars($smilies_dir);
        for ($j = 0; $j < count($smilies); $j++)
        {
            $text = htmlspecialchars($smilies[$j]);
            $url = urlencode($smilies[$j]);
            if ($j % $params['width'] == 0 && $j > 0) {
                $table .= '</tr><tr>';
            }
            $table .= '<td><a title="'.$title.'" href="javascript://:'.$url.':"><img alt="'.$text.'" src="'.$dir.$text.'.gif" onmousedown="yanaAddIcon(\':'.$text.':\',event)"/></a></td>'."\n";
        }

        return $table."</tr></table>";
    }

    /**
     * load smilies
     *
     * Note: this function is called automatically by
     * the framework on start-up.
     * So there is no need to call it by yourself.
     *
     * @static
     * @access  public
     * @name    SmartUtility::loadSmilies()
     * @param   string  $user_dir (optional)
     *
     * @see     SmartUtility::guiSmilies()
     */
    function loadSmilies($user_dir = "./")
    {
        assert('is_dir($user_dir); /* The value \''.$user_dir.'\' is not a directory */');
        if (!is_string($user_dir)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($user_dir)), E_USER_WARNING);
        } else {
            global $YANA;
            if (isset($YANA)) {
                $smilies_dir = $YANA->getVar('PROFILE.SMILEYDIR');
            } else {
                global $smilies, $smilies_dir;
                $smilies_dir = $user_dir;
            }

            if (is_dir($smilies_dir)) {
                if (empty($smilies) || is_null($smilies)) {
                    $smilies = array();

                    $dir = dir($smilies_dir);
                    while ($txt = $dir -> read())
                    {
                        if (preg_match("/\.gif$/i", $txt)) {
                            $smilies[count($smilies)] = substr($txt, 0, strrpos($txt, "."));
                        }
                    }
                    $dir->close();
                    sort($smilies);
                }

                if (isset($YANA)) {
                    $YANA->setVar('SMILIES', $smilies);
                }
            } else {
                trigger_error("Unable to load smilies. The directory '${smilies_dir}' does not exist.", E_USER_WARNING);
            }
        } /* end if */
    }

    /**
     * smarty function import
     *
     * Import another template.
     * This replaces Smarty's default import function 'include'.
     *
     * In opposite to 'include' this function allows the file parameter
     * to use a relative path and does not force the template designer
     * to work with absolute paths.
     *
     * @static
     * @access  public
     * @param   array   $param
     * @param   Smarty  &$smarty
     * @return  string
     */
    function import($param, &$smarty)
    {
        if (isset($param['file'])) {

            $filename = $param['file'];
            /* input check */
            if (!is_string($filename)) {
                trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 'file', 'String', gettype($filename)), E_USER_WARNING);
                return false;
            } elseif (!is_file($filename)) {
                trigger_error("Invalid argument 'file'. The file '${filename}' does not exist.", E_USER_WARNING);
                return false;
            }

        } elseif (isset($param['id'])) {

            /* input check */
            if (!is_string($param['id'])) {
                trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 'id', 'String', gettype($param['id'])), E_USER_WARNING);
                return false;
            } else {
                $id = $param['id'];
            }

            global $YANA;
            if (isset($YANA)) {
                $filename = $YANA->skin->getFile($id);
            } else {
                $filename = "";
            }

            if (!is_file($filename)) {
                trigger_error("Invalid argument 'id'. There is not file associated to the id '${id}'.", E_USER_WARNING);
                return false;
            }

        } else {

            trigger_error("Missing argument. You need to provide either the argument 'file' or 'id'.", E_USER_WARNING);
            return false;

        }

        $document = new SmartTemplate($filename);
        unset($param['file']);
        if (count($param)>0) {
            $document->insertByReference('*', $param);
        }
        $document->insert('FILE_IS_INCLUDE', "true");
        $include = $document->get();
        return $include;
    }

    /**
     * smarty function captcha
     *
     * Inserts a captcha image into current template.
     * Note: you still need to check the value of this
     * in your script, otherwise it will have no effect.
     *
     * @static
     * @access  public
     * @param   array   $param
     * @param   Smarty  &$smarty
     * @return  string
     * @since   2.9.3
     */
    function captcha($param, &$smarty)
    {
        if (isset($param['id']) && is_string($param['id'])) {
            $id = ' id="' . htmlentities($param['id']) . '"';
        } else {
            $id = '';
        }
        $index = rand(1, 9);
        
        global $YANA;
        if (isset($YANA)) {
            $title = $YANA->getVar('LANGUAGE.SECURITY_IMAGE.DESCRIPTION');
        } else {
            $title = "";
        }

        return '<input type="hidden" name="security_image_index" value="' . $index . '"/>' .
        '<img alt="" hspace="5" src=' .
        SmartUtility::href("action=security_get_image&security_image_index=${index}") . '/>' .
        '<span class="roundbtn"><input maxlength="5" size="5"' . $id .
        ' title="' . $title . '" type="text" name="security_image"/></span>';
    }

    /**
     * create a token with quoted delimiters
     *
     * @static
     * @access  private
     * @param   string  $input
     * @return  string
     */
    function _regExpToken($input)
    {
        assert('is_string($input);');
        /* settype to STRING */
        $input = (string) $input;
        return preg_quote(YANA_LEFT_DELIMITER).$input.preg_quote(YANA_RIGHT_DELIMITER);
    }

    /**
     * create a token refering to a variable with quoted delimiters
     *
     * @static
     * @access  private
     * @param   string  $input
     * @return  string
     */
    function _regExpVarToken($input)
    {
        assert('is_string($input);');
        /* settype to STRING */
        $input = (string) $input;
        return preg_quote(YANA_LEFT_DELIMITER).'\$'.strtoupper($input).preg_quote(YANA_RIGHT_DELIMITER);
    }

    /**
     * create a token refering to a variable
     *
     * @static
     * @access  private
     * @param   string  $input
     * @return  string
     */
    function _varToken($input)
    {
        assert('is_string($input);');
        /* settype to STRING */
        $input = (string) $input;
        return YANA_LEFT_DELIMITER.'$'.strtoupper($input).YANA_RIGHT_DELIMITER;
    }

    /**
     * create a token
     *
     * @static
     * @access  private
     * @param   string  $input
     * @return  string
     */
    function _token($input)
    {
        assert('is_string($input);');
        /* settype to STRING */
        $input = (string) $input;
        return YANA_LEFT_DELIMITER.strtoupper($input).YANA_RIGHT_DELIMITER;
    }
}
// Copyright  2009  Thomas Meyer  http://www.yanaframework.net
?>