<?php
/**
 * YANA library
 *
 *
 * Software:        Yana PHP-Framework 3.0.0
 * Author:          Dipl. Inf. Thomas Meyer
 * Homepage:        http://www.yanaframework.net
 * Copyright:       (c) 2009  Thomas Meyer
 * License:         GNU GPL  http://www.gnu.org/licenses/
 *
 * This program: can be redistributed and/or modified under the
 * terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see http://www.gnu.org/licenses/.
 *
 * You may use this software for free. In case you like it, or want to
 * use it for a commercial purpose, please consider making a donation.
 * http://yanaframework.net/donate.php
 *
 * This notice MAY NOT be removed.
 *
 *
 * @package  yana
 * @license  http://www.gnu.org/licenses/gpl.txt
 *
 * @ignore
 */

/**
 * Singleton View
 *
 * A manager class to automate searching and loading of templates, belonging to view layer.
 *
 * {@internal
 *
 * The following two system vars have been added as of version 2.9.2:
 *
 * <ol>
 *   <li> %SYSTEM_TEMPLATE% = id of current base template </li>
 *   <li> %SYSTEM_INSERT% = id of current extensional template </li>
 * </ol>
 *
 * }}
 *
 * @access      public
 * @package     yana
 * @subpackage  core
 */
class SmartView extends SmartTemplate
{
    /**#@+
     * @ignore
     * @access  private
     */

    /** @var array */ var $views = array(null, null);

    /**#@-*/

    /**
     * select template by template id
     *
     * Will select the templates for frame and content of the view,
     * where $frameTemplate and $frameTemplate are the ids of the
     * selected templates.
     *
     * Returns bool(true) on success or bool(false) on error.
     *
     * Example of usage:
     * <code>
     * $GLOBALS['YANA']->view->setTemplates('index');
     * </code>
     *
     * @access  public
     * @param   string  $frameTemplate
     * @param   string  $embeddedTemplate
     * @return  bool
     */
    function setTemplates($frameTemplate = null, $embeddedTemplate = null)
    {
        global $YANA;

        if (func_num_args() === 0) {
            /**
             * 1) reset templates
             */
            $this->views = array(null, null);
        } else {
            assert('!isset($i); // Cannot redeclare variable $i');
            assert('!isset($views); // Cannot redeclare variable $views');
            $i = 0;
            $views = $this->views;
            foreach (func_get_args() as $id)
            {
                /**
                 * 2) check input
                 */
                if (!is_null($id) && (!isset($YANA) || !$YANA->skin->isId($id))) {
                    trigger_error("Invalid view name '$frameTemplate' supplied in ".__CLASS__."->".__FUNCTION__."(). View is not registered.", E_USER_WARNING);
                    return false;
                /**
                 * 3) remember setting
                 */
                } else {
                    $views[$i] = $id;
                }
                $i++;
            }
            /**
             * 4) apply changes
             */
            $this->views = $views;
            unset($i, $views);
        }
        /**
         * 5) restore associations
         */
        $this->insertByReference('SYSTEM_TEMPLATE', $this->views[0]);
        $this->insertByReference('SYSTEM_INSERT', $this->views[1]);
        return true;
    }

    /**
     * get id of selected template
     *
     * This returns the id of the template at index $i,
     * where $i=0 is the main template and $i=1 is the embedded
     * template.
     *
     * Returns bool(false) if the requested template is not set.
     *
     * Note that this function does not recurse inside a template
     * to find imported templates as well, because this would
     * include parsing the templates, what is beyond the scope
     * of this function.
     *
     * @access  public
     * @param   int  $i
     * @return  string|bool(false)
     */
    function getTemplate($i = 0)
    {
        if (isset($this->views[$i])) {
            return $this->views[$i];
        } else {
            return false;
        }
    }

    /**
     * prepare document for output
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * @access  public
     * @return  bool
     *
     * @ignore
     */
    function buildPage()
    {
        global $YANA;
        $errMessage = "The value '%s' is not a valid template id in ".__CLASS__."->".__FUNCTION__."().";
        if (isset($YANA)) {

            /* check baseView */
            if (!is_string($this->views[0])) {
                trigger_error(sprintf($errMessage, $this->views[0]), E_USER_ERROR);
                return false;
            } elseif (!$this->views[0] === '') {
                trigger_error(sprintf($errMessage, $this->views[0]), E_USER_ERROR);
                return false;
            } elseif ((false === ($baseView = $YANA->skin->getFile($this->views[0])))) {
                trigger_error(sprintf($errMessage, $this->views[0]), E_USER_ERROR);
                return false;
            } elseif (!file_exists($baseView)) {
                trigger_error(sprintf($errMessage, $this->views[0]), E_USER_ERROR);
                return false;
            } else {
                /* all fine - proceed */
            }
    
            /* check view */
            if (is_null($this->views[1])) {
                /* this argument is optional! - proceed */
            } elseif (!is_string($this->views[1])) {
                trigger_error(sprintf($errMessage, $this->views[1]), E_USER_ERROR);
                return false;
            } elseif (!$this->views[1] === '') {
                trigger_error(sprintf($errMessage, $this->views[1]), E_USER_ERROR);
                return false;
            } elseif (false === ($view = $YANA->skin->getFile($this->views[1]))) {
                trigger_error(sprintf($errMessage, $this->views[1]), E_USER_ERROR);
                return false;
            } elseif (!file_exists($view)) {
                trigger_error(sprintf($errMessage, $this->views[1]), E_USER_ERROR);
                return false;
            } else {
                /* all fine - proceed */
            }

        } else {
            if (!is_string($this->views[0]) || !file_exists($this->views[0])) {
                trigger_error(sprintf($errMessage, $this->views[0]), E_USER_ERROR);
                return false;
            } else {
                $baseView = $this->views[0];
            }
            if (is_null($this->views[1])) {
                /* this argument is optional! - proceed */
            } elseif (!is_string($this->views[1]) || !file_exists($this->views[1])) {
                trigger_error(sprintf($errMessage, $this->views[1]), E_USER_ERROR);
                return false;
            }
        } /* end if */

        if (!$this->setPath($baseView)) {
            return false;
        } else {
            return true;
        }
    }

    /**
     * output a message to browser and terminate the program
     *
     * NOTE: This function will terminate program execution
     *
     * @access  public
     * @param   string  $type      alert, error, aso.
     * @param   string  $errcode   404, 500, aso.
     * @param   string  $event     event to trigger, e.g. 'index'
     * @param   string  $template  id of template to use
     */
    function showMessage($type, $errcode, $event = "", $template = "message")
    {
        /**
         * Input checking
         */
        if (!is_string($type)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($type)), E_USER_WARNING);
            return false;
        }        
        if (!is_string($errcode)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 2, 'String', gettype($errcode)), E_USER_WARNING);
            return false;
        }
        if (!is_string($event)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 3, 'String', gettype($event)), E_USER_WARNING);
            return false;
        }
        if (!is_string($template)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 4, 'String', gettype($template)), E_USER_WARNING);
            return false;
        }

        if (!$this->setTemplates($template)) {
            trigger_error("Invalid filename supplied. Unable to create page.", E_USER_ERROR);
            exit(1);
        }

        /* argument 3 */
        if (is_null($event)) {
            $event = 'null';
        } else {
            $event = strtolower($event);
        }

        $this->buildPage();

        global $YANA;
        if (isset($YANA)) {
            $YANA->registry->mount('language_message_file');
            $YANA->setVar('ACTION', $event);
            $YANA->setVar('STDOUT.LEVEL', strtolower($type));
            if (!empty($errcode)) {
                $YANA->setVar("STDOUT.MESSAGES.0", $YANA->getVar("TEXT.$errcode"));
            }
        }

        assert('!isset($a); /* cannot redeclare variable $a */');
        assert('!isset($b); /* cannot redeclare variable $b */');
        foreach ($_GET as $a => $b)
        {
            if (is_null($this->get(strtoupper($a)))) {
                $this->insert(strtoupper($a), $b);
            }
        }
        unset($a,$b);
        assert('!isset($a); /* cannot redeclare variable $a */');
        assert('!isset($b); /* cannot redeclare variable $b */');
        foreach ($_POST as $a => $b)
        {
            if (is_null($this->get(strtoupper($a)))) {
                $this->insert(strtoupper($a), $b);
            }
        }
        unset($a, $b);
        $this->write();
        exit(0);
    }
}
// Copyright  2009  Thomas Meyer  http://www.yanaframework.net
?>