<?php
/**
 * Plugin for Yana PHP-Framework
 *
 * Software:        Yana PHP-Framework 3.0.0
 * Author:          Dipl. Inf. Thomas Meyer
 * Homepage:        http://www.yanaframework.net
 * Copyright:       (c) 2009  Thomas Meyer
 * License:         GNU GPL  http://www.gnu.org/licenses/
 *
 * This program: can be redistributed and/or modified under the
 * terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see http://www.gnu.org/licenses/.
 *
 * You may use this software for free. In case you like it, or want to
 * use it for a commercial purpose, please consider making a donation.
 * http://yanaframework.net/donate.php
 *
 * This notice MAY NOT be removed.
 *
 *
 * @author     Thomas Meyer
 * @license    http://www.gnu.org/licenses/gpl.txt
 * @package    yana
 * @subpackage plugins
 */

/**
 * automated spam protection
 *
 * @access     public
 * @package    yana
 * @subpackage plugins
 *
 * @ignore
 */
class plugin_antispam extends Plugin
{

    /**
     * Default event handler
     *
     * returns bool(true) on success and bool(false) on error
     *
     * {@internal
     *
     * The corresponding form fields are created in SmartUtility::htmlPostProcessor();
     *
     * }}
     *
     * @access  public
     * @return  bool
     * @param   string  $event  name of the called event in lower-case
     * @param   array   &$ARGS  array of arguments passed to the function
     */
    function _default($event, &$ARGS)
    {
        assert('is_string($event);');
        assert('is_array($ARGS);');
        /* check input data */
        settype($event, "string");
        if (!is_array($ARGS)) {
            settype($ARGS, "array");
        }

        /* global variables */
        global $YANA;  

        $eventType = strtolower($YANA->plugins->getEventType($event));

        /**
         * 0) prepare settings
         */
        $settings = $YANA->getVar('PROFILE.SPAM');
        $permission = $YANA->getVar('PERMISSION');

        /**
         * Register ouput filter if option has been activated
         */
        if (!empty($settings['FORM_ID'])) {
            /**
             * Note a difference between PHP 4 and 5 here.
             * PHP 5 returns and assigns the object by reference.
             * PHP 4 returns the object by reference but then
             * assigns it by value.
             *
             * It is important to use '=&' to get the same
             * behaviour for both versions.
             */
            $smarty =& $YANA->view->getSmarty();
            $smarty->register_outputfilter(array(__CLASS__, '_outputFilter'));
        }

        /**
         * 1) Permission setting
         *
         * {@internal
         *
         * Let's say: a robot would'nt have been able to log in,
         * because it would have had to defeat the login form
         * which is also protected by this implementation.
         * In addition if it DID conquer the login form already,
         * then it is quite obvious it would conquer them all.
         *
         * Thus, the "permission" option allows to disable this
         * check, if the user is already logged in.
         *
         * }}
         */
        if (empty($settings['PERMISSION']) && is_int($permission) && $permission > 0) {
            return true;

        } else if ($eventType === 'write') {

            /**
             * 2) create header information for review by administrator
             *
             * Note: remote address is logged automatically - no need to add this here
             */
            $header_data = array();
            if (isset($_SERVER['HTTP_REFERER'])) {
                $header_data['REFERER'] = $_SERVER['HTTP_REFERER'];
            } else {
                $header_data['REFERER'] = '-';
            }
            if (isset($_SERVER['REQUEST_METHOD'])) {
                $header_data['REQUEST_METHOD'] = $_SERVER['REQUEST_METHOD'];
            } else {
                $header_data['REQUEST_METHOD'] = '-';
            }
            if (isset($_SERVER['HTTP_USER_AGENT'])) {
                $header_data['USER_AGENT'] = $_SERVER['HTTP_USER_AGENT'];
            } else {
                $header_data['USER_AGENT'] = '-';
            }
            if (isset($_SERVER['REMOTE_PORT'])) {
                $header_data['PORT'] = $_SERVER['REMOTE_PORT'];
            } else {
                $header_data['PORT'] = '-';
            }

            /**
             * 3) Form-id setting
             *
             * check the time the user required to fill out the form
             */
            if (!empty($settings['FORM_ID'])) {
                $time =& $_SESSION['transaction_isolation_created'];
                /**
                 * 3.1) if the time is not set or = 0,
                 * then we assume the form has never been displayed and we just encountered a direct request
                 */
                if (empty($time)) {

                    if (!empty($settings['LOG'])) {
                        $YANA->report(new Log('SPAM: blocked entry because no timestamp present (possibly the form was never displayed).',$header_data));
                    }
                    $YANA->report( new Alert('SPAMPROTECT'));
                    $YANA->exitTo('');
                    return false;

                /**
                 * 3.2) assume a user would need at least 5 seconds and at most 30 minutes to fill the form
                 */
                } else if (time() - $time < 5) {

                    if (!empty($settings['LOG'])) {
                        $YANA->report(new Log('SPAM: blocked entry because a previous entry has been issued within the last 5 seconds.',$header_data));
                    }
                    $YANA->report( new Alert('SPAMPROTECT_TIME'));
                    $YANA->exitTo('');
                    return false;

                } else if (time() - $time > 1800) {

                    if (!empty($settings['LOG'])) {
                        $YANA->report(new Log('SPAM: blocked entry because maximum time of life (30 minutes) for the form has been exceeded.',$header_data));
                    }
                    $YANA->report( new Alert('SPAMPROTECT_TIMEOUT'));
                    $YANA->exitTo('');
                    return false;

                /**
                 * 3.3) if an invisible field has been filled,
                 * then we assume that the request was not send via a web browser
                 */
                } else if (!empty($ARGS['yana_url'])) {

                    if (!empty($settings['LOG'])) {
                        $YANA->report(new Log('SPAM: blocked entry because a field that is not visible to human visitors has been filled.',$header_data));
                    }
                    $YANA->report( new Alert('SPAMPROTECT'));
                    $YANA->exitTo('');
                    return false;

                /**
                 * 4.1) check if input has a valid form id
                 */
                } else if ($YANA->getVar('DISABLE_FORM_ID') !== true) {
                    if (!isset($_SESSION['yana_form_id']) || $_SESSION['yana_form_id'] === 'expired' || empty($ARGS['yana_form_id']) || strcasecmp($ARGS['yana_form_id'], $_SESSION['yana_form_id']) !== 0) {
                        if (!empty($settings['LOG'])) {
                            $YANA->report(new Log('SPAM: blocked entry because no valid form Id has been found.'));
                        }
                        if ($_SESSION['yana_form_id'] === 'expired') {
                            $YANA->report( new Alert('NOSPAM'));
                            $YANA->exitTo('');
                        } else {
                            $YANA->report( new Alert('INPUT_INVALID'));
                            $YANA->exitTo('');
                        }
                        $_SESSION['yana_form_id'] = 'expired';
                        return false;
                    } else {
                        $_SESSION['yana_form_id'] = 'expired';
                    }
                } else {
                    $_SESSION['yana_form_id'] = 'expired';
                }
            } /* end if */

            /**
             * 4) header setting
             *
             * The following was added in version 2.8.6
             * This checks the header sent by the user for suspicious combinations of data.
             */
            if (!empty($settings['HEADER'])) {
                if (!defined('STDIN') && isset($_SERVER['REQUEST_METHOD']) && empty($_COOKIE) && !empty($_SERVER['HTTP_REFERER']) && empty($_SERVER['HTTP_USER_AGENT'])) {

                    if (!empty($settings['LOG'])) {
                        $YANA->report(new Log("SPAM: blocked entry because of suspicious header. 'Referer' is set while 'user agent' is missing.",$header_data));
                    }
                    $YANA->report( new Alert('SPAMPROTECT'));
                    $YANA->exitTo('');
                    return false;

                }
            }

            /**
             * 5) bad word and regular expressions filter
             */
            if (!empty($settings['WORD_FILTER']) && !empty($settings['WORDS'])) {
                if (!empty($settings['REG_EXP'])) {
                    $words = join('|', $settings['WORDS']);
                    $words = html_entity_decode($words);
                    $words = str_replace('||', '|', $words);
                    if (@preg_match("/${words}/Usi", print_r($ARGS, true), $m)) {

                        if (!empty($settings['LOG'])) {
                            $YANA->report(new Log("SPAM: blocked entry because a blacklisted phrase '" . $m[0]  . "' has been found.",$header_data));
                        }
                        $YANA->report( new Alert('SPAMPROTECT'));
                        $YANA->exitTo('');
                        return false;

                    }
                    unset($words);

                } else {
                    $string = print_r($ARGS, true);
                    foreach ($settings['WORDS'] as $words)
                    {
                        $words = html_entity_decode($words);
                        if (!empty($words) && stripos($string, $words) !== false) {

                            if (!empty($settings['LOG'])) {
                                $YANA->report(new Log("SPAM: blocked entry because a blacklisted phrase '${words}' has been found.",$header_data));
                            }
                            $YANA->report( new Alert('SPAMPROTECT'));
                            $YANA->exitTo('');
                            return false;

                        }
                    }
                    unset($words);
                }
            }

            /**
             * 6) clean up the garbage
             */
            if (isset($ARGS['yana_form_id'])) {
                unset($ARGS['yana_form_id']);
            }
            return true;

        } else {

            return true;

        }
    }

    /**
     * smarty outputfilter outputfilter
     *
     * @access  public
     * @static
     * @param   string  $source
     * @return  string
     */
    function _outputFilter($source)
    {
        global $YANA;
        /* Create form id */
        if ($YANA->getVar('DISABLE_FORM_ID') !== true) {
            $yana_form_id = md5(uniqid(rand(), true));
            if (strpos($source, "</form>") !== false) {
                /* insert form id */
                $source   = str_replace("</form>", "<span class=\"yana_button\"><input type=\"text\" name=\"yana_form_id\" value=\"$yana_form_id\" /></span>\n</form>", $source);
                $_SESSION['yana_form_id'] = $yana_form_id;
                $YANA->setVar('DISABLE_FORM_ID', true);
            }
        }
        return $source;
    }

}
// Copyright  2009  Thomas Meyer  http://www.yanaframework.net
?>