<?php
/**
 * Plugin for Yana PHP-Framework
 *
 * A Plugin for managing and billing of projects.
 *
 *
 * Software:        Yana PHP-Framework 3.0.0
 * Author:          Dipl. Inf. Thomas Meyer
 * Homepage:        http://www.yanaframework.net
 * Copyright:       (c) 2009  Thomas Meyer
 * License:         GNU GPL  http://www.gnu.org/licenses/
 *
 * This program: can be redistributed and/or modified under the
 * terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see http://www.gnu.org/licenses/.
 *
 * You may use this software for free. In case you like it, or want to
 * use it for a commercial purpose, please consider making a donation.
 * http://yanaframework.net/donate.php
 *
 * This notice MAY NOT be removed.
 *
 *
 * @author      Thomas Meyer
 * @package     plugins
 */

/**
 * plugin class "plugin_project"
 *
 * @access      public
 * @package     plugins
 * @subpackage  project
 */
class plugin_project extends Plugin
{

    /**
     * Connection to data source (API)
     *
     * Is created by a factory function and can be either a connection to "some" database or a text-file,
     * depending on user settings. The correct type will be determined automatically. Take a look at the
     * documentation of class 'DbStream' for details on this API. (all alternative classes implement
     * the same interface, regardless whether the source is a file or a databse)
     *
     * {@internal
     *
     * NOTE:
     * As of PHP 5 the notation "var $database" became deprecated, resulting in an E_STRICT_NOTICE.
     * So you may want to change this to "private $database" for PHP 5. Just be aware that doing so will
     * break backwards compatibility with PHP 4.
     *
     * }}
     *
     * @access  private
     * @var     object serializeableObject  Database-API with Query-Builder (also works with text-files)
     *
     * @ignore
     */
    var $database;

    /**
     * Constructor
     *
     * @access  public
     * @name    plugin_project::__construct()
     * @param   string  $name  a name for this plugin
     */
    function plugin_project($name)
    {
        assert('is_string($name);');

        $this->name = $name;
        global $YANA;
        $this->database = $YANA->connect("project");
    }

    /**
     * Default event handler
     *
     * returns bool(true) on success and bool(false) on error
     *
     * @access  public
     * @param   string  $event  name of the called event in lower-case
     * @param   array   $ARGS   array of arguments passed to the function
     * @return  bool
     * @name    plugin_project::_default()
     */
    function _default($event, $ARGS)
    {
        /* NOTE:
         * Fnction "assert()" is a helper for debugging your plugins.
         * It evaluates the given string as an PHP-expression and issues
         * an E_NOTICE to screen / log-file if it evaluates to bool(false).
         *
         * "Assertions" can be turned off if your application has reached
         * production stage by setting the appropriate debug-constant in
         * index.php
         */

        /* check input data */
        assert('is_string($event);');
        assert('is_array($ARGS);');
        settype($event, "string");
        settype($ARGS, "array");
        $event = strtolower($event);

        /* global variables */
        global $YANA;

        /* do something */

        /* NOTE: event handlers should always return a boolean value.
         * Return "true" on success or "false" on error.
         */
        return true;
    }

    /**
     * project_list
     *
     * returns bool(true) on success and bool(false) on error
     *
     * Type:        read
     * Permission:  1
     * Templates:   PROJECT_LIST
     *
     * @access  public
     * @param   array  $ARGS  array of params passed to the function
     * @return  bool
     * @name    plugin_project::project_list()
     */
    function project_list ($ARGS)
    {
        /* check input data */
        assert('is_array($ARGS);');
        settype($ARGS, "array");

        /* global variables */
        global $YANA;
        /* get entries from database */
        $query = array(
        'key' => 'project.*',
        'order_by' => 'project_created',
        'desc' => true,
        'limit' => 50
        );
        $rows = $this->database->get($query);
        $YANA->setVar('PROJECT', $rows);
        return true;
    }

    /**
     * calculate sum of hours spent on a project
     *
     * returns bool(true) on success and bool(false) on error
     *
     * Type:        read
     * Permission:  1
     * Templates:   NULL
     *
     * @access  public
     * @param   array  $ARGS  array of params passed to the function
     * @return  bool
     * @name    plugin_project::project_list()
     */
    function project_sum ($ARGS)
    {
        /* check input data */
        assert('is_array($ARGS);');
        settype($ARGS, "array");

        /* global variables */
        global $YANA;

        if (isset($ARGS['target']) && is_numeric($ARGS['target'])) {
            $projectId = (int) $ARGS['target'];
            /* get entries from database */
            $query = array(
            'key' => 'effort.*.effort_duration',
            'where' => array(
            array('project_id', $projectId, '='),
            array('effort_state', 1, '!=')
            )
            );
            $rows = $this->database->get($query);
            if (empty($rows)) {

                $YANA->language->read('project');
                $description = $YANA->getVar('LANGUAGE.PRJ.SUM.HOURS');
                $values = array('SUM' => 0, 'COUNT' => 0);
                $description = SmartUtility::fillTemplate($description, $values);

                exit($description);
            } else {
                $sum = array_sum($rows);
                if (is_numeric($sum)) {
                    $sum = (float) $sum;
                    $loan = $this->database->get("project.${projectId}.project_loan");
                    if (is_numeric($loan)) {
                        $loan = (float) $loan;

                        $YANA->language->read('project');
                        $description = $YANA->getVar('LANGUAGE.PRJ.SUM.HOURS') . ",\n" .
                        $YANA->getVar('LANGUAGE.PRJ.SUM.LOAN') . ",\n" .
                        $YANA->getVar('LANGUAGE.PRJ.SUM.TOTAL');
                        $values = array('SUM' => $sum, 'COUNT' => count($rows), 'LOAN' => $loan, 'TOTAL' => ($sum * $loan));
                        $description = SmartUtility::fillTemplate($description, $values);

                        exit($description);
                    } else {

                        $YANA->language->read('project');
                        $description = $YANA->getVar('LANGUAGE.PRJ.SUM.HOURS');
                        $values = array('SUM' => $sum, 'COUNT' => count($rows));
                        $description = SmartUtility::fillTemplate($description, $values);

                        exit($description);
                    }
                } else {
                    exit('Error: invalid data.');
                }
            }
        } else {
            exit('Error: invalid input.');
        }
    }

/* BEGIN auto-generated Event-handlers */

    /**
     * project_write_edit_project
     *
     * returns bool(true) on success and bool(false) on error
     *
     * Type:        write
     * Permission:  75
     * Templates:   MESSAGE
     *
     * @access  public
     * @param   array  $ARGS  array of params passed to the function
     * @return  bool
     * @name    plugin_project::project_write_edit_project()
     */
    function project_write_edit_project ($ARGS)
    {
        /* check input data */
        assert('is_array($ARGS);');
        settype($ARGS, "array");

        /* global variables */
        global $YANA;


        $updatedEntries = FormCreator::getFormdata(FORM_EDIT, $this->database->structure, 'project');

        /* no data has been provided */
        if (empty($updatedEntries)) {
            $YANA->report(new Warning('INPUT_INVALID'));
            return false;
        }

        foreach ($updatedEntries as $id => $entry)
        {
            $id = strtolower($id);

            /* before doing anything, check if entry exists */
            if (!$this->database->exists("project.${id}")) {

                /* error - no such entry */
                $YANA->report(new Warning('INPUT_INVALID'));
                return false;

            /* update the row */
            } else if (!$this->database->update("project.${id}", $entry)) {
                /* error - unable to perform update - possibly readonly */
                return false;
            }
        } /* end for */
        /* commit changes */
        return $this->database->write();
    }


    /**
     * project_write_delete_project
     *
     * returns bool(true) on success and bool(false) on error
     *
     * Type:        write
     * Permission:  75
     * Templates:   MESSAGE
     *
     * @access  public
     * @param   array  $ARGS  array of params passed to the function
     * @return  bool
     * @name    plugin_project::project_write_delete_project()
     */
    function project_write_delete_project ($ARGS)
    {
        /* check input data */
        assert('is_array($ARGS);');
        settype($ARGS, "array");

        /* global variables */
        global $YANA;

        /* check if user forgot to mark at least 1 row */
        if (empty($ARGS['selected_entries']) || !is_array($ARGS['selected_entries'])) {
            $YANA->report(new Warning('INPUT_INVALID'));
            return false;
        }
        /* remove entry from database */
        foreach ($ARGS['selected_entries'] as $id)
        {
            if (!$this->database->remove("project.${id}")) {
                /* entry does not exist */
                $YANA->report(new Warning('INPUT_INVALID'));
                return false;
            }
        } /* end for */
        /* commit changes */
        return $this->database->write();
    }

    /**
     * project_write_new_project
     *
     * returns bool(true) on success and bool(false) on error
     *
     * Type:        write
     * Permission:  0
     * Templates:   MESSAGE
     *
     * @access  public
     * @param   array  $ARGS  array of params passed to the function
     * @return  bool
     * @name    plugin_project::project_write_new_project()
     */
    function project_write_new_project ($ARGS)
    {
        /* check input data */
        assert('is_array($ARGS);');
        settype($ARGS, "array");

        /* global variables */
        global $YANA;


        $newEntry = FormCreator::getFormdata(FORM_NEW, $this->database->structure, 'project');

        /* no data has been provided */
        if (empty($newEntry)) {
            $YANA->report(new Warning('INPUT_INVALID'));
            return false;
        }

        /* insert new entry into table */
        if (!$this->database->insert("project.*", $newEntry)) {
            $YANA->report(new Warning('INPUT_INVALID'));
            return false;
        } else {
            return $this->database->write();
        }
    }

    /**
     * project_write_edit_effort
     *
     * returns bool(true) on success and bool(false) on error
     *
     * Type:        write
     * Permission:  75
     * Templates:   MESSAGE
     *
     * @access  public
     * @param   array  $ARGS  array of params passed to the function
     * @return  bool
     * @name    plugin_project::project_write_edit_effort()
     */
    function project_write_edit_effort ($ARGS)
    {
        /* check input data */
        assert('is_array($ARGS);');
        settype($ARGS, "array");

        /* global variables */
        global $YANA;


        $updatedEntries = FormCreator::getFormdata(FORM_EDIT, $this->database->structure, 'effort');

        /* no data has been provided */
        if (empty($updatedEntries)) {
            $YANA->report(new Warning('INPUT_INVALID'));
            return false;
        }

        foreach ($updatedEntries as $id => $entry)
        {
            $id = strtolower($id);

            /* before doing anything, check if entry exists */
            if (!$this->database->exists("effort.${id}")) {

                /* error - no such entry */
                $YANA->report(new Warning('INPUT_INVALID'));
                return false;

            /* update the row */
            } else if (!$this->database->update("effort.${id}", $entry)) {
                /* error - unable to perform update - possibly readonly */
                return false;
            }
        } /* end for */
        /* commit changes */
        return $this->database->write();
    }


    /**
     * project_write_delete_effort
     *
     * returns bool(true) on success and bool(false) on error
     *
     * Type:        write
     * Permission:  75
     * Templates:   MESSAGE
     *
     * @access  public
     * @param   array  $ARGS  array of params passed to the function
     * @return  bool
     * @name    plugin_project::project_write_delete_effort()
     */
    function project_write_delete_effort ($ARGS)
    {
        /* check input data */
        assert('is_array($ARGS);');
        settype($ARGS, "array");

        /* global variables */
        global $YANA;

        /* check if user forgot to mark at least 1 row */
        if (empty($ARGS['selected_entries']) || !is_array($ARGS['selected_entries'])) {
            $YANA->report(new Warning('INPUT_INVALID'));
            return false;
        }
        /* remove entry from database */
        foreach ($ARGS['selected_entries'] as $id)
        {
            if (!$this->database->remove("effort.${id}")) {
                /* entry does not exist */
                $YANA->report(new Warning('INPUT_INVALID'));
                return false;
            }
        } /* end for */
        /* commit changes */
        return $this->database->write();
    }

    /**
     * project_write_new_effort
     *
     * returns bool(true) on success and bool(false) on error
     *
     * Type:        write
     * Permission:  0
     * Templates:   MESSAGE
     *
     * @access  public
     * @param   array  $ARGS  array of params passed to the function
     * @return  bool
     * @name    plugin_project::project_write_new_effort()
     */
    function project_write_new_effort ($ARGS)
    {
        /* check input data */
        assert('is_array($ARGS);');
        settype($ARGS, "array");

        /* global variables */
        global $YANA;


        $newEntry = FormCreator::getFormdata(FORM_NEW, $this->database->structure, 'effort');

        /* no data has been provided */
        if (empty($newEntry)) {
            $YANA->report(new Warning('INPUT_INVALID'));
            return false;
        }

        /* insert new entry into table */
        if (!$this->database->insert("effort.*", $newEntry)) {
            $YANA->report(new Warning('INPUT_INVALID'));
            return false;
        } else {
            return $this->database->write();
        }
    }


/* END auto-generated Event-handlers */

/* BEGIN auto-generated functions */

    /* NOTE:
     * By convention you can hide any member function from the plugin manager
     * by using an underscore '_' as the first character of the function's name.
     * This results in: any such function (event) can not be triggered directly
     * by the user through pointing the "action" parameter in the url to that
     * function name, like: index.php?action=function_name
     *
     * (Using underscores is a common PHP4 notation for "private")
     *
     * This also means, if for some reason you need to create a public event handler
     * whose name somehow has to begin with an underscore, you need to declare this
     * within the class' default event handler.
     */

    /**
     * remove tags from user form data
     *
     * @static
     * @access  private
     * @param   array   $input  associative array containing user form data
     * @name    plugin_project::_untaintInput()
     *
     * @ignore
     */
    function _untaintInput(&$input)
    {
        if (is_array($input)) {
            foreach ($input as $a => $b)
            {
                $input[$a] = untaintInput($b, "string", 0, YANA_ESCAPE_USERTEXT);
            }
        }
    }

/* END auto-generated functions */

}
// Copyright  2009  Thomas Meyer  http://www.yanaframework.net
?>